/* $Id: entity.c,v 1.1.1.1 2002/07/07 14:03:34 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * Entities are string objects with the following operations:
 *	* new		- create a new entity
 *	* rewind	- rewind to beginning
 *	* getch		- get next wchar_t in entity
 *	* eof		- return true on EOF
 *	* free		- dispose of a created entity
 *
 * Note that (wchar_t)-1 may be returned for getch when at EOF.
 * However, this is just a placeholder character; EOF should be
 * detected via the eof operation.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "utf8.h"
#include "entity.h"

static	void	no_rewind(struct entity *);
static	int	no_eof(struct entity *);
static	const char *no_name(struct entity *);
static	void	no_setenc(struct entity *, const char *);

static	void	mem_rewind(struct entity *);
static	wchar_t	mem_getch(struct entity *);
static	int	mem_eof(struct entity *);
static	void	mem_free(struct entity *);

static	void	utf8mem_rewind(struct entity *);
static	wchar_t	utf8mem_getch(struct entity *);
static	int	utf8mem_eof(struct entity *);
static	void	utf8mem_free(struct entity *);

static	wchar_t	eof_getch(struct entity *);
static	void	eof_free(struct entity *);
static	const char *eof_name(struct entity *);

static	void	file_rewind(struct entity *);
static	wchar_t	file_getch(struct entity *);
static	int	file_eof(struct entity *);
static	void	file_free(struct entity *);
static	const char *file_name(struct entity *);

/* function form of a macro so it can be used as a free function */
void
entity_decref(e)
	struct entity *e;
{
	ENTITY_DECREF(e);
}

/*------------------------------------------------------------
 * no-op functions
 */

static void
no_rewind(e)
	struct entity *e;
{
}

static int
no_eof(e)
	struct entity *e;
{
	return 0;
}

static const char *
no_name(e)
	struct entity *e;
{
	return NULL;
}

static void
no_setenc(e, enc)
	struct entity *e;
	const char *enc;
{
}

/*------------------------------------------------------------
 * Memory entity of UCS-4 (wchar_t) characters
 */
struct mementity {
	struct entity	e;
	wchar_t *start, *end, *pos;
	int isstatic;
};

static struct entityop mem_ops = {
	mem_rewind,
	mem_getch,
	mem_eof,
	mem_free,
	no_name,
	no_setenc
};

static void
mem_rewind(e)
	struct entity *e;
{
	struct mementity *m = (struct mementity *)e;

	m->pos = m->start;
}

static int
mem_eof(e)
	struct entity *e;
{
	struct mementity *m = (struct mementity *)e;

	return m->pos >= m->end;
}

static void
mem_free(e)
	struct entity *e;
{
	struct mementity *m = (struct mementity *)e;

	if (!m->isstatic)
		free(m->start);
	free(m);
}

static wchar_t
mem_getch(e)
	struct entity *e;
{
	struct mementity *m = (struct mementity *)e;

	if (m->pos >= m->end)
		return -1;
	return *m->pos++;
}

struct entity *
entity_mem(start, end, isstatic)
	const wchar_t *start, *end;
	int isstatic;
{
	struct mementity *m;

	m = (struct mementity *)malloc(sizeof (struct mementity));
	if (!m)
		return NULL;
	ENTITY_INIT(&m->e, &mem_ops);
	if (isstatic) {
		m->start = (wchar_t *)start;
		m->end = (wchar_t *)end;
	} else {
		m->start = (wchar_t *)malloc((end - start) * sizeof (wchar_t));
		m->end = m->start + (end - start);
		memcpy(m->start, start, (end - start) * sizeof (wchar_t));
	}
	m->isstatic = isstatic;
	m->pos = m->start;
	return (struct entity *)m;
}

/*------------------------------------------------------------
 * UTF8 memory entity; data encoded using UTF8
 */
struct utf8mementity {
	struct entity	e;
	char *start, *end, *pos;
	int isstatic;
};

static struct entityop utf8mem_ops = {
	utf8mem_rewind,
	utf8mem_getch,
	utf8mem_eof,
	utf8mem_free,
	no_name,
	no_setenc

};

static void
utf8mem_rewind(e)
	struct entity *e;
{
	struct utf8mementity *m = (struct utf8mementity *)e;

	m->pos = m->start;
}

static int
utf8mem_eof(e)
	struct entity *e;
{
	struct utf8mementity *m = (struct utf8mementity *)e;

	return m->pos >= m->end;
}

static void
utf8mem_free(e)
	struct entity *e;
{
	struct utf8mementity *m = (struct utf8mementity *)e;

	if (!m->isstatic)
		free(m->start);
	free(m);
}

static wchar_t
utf8mem_getch(e)
	struct entity *e;
{
	struct utf8mementity *m = (struct utf8mementity *)e;

	if (m->pos >= m->end)
		return -1;
	return utf8_getn((const char **)&m->pos, m->end - m->pos);
}

struct entity *
entity_utf8mem(start, end, isstatic)
	const char *start, *end;
	int isstatic;
{
	struct utf8mementity *m;

	m = (struct utf8mementity *)malloc(sizeof (struct utf8mementity));
	if (!m)
		return NULL;
	ENTITY_INIT(&m->e, &utf8mem_ops);
	if (isstatic) {
		m->start = (char *)start;
		m->end = (char *)end;
	} else {
		m->start = malloc((end - start) * sizeof (char));
		m->end = m->start + (end - start);
		memcpy(m->start, start, (end - start) * sizeof (char));
	}
	m->isstatic = isstatic;		/* don't free(start) when done */
	m->pos = m->start;
	return (struct entity *)m;
}

/*------------------------------------------------------------
 * EOF entity: an infinite source of EOFs (-1)
 */
static struct entityop eof_ops = {
	no_rewind,
	eof_getch,
	no_eof,
	eof_free,
	eof_name
};

static wchar_t
eof_getch(e)
	struct entity *e;
{
	return -1;
}

static void
eof_free(e)
	struct entity *e;
{
	free(e);
}

static const char *
eof_name(e)
	struct entity *e;
{
	return "EOF";
}

struct entity *
entity_eof()
{
	struct entity *e;

	e = (struct entity *)malloc(sizeof (struct entity));
	if (!e)
		return NULL;
	ENTITY_INIT(e, &eof_ops);
	return e;
}

/*------------------------------------------------------------
 * file: auto-type detecting file descriptor
 */
struct fileentity {
	struct entity	e;
	FILE	*f;
	char	*name;
	enum {  UCS4BE, UCS4LE, UCS4U1, UCS4U2, 
		UTF16LE, UTF16BE, UTF8 }	encoding;
	int	lapos;
	int	buf[4];
};

static struct entityop file_ops = {
	file_rewind,
	file_getch,
	file_eof,
	file_free,
	file_name
};

static void
file_rewind(e)
	struct entity *e;
{
	struct fileentity *fe = (struct fileentity *)e;
	int i;

	rewind(fe->f);

	for (i = 0; i < 4; i++)
		fe->buf[i] = fgetc(fe->f);
	fe->lapos = 0;

#define CHK(buf, a0, a1, a2, a3) \
		((buf)[0] == a0 && (buf)[1] == a1 && \
		 (buf)[0] == a2 && (buf)[1] == a3)

	if        (CHK(fe->buf, 0x00, 0x00, 0xfe, 0xff)) {
		fe->encoding = UCS4BE;
		fe->lapos = 4;
	} else if (CHK(fe->buf, 0xff, 0xfe, 0x00, 0x00)) {
		fe->encoding = UCS4LE;
		fe->lapos = 4;
	} else if (CHK(fe->buf, 0x00, 0x00, 0xff, 0xfe)) {
		fe->encoding = UCS4U1;
		fe->lapos = 4;
	} else if (CHK(fe->buf, 0xfe, 0xff, 0x00, 0x00)) {
		fe->encoding = UCS4U2;
		fe->lapos = 4;
	} else if (fe->buf[0] == 0xfe && fe->buf[1] == 0xff) {
		fe->encoding = UTF16BE;
		fe->lapos = 2;
	} else if (fe->buf[0] == 0xff && fe->buf[1] == 0xfe) {
		fe->encoding = UTF16LE;
		fe->lapos = 2;
	} else if (fe->buf[0] == 0xef && fe->buf[1] == 0xbb
			&& fe->buf[2] == 0xbf) {
		fe->encoding = UTF8;
		fe->lapos = 3;

	} else if (CHK(fe->buf, 0x00, 0x00, 0x00, 0x3c)) {
		fe->encoding = UCS4BE;
		fe->lapos = 0;
	} else if (CHK(fe->buf, 0x3c, 0x00, 0x00, 0x00)) {
		fe->encoding = UCS4LE;
		fe->lapos = 0;
	} else if (CHK(fe->buf, 0x00, 0x00, 0x3c, 0x00)) {
		fe->encoding = UCS4U1;
		fe->lapos = 0;
	} else if (CHK(fe->buf, 0x00, 0x3c, 0x00, 0x00)) {
		fe->encoding = UCS4U2;
		fe->lapos = 0;
	} else if (CHK(fe->buf, 0x00, 0x3c, 0x00, 0x3f)) {
		/* XXX unsupported - should error */
		fe->encoding = UTF16BE;
		fe->lapos = 0;
	} else if (CHK(fe->buf, 0x3c, 0x00, 0x3f, 0x00)) {
		/* XXX unsupported - should error */
		fe->encoding = UTF16LE;
		fe->lapos = 0;
	} else if (CHK(fe->buf, 0x3c, 0x3f, 0x78, 0x6d)) {
		fe->encoding = UTF8;
		fe->lapos = 0;
	} else if (CHK(fe->buf, 0x4c, 0x6f, 0xa7, 0x94)) {
		/* XXX EBCDIC! - should error */
		fe->encoding = UTF8;
		fe->lapos = 0;
	} else {
		fe->encoding = UTF8;
		fe->lapos = 0;
	}

#undef CHK

}

static wchar_t
file_getch(e)
	struct entity *e;
{
	char buf[6] = { 0,0,0,0,0,0 };
	const char *p;
	int i, c;

	struct fileentity *fe = (struct fileentity *)e;

#define GB(fe) ((fe)->lapos < 4 ? (fe)->buf[(fe)->lapos++] :  fgetc((fe)->f))
#define UGB(fe, c) (fe)->buf[--(fe)->lapos] = (c)

	if (fe->encoding == UTF8) {
		c = GB(fe);
		if (c == -1)
			return -1;
		buf[0] = c;
		for (i = 1; i < 6; i++) {
			c = GB(fe);
			if (c == -1)
				return -1;		/* truncated char */
			if ((c & 0xc0) != 0x80) {	/* new char start */
				UGB(fe, c);		/* undo */
				break;
			}
			buf[i] = c;
		}
		p = buf;
		return utf8_getn(&p, i);
	}

	if (fe->encoding == UTF16LE || fe->encoding == UTF16BE) {
		/* XXX impossible */
	}

	/* Anything else must be UCS-4 */
	for (i = 0; i < 4; i++)
		buf[i] = GB(fe);

#define U4(a,b,c,d)	( ((wchar_t)buf[a] << 24) \
			| ((wchar_t)buf[b] << 16) \
			| ((wchar_t)buf[c] << 8)  \
			| ((wchar_t)buf[d]) )

	switch (fe->encoding) {
	case UCS4BE:	/* 1234 */
		return U4(1,2,3,4); break;
	case UCS4LE:	/* 4321 */
		return U4(4,3,2,1); break;
	case UCS4U1:	/* 2143 */
		return U4(2,1,4,3); break;
	case UCS4U2:	/* 3412 */
		return U4(3,4,1,2); break;
	default:
		/* XXX impossible */
		return -1;
	}
#undef U4
#undef UGB
#undef GB

}

static int
file_eof(e)
	struct entity *e;
{
	struct fileentity *fe = (struct fileentity *)e;

	return fe->lapos >=4 && feof(fe->f);
}

static void
file_free(e)
	struct entity *e;
{
	struct fileentity *fe = (struct fileentity *)e;

	free(fe->name);
	fclose(fe->f);
	free(fe);
}

static const char *
file_name(e)
	struct entity *e;
{
	struct fileentity *fe = (struct fileentity *)e;

	return fe->name;
}

struct entity *
entity_file(f, name)
	FILE *f;
	const char *name;
{
	struct fileentity *fe;

	fe = (struct fileentity *)malloc(sizeof (struct fileentity));
	if (!fe)
		return NULL;
	ENTITY_INIT(&fe->e, &file_ops);

	fe->f = f;
	fe->name = strdup(name);
	fe->encoding = 0;

	return (struct entity *)fe;
}
