/* $Id: test.c,v 1.6 2002/07/21 04:23:30 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* 
 * XPath test program. Read an XML file and apply an XPath expression to it.
 *
 * usage: xpath [-dvf] xpath-expr [xml-filename]
 *	-d 	increasing library debugging level
 *	-v	enable validation (process and check DTDs match)
 *	-f	show full node content (when printing out node-sets)
 *
 * Exit code is 255 on error.
 * If result is boolean, then exit code is 0 for true, or 1 for false.
 * Otherwise exit code is 0 on successful execution.
 *
 * If the xpath-expr is the empty string, the XML document is copied through
 * after DTD evaluation, whitespace stripping and implied attribute generation.
 *
 * If xml-filename is absent (or "-") then stdin is assumed.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <xml/xml.h>
#include <xml/xpath.h>
#include <xml/entity.h>

extern int xml_debug;	/* enables library debugging */

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct entity *e;
	xml_t root;
	xpath_t path;
	int type, i;
	int ch;
	int error = 0;
	int vflag = 0;	/* validate */
	int fflag = 0;	/* full node */
	char *path_expr;
	char *filename;
	FILE *f;
	union xpath_value xv;
	int exitcode = 0;

	/* Process command line options */
	while ((ch = getopt(argc, argv, "dvf")) != -1)
	    switch (ch) {
	    case 'd':
		xml_debug++;
		break;
	    case 'v':
		vflag = 1;
		break;
	    case 'f':
		fflag = 1;
		break;
	    default:
		error = 1;
		break;
	}

	/* Pick out the xpath expression from the command line */
	if (optind < argc) {
		path_expr = argv[optind];	
		if (*path_expr == '\0')
			path_expr = NULL;
		optind++;
	} else
		error = 1;

	/* Open the XML file for input */
	if (!error && optind < argc) {
		if (strcmp(argv[optind], "-") == 0) {
			f = stdin;
			filename = "stdin";
		} else {
			f = fopen(argv[optind], "r");
			filename = argv[optind];
			if (!f) {
				perror(argv[optind]); 
				exit(255);
			}
		}
		optind++;
	} else {
		f = stdin;
		filename = "stdin";
	}

	/* There should be no more arguments left */
	if (optind != argc)
		error = 1;

	/* Show usage information */
	if (error) {
		fprintf(stderr, "usage: %s [-dvf] xpath-expr [xml-file]\n",
			argv[0]);
		exit(1);
	}

	/* Compile the XPath expression into an xpath_t expression tree */
	if (path_expr) {
		path = xpath_new_utf8(path_expr);
		if (!path) {
			fprintf(stderr, "error in XPath expression: %s\n",
			    xpath_last_error);
			exit(1);
		}
		if (xml_debug) {
			printf("xpath expression is: ");
			xpath_puts(path, puts_stdio, stdout);
			printf("\n");
		}
	} else
		path = NULL;

	/* Convert the XML file descriptor into an enity object */
	e = entity_file(f, filename);
	if (!e) {
		fprintf(stderr, "error creating entity\n");
		exit(255);
	}

	/* Parse the XML file-entity into an XML document */
	root = xml_newe(e, NULL, !vflag);
	if (!root) {
		fprintf(stderr, "%s\n", xml_last_error);
		if (path)
			xpath_free(path);
		exit(255);
	}

	/*
	 * Evaluate the compiled XPath expression 
	 * in the context of the root node.
	 */
	if (path) {
		type = xpath_eval(path, root, NULL, &xv);

		/* Print the evaluation result */
		switch (type) {
		case XPATH_VALUE_ERROR:
			fprintf(stderr, "%s\n", xpath_last_error);
			exitcode = 255;
			break;
		case XPATH_VALUE_BOOLEAN:
			printf("%s\n", xv.x_boolean ? "true" : "false");
			exitcode = xv.x_boolean ? 0 : 1;
			break;
		case XPATH_VALUE_NUMBER:
			printf("%g\n", xv.x_number);
			break;
		case XPATH_VALUE_STRING_UTF8:
			printf("%.*s\n", xv.x_string.size, xv.x_string.data);
			free(xv.x_string.data);
			break;
		case XPATH_VALUE_NODESET:
			for (i = 0; xv.x_nodeset[i]; i++) {
				printf("%d: ", i+1);
				/* Print only the node's short form */
				if (fflag)
					xml_puts(xv.x_nodeset[i], puts_stdio,
						stdout);
				else
					xml_puts1(xv.x_nodeset[i], puts_stdio,
						stdout);
				printf("\n");
			}
			free(xv.x_nodeset);
			break;
		}
		xpath_free(path);

	} else {
		/*
		 * Print the in-memory XML document to stdout,
		 * including all inferred attributes, namespaces,
		 * and omitting elided whitespace.
		 */
		xml_puts(root, puts_stdio, stdout);
		printf("\n");
	}

	xml_free(root);

	exit(exitcode);
}
