/* $Id: html2xhtml.c,v 1.4 2002/07/21 04:29:02 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * Convert something looking like HTML into something looking like XHTML.
 * The HTML4 DTD is used to insert implicit elements, and to close
 * unclosed elements.
 *
 * Note: this doesn't actually check the HTML too closely. Also, it
 * might stuff up PRE elements.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "loose.inc"

static struct M root =
        { NULL, NULL, NULL, { PCData, HTML, NULL } };

/*------------------------------------------------------------
 * tag stack - a stack of currently-opened elements
 */

static char *stack[8192];
static struct M *stackm[8192];
static int stackpos = 0;
static int cdataok = 1;

static void
stack_push(sym, m)
	char *sym;
	struct M *m;
{
	stack[++stackpos] = sym;
	stackm[stackpos] = m;
	cdataok =  !m || m->content[0] == PCData;
}

static char *
stack_pop()
{
	static char *sym;
	sym = stack[stackpos--];
	/* note PCData must be first in element list by the DTD grammar! */
	cdataok = !stackm[stackpos] || stackm[stackpos]->content[0] == PCData;
	return sym;
}

static char *
stack_top()
{
	return stack[stackpos];
}

static struct M *
stack_topm()
{
	return stackm[stackpos];
}

/*------------------------------------------------------------
 * intern strings so we can use == instead of strcasecmp
 */

/* lowercase a string */
static void
lc(buf)
	char *buf;
{
	for (; *buf; buf++)
		if (*buf >= 'A' && *buf <= 'Z')
			*buf += 'a' - 'A';
}

struct intern {
	char *str;
	struct intern *next;
} *intern_head;

static char *
str_intern(s, dup)
	char *s;
	int dup;
{
	struct intern *i;

	if (s == NULL)
		return NULL;
	for (i = intern_head; i; i = i->next)
		if (s == i->str || strcasecmp(s, i->str) == 0)
			return i->str;
	i = (struct intern *)malloc(sizeof (struct intern));
	if (dup)
		s = strdup(s);
	i->str = s;
	lc(i->str);
	i->next = intern_head;
	intern_head = i;
	return i->str;
}

static char *
intern(s)
	char *s;
{
	return str_intern(s, 1);
}

/* Internalise all the constants in the M and A tables */
static void
intern_init()
{
	int i, j;

#define fix(v)	v = str_intern(v, 0)

	intern_head = NULL;
	for (i = 0; i < nMtab; i++) {
		fix(Mtab[i].tag);
		fix(Mtab[i].open);
		fix(Mtab[i].close);
		for (j = 0; Mtab[i].content[j]; j++)
			fix(Mtab[i].content[j]);
	}

	for (i = 0; i < nAtab; i++) {
		fix(Atab[i].tag);
		fix(Atab[i].attr);
	}

	fix(root.content[0]);
	fix(root.content[1]);
#undef fix
}

/*------------------------------------------------------------
 * single-character lookahead and line number counter
 */

int nextch = 0;
int lineno = 1;

static int
next() {
	int ret = nextch;
	if (ret != -1)
		nextch = getc(stdin);
	if (ret == '\n')
		lineno++;
	return ret;
}

/* read an identifier */
static char *
name() {
	static char buf[80];
	int len;

	len = 0;
	while ((nextch == '-' || nextch == '.' || nextch == '_' ||
		(nextch >= '0' && nextch <= '9') ||
	 	(nextch >= 'A' && nextch <= 'Z') ||
		(nextch >= 'a' && nextch <= 'z')) && len < sizeof buf - 1)
	    buf[len++] = next();
	buf[len] = '\0';

	lc(buf);
	return intern(buf);
}

/*------------------------------------------------------------
 * read a (possibly quoted) attribute value
 * also inserts extra entities if required
 */

static char *
value() {
	static char buf[8192];
	char buf2[8192];
	char *p;
	int len = 0;
	int quote = 0;

	if (nextch == '\'' || nextch == '\"')
		quote = next();
	len = 0;
	while (nextch != -1 && nextch != quote && 
	       len < sizeof buf2 - 1 &&
		      !(quote == 0 &&
			(nextch == '>' || nextch == ' ' || 
			 nextch == '\r' || nextch == '\t' ||
			 nextch == '\n')))
		buf2[len++] = next();
	if (quote)
		next();
	buf2[len] = '\0';

	len = 0;
	for (p = buf2; *p && len < sizeof buf - 1; p++) {
		if (*p == '&' && 
		    (p[1] == '#' ||
		     memcmp(p, "&lt;", 4) == 0 ||
		     memcmp(p, "&gt;", 4) == 0 ||
		     memcmp(p, "&amp;", 5) == 0 ||
		     memcmp(p, "&quot;", 6) == 0 ||
		     memcmp(p, "&apos;", 6) == 0))
			buf[len++] = *p;
		else if (*p == '&') {
			memcpy(buf+len, "&amp;", 5);
			len += 5;
		} else if (*p == '\'') {
			memcpy(buf+len, "&apos;", 6);
			len += 6;
		} else if (*p == '\"') {
			memcpy(buf+len, "&quot;", 6);
			len += 6;
		} else if (*p == '<') {
			memcpy(buf+len, "&lt;", 4);
			len += 4;
		} else if (*p == '>') {
			memcpy(buf+len, "&gt;", 4);
			len += 4;
		} else
			buf[len++] = *p;
	}
	buf[len]='\0';

	return buf;
}

/* return true if argument is whitespace */
static int
isspace(c)
	int c;
{
	return c == ' ' 
	    || c == '\n'
	    || c == '\r' 
	    || c == '\t';
}

/* parse optional whitespace */
static void 
space() {
	while (isspace(nextch))
		next();
}


/* Find a slot in the M table belonging to the given tag */
static struct M *
findM(tag)
	char *tag;
{
	int i;

	for (i = 0; i < nMtab; i++)
		if (Mtab[i].tag == tag)
			return &Mtab[i];
	return NULL;
}

/*------------------------------------------------------------
 * output reconstruction, indenting and optimising
 */

static int eopending = 0;
static int seenspace = 0;
static int hadcontent = 0;
static int lastwasspace = 0;

static void
checkeopen()
{
	if (eopending) {
		printf(">");
		lastwasspace = 0;
		if (seenspace) {
			printf(" ");
			lastwasspace = 1;
		}
		eopending = 0;
	}
}

static void
indent(offset)
	int offset;
{
	printf("\n%*s", offset + stackpos, "");
	lastwasspace = 1;
}

static void
opentag(tag, tagm)
	char *tag;
	struct M *tagm;
{
	checkeopen();
	if (!cdataok || lastwasspace)
		indent(0);
	printf("<%s", tag);
	stack_push(tag, tagm);
	eopending = -1;
}

static void
eopentag()
{
	eopending = 1;
	seenspace = 0;
	hadcontent = 0;
}

static void
closetag()
{
	char *tag;
	int cdatawasok = cdataok;

	tag = stack_pop();
	if (eopending && !seenspace)  {
		/* Special case of XML shortcut closure */
		printf("/>");
		eopending = 0;
		lastwasspace = 0;
	} else {
		checkeopen();
		if ((hadcontent == 2 && !cdatawasok) || lastwasspace)
			indent(0);
		printf("</%s>", tag);
		lastwasspace = 0;
	}
	seenspace = 0;
	hadcontent = 2;
}

static void
putdata(c)
{
	if (eopending && isspace(c)) {
		if (cdataok)
			seenspace = 1;
		return;
	}
	checkeopen();
	if (c == '\n')
		indent(0);
	else
		putchar(c);
	hadcontent = 1;
	lastwasspace = isspace(c);
}

/*------------------------------------------------------------
 * A breadth-first search for the shortest chain of tags to open
 * in order to reach the required tag
 */

static int
search_sub(req, spath, spathlen, opens)
	char *req;
	struct M **spath;
	int spathlen;
	struct M **opens;
{
	struct M *path[1024], *im, *jm;
	int pathsrc[1024];
	int i, j, pathlen = 0;

	for (i = 0; i < spathlen; i++) {
		im = spath[i];
		for (j = 0; im->content[j]; j++) {
			if (im->content[j] == req) {
				/* found */
				*opens = NULL;
				return i;
			}
			jm = findM(im->content[j]);
			if (!jm || jm->open != O)
				continue;
			path[pathlen] = jm;
			pathsrc[pathlen] = i;
			pathlen++;
		}
	}
	if (pathlen == 0)
		i = -1;
	else
		i = search_sub(req, path, pathlen, opens+1);
	if (i >= 0) {
		*opens = path[i];
		return pathsrc[i];
	}
	*opens = NULL;
	return -1;
}

static void
searchopen(curm, req, opens)
	struct M *curm;
	char *req;
	struct M **opens;
{
	search_sub(req, &curm, 1, opens);
}

/*------------------------------------------------------------
 * Enter or leave tags so that the next tag will be in a valid context
 * Prefer to enter tags rather than leave, since we don't know what's
 * coming up.
 */

static void
tryfor(tag)
	char *tag;
{
	int j;
	struct M *opens[80];
	struct M *curm;

	while (1) {
		curm = stack_topm();
		if (curm) {
			for (j = 0; curm->content[j]; j++)
				if (curm->content[j] == tag)
					break;
			if (curm->content[j]) 
				break;
			searchopen(curm, tag, opens);
			if (!opens[0] && curm->close == O) {
				closetag();
				continue;
			}
		} else
			opens[0] = NULL;
		for (j = 0; opens[j]; j++) {
			opentag(opens[j]->tag, opens[j]);
			eopentag();
		}
		break;
	}
}
/*------------------------------------------------------------
 * Tag events: called when parts of a tag from the source are
 * encountered (start-tag-begin, attribute, start-tag-finish, 
 * cdata, end-tag)
 */
static void
STagBegin(tag)
	char *tag;
{
	struct M *tagm;
	tryfor(tag);
	tagm = findM(tag);
	if (!tagm)
		fprintf(stderr, "line %d: unknown tag <%s>\n", lineno, tag);
	opentag(tag, tagm);
}

static void
Attr(tag, attr, value)
	char *tag, *attr, *value;
{
	int i;

	for (i = 0; i < nAtab; i++)
		if (Atab[i].tag == tag && Atab[i].attr == attr)
		{
			lc(value);
			break;
		}
	printf(" %s=\"%s\"", attr, value ? value : attr);
}


static void
STagEnd(tag)
	char *tag;
{
	eopentag();
}

static void
CData(c)
	int c;
{
	if (!isspace(c)) 
		tryfor(PCData);
	putdata(c);
}

static void
ETag(tag)
	char *tag;
{
	char *cur;
	struct M *curm;

	while (1) {
		cur = stack_top();
		if (cur == tag) {
			closetag();
			return;
		}
		curm = findM(cur);
		if (curm && curm->close == O) {
			closetag();
			continue;
		}
		fprintf(stderr, "line %d: ignoring bad close tag </%s>\n", 
			lineno, tag);
		return;
	}
}

/*------------------------------------------------------------
 * handle a specification that starts with <! (eg <!ENTITY...>)
 */
static void
spec()
{
	char buf[8192];
	int len;

	len = 0;
	while (nextch != '>' && nextch != -1) 
		buf[len++] = next();
	buf[len] = '\0';
	next();
	if (memcmp(buf, "!DOCTYPE", 8) == 0 ||
	    memcmp(buf, "!doctype", 8) == 0)
		return;
	printf("<%s>", buf);
}


int
main(argc, argv)
	int argc;
	char *argv[];
{
	int c;
	char *attr, *tag, *val;
	struct M *m;

	if (argc != 1) {
		fprintf(stderr, "usage: %s <foo.html >foo.xml\n", argv[0]);
		exit(1);
	}

	intern_init();
	stack_push(NULL, &root);

	next();
	printf("<?xml version=\"1.0\"?>\n");
	printf("<!DOCTYPE html\n"
		 "\tPUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
		 "\t\"DTD/xhtml1-transitional.dtd\">\n");

	while (nextch != -1) {
		c = next();

		if (c != '<') {
			/* CData */
			CData(c);
			continue;
		}

		if (nextch == '!') {
			spec();
			continue;
		}

		/* </tag> */
		if (nextch == '/') {
			next();
			tag = name();
			while (nextch != '>' && nextch != -1)
				next();
			next();	/* consume the > */
			ETag(tag);
			continue;
		}

		/* <tag ... */
		tag = name();
		STagBegin(tag);
		space();
		while (nextch != -1 && nextch != '>') {
			attr = name();
			space();
			if (nextch == '=') {
				next();	/* = */
				space();
				val = value();
				space();
			} else
				val = NULL;
			Attr(tag, attr, val);
		}
		next();	/* > */
		STagEnd(tag);
	}

	while ((m = stack_topm()) != &root)
		closetag();
	printf("\n");

	exit(0);
}
