/* $Id: dxmlmodule.c,v 1.4 2002/07/17 13:19:00 d Exp $ */
/* David Leonard, 2002. Public domain. */

#include <stdlib.h>
#include "Python.h"
#include <xml/entity.h>
#include <xml/xml.h>
#include <xml/xpath.h>
#include <xml/xmlns.h>

static PyObject *ErrorObject;

/* ------------------------------------------------------------ */

/* a puts function that appends to a python string */

static int puts_pystring(const char *str, void *ctxt) {
	PyObject **s = (PyObject **)ctxt;
	PyString_ConcatAndDel(s, 
	    PyString_Decode(str, strlen(str), "utf-8", "ignore"));
	return *s != NULL;
}

/* --------------------------------------------------------------------- */

/* an entity to read Python unicode strings */

struct pu_entity {
	struct entity e;
	PyUnicodeObject *o;
	int pos;
};

static void pu_rewind(struct entity *e) {
	struct pu_entity *pe = (struct pu_entity *)e;
	pe->pos = 0;
}

static wchar_t pu_getch(struct entity *e) {
	struct pu_entity *pe = (struct pu_entity *)e;
	return PyUnicode_AS_UNICODE(pe->o)[pe->pos++];
}

static int pu_eof(struct entity *e) {
	struct pu_entity *pe = (struct pu_entity *)e;
	return pe->pos >= PyUnicode_GET_SIZE(pe->o);
}

static void pu_free(struct entity *e) {
	struct pu_entity *pe = (struct pu_entity *)e;
	Py_DECREF((PyObject *)pe->o);
	PyMem_Del(pe);
}

static const char *pu_name(struct entity *e) {
	return NULL;
}

static struct entityop pu_op = 
	{ pu_rewind, pu_getch, pu_eof, pu_free, pu_name, NULL };

struct entity *
pu_new(PyUnicodeObject *o) {
	struct pu_entity *pe = PyMem_New(struct pu_entity, 1);
	if (pe) {
		ENTITY_INIT(&pe->e, &pu_op);
		Py_INCREF((PyObject *)o);
		pe->o = o;
		pe->pos = 0;
	}
	return (struct entity *)pe;
}

/*------------------------------------------------------------
 * Node object - wraps an xml_t
 *  if root is NULL, then this node is also a root object,
 *  which when decref'd releases everything
 */

typedef struct {
	PyObject_HEAD
	xml_t	node;
	PyObject *root;
} NodeObject;

staticforward PyTypeObject Node_Type;

#define NodeObject_Check(v)	((v)->ob_type == &Node_Type)

static PyObject *eval_path(xpath_t path, NodeObject *node, PyObject *bindings);

static NodeObject *
newNodeObject(xml_t node, PyObject *root)
{
	NodeObject *self;

	if (root && node == ((NodeObject *)root)->node) {
		/* We are the root */
		Py_INCREF(root);
		return (NodeObject *)root;
	}
	self = PyObject_New(NodeObject, &Node_Type);
	if (self == NULL)
		return NULL;
	self->node = node;
	if (root) {
		self->root = root;
		Py_INCREF(root);
	} else
		self->root = (PyObject *)self;
	return self;
}

static void
Node_dealloc(NodeObject *self)
{
	if (self->root == (PyObject *)self)
		xml_free(self->node);
	else 
		Py_DECREF(self->root);
	PyObject_Del(self);
}

/* A unicode converter */
static int to_unicode(PyObject *o, PyUnicodeObject **ret)
{
	PyObject *u;
	u = PyObject_Unicode(o);
	if (!u) return 0;
	if (!PyUnicode_Check(u)) {
		Py_DECREF(u);
		return 0;
	}
	*ret = (PyUnicodeObject *)u;
	return 1;
}

static PyObject *
Node_select(NodeObject *self, PyObject *args)
{
	PyUnicodeObject *buf;
	xpath_t path;
	PyObject *ret;
	PyObject *bindings = NULL;

	if (!PyArg_ParseTuple(args, "O&|O:select", to_unicode, &buf, &bindings))
		return NULL;

	path = xpath_newe(pu_new(buf));
	if (!path) {
		PyErr_SetString(ErrorObject, xpath_last_error);
		return NULL;
	}
	ret = eval_path(path, self, bindings);
	xpath_free(path);
	return ret;
}
static char Node_select_doc[] =
"select(str) -> value\n"
"Evaluate the argument as an XPath expression in the context of this node.";

static PyMethodDef Node_methods[] = {
	{"select",	(PyCFunction)Node_select, METH_VARARGS,
					Node_select_doc},
	/* "stringvalue" */
	{NULL,		NULL}		/* sentinel */
};

static PyObject *
Node_repr(NodeObject *self)
{
	PyObject *t, *s = PyString_FromString("");
	if (!xml_puts1(self->node, puts_pystring, &s)) {
		Py_XDECREF(s);
		s = NULL;
	}
	if (!s) return NULL;
	t = PyObject_Repr(s);
	Py_DECREF(s);
	if (!t) return NULL;
	s = PyString_FromString("<Node ");
	PyString_ConcatAndDel(&s, t);
	PyString_ConcatAndDel(&s, PyString_FromString(">"));
	return s;
}

static PyObject *
Node_str(NodeObject *self)
{
	PyObject *s = PyString_FromString("");
	if (!xml_puts(self->node, puts_pystring, &s)) {
		Py_XDECREF(s);
		s = NULL;
	}
	return s;
}

/*
 * Attach attributes to the node in a persistent way
 */
static PyObject *
Node_getattr(NodeObject *self, char *name)
{
	PyObject *dict = (PyObject *)xml_getuser(self->node);
	if (dict) {
		PyObject *v = PyDict_GetItemString(dict, name);
		if (v != NULL) {
			Py_INCREF(v);
			return v;
		}
	}
	return Py_FindMethod(Node_methods, (PyObject *)self, name);
}

/* Called by the xml library when node is about to be deallocated */
static void
dictfree(void *data, xml_t node)
{
	Py_XDECREF((PyObject *)data);
}

static int
Node_setattr(NodeObject *self, char *name, PyObject *v)
{
	PyObject *dict = (PyObject *)xml_getuser(self->node);
	if (dict == NULL) {
		dict = PyDict_New();
		if (dict == NULL)
			return -1;
		xml_setuser(self->node, dict, dictfree);
	}
	if (v == NULL) {
		int rv = PyDict_DelItemString(dict, name);
		if (rv < 0)
			PyErr_SetString(PyExc_AttributeError,
			        "delete non-existing Node attribute");
		return rv;
	}
	else
		return PyDict_SetItemString(dict, name, v);
}

statichere PyTypeObject Node_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"xml.Node",		/*tp_name*/
	sizeof(NodeObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)Node_dealloc,/*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)Node_getattr, /*tp_getattr*/
	(setattrfunc)Node_setattr, /*tp_setattr*/
	0,			/*tp_compare*/
        (reprfunc)Node_repr,    /*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        (reprfunc)Node_str,     /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        0,                      /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        0,                      /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        0,                      /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/
};

/*------------------------------------------------------------
 * Path
 */

typedef struct {
	PyObject_HEAD
	xpath_t	path;
} PathObject;

staticforward PyTypeObject Path_Type;

#define PathObject_Check(v)	((v)->ob_type == &Path_Type)

static PathObject *
newPathObject(xpath_t path)
{
	PathObject *self;

	self = PyObject_New(PathObject, &Path_Type);
	if (self == NULL)
		return NULL;
	self->path = path;
	return self;
}

static void
Path_dealloc(PathObject *self)
{
	xpath_free(self->path);
	PyObject_Del(self);
}


static PyObject *
Path_eval(PathObject *self, PyObject *args)
{
	NodeObject *node;

	if (!PyArg_ParseTuple(args, "O!:eval", &node, &Node_Type))
		return NULL;

	return eval_path(self->path, node, NULL);
}
static char Path_eval_doc[] =
"eval(node) -> resuult\n"
"Evaluate the XPath expression in the context of the given node";

static PyObject *
eval_path(path, node, bindings)
	xpath_t path;
	NodeObject *node;
	PyObject *bindings;
{
	PyObject *o;
	struct xmlns *ns;
	int vtype, len, i;
	union xpath_value pv;
	xml_t *xp;

	ns = NULL;	/* XXX should build from bindings */

	vtype = xpath_eval(path, node->node, ns, &pv);

	while (ns)
		ns = xmlns_free(ns);

	switch (vtype) {
	case XPATH_VALUE_NODESET:
		len = 0;
		for (xp = pv.x_nodeset; *xp; xp++)
			len++;
		o = PyList_New(len);
		for (i = 0; i < len; i++)
			PyList_SET_ITEM(o, i, (PyObject *)newNodeObject(
				pv.x_nodeset[i], node->root));
		free(pv.x_nodeset);
		return o;
	case XPATH_VALUE_BOOLEAN:
		o = pv.x_boolean ? Py_True : Py_False;
		Py_INCREF(o);
		return o;
	case XPATH_VALUE_NUMBER:
		return PyFloat_FromDouble(pv.x_number);
	case XPATH_VALUE_STRING_UTF8:
		o = PyString_Decode(
			pv.x_string.data,
			pv.x_string.size,
			"utf-8",
			"ignore");
		free(pv.x_string.data);
		return o;
	default:
	case XPATH_VALUE_ERROR:
		PyErr_SetString(ErrorObject, xpath_last_error
			? xpath_last_error : "unknown xpath error");
		return NULL;

	}
}

static PyObject *
Path_bind(PathObject *self, PyObject *args)
{
	wchar_t *name, *uri = NULL;
	int namelen, urilen = 0;
	size_t namesz, urisz;
	PyObject *o;

	if (!PyArg_ParseTuple(args, "Ou#|u#:bind", &o,
			&name, &namelen, &uri, &urilen))
		return NULL;

	namesz = namelen * sizeof (Py_UNICODE);
	urisz = urilen * sizeof (Py_UNICODE);

	/* special case of a single node */
	if (NodeObject_Check(o))
		xpath_bind_nodeset(self->path,
			uri, urisz, name, namesz,
			&((NodeObject *)o)->node, 1);
	else if (o == Py_False || o == Py_True)
		xpath_bind_boolean(self->path,
			uri, urisz, name, namesz,
			o == Py_True ? 1 : 0);
	else if (PyFloat_Check(o))
		xpath_bind_number(self->path,
			uri, urisz, name, namesz,
			PyFloat_AsDouble(o));
	else if (PyList_Check(o)) {
		int len = PyList_Size(o);
		int i;
		xml_t *nodes;

		nodes = len ? malloc(len * sizeof (xml_t)) : NULL;
		if (len && !nodes)
			return PyErr_NoMemory();
		for (i = 0; i < len; i++) {
			PyObject *n = PyList_GetItem(o, i);
			if (n && !NodeObject_Check(n)) {
				Py_DECREF(n);
				PyErr_SetString(PyExc_TypeError, 
				   "expected list of nodes");
				n = NULL;
			}
			if (!n) {
				free(nodes);
				return NULL;
			}
			nodes[i] = ((NodeObject *)n)->node;
			Py_DECREF(n);
		}
		xpath_bind_nodeset(self->path,
			uri, urisz, name, namesz,
			nodes, len);
		if (len)
			free(nodes);
	} else {
		PyObject *u = PyObject_Unicode(o);
		if (!u)
			return NULL;
		xpath_bind_string(self->path,
			uri, urisz, name, namesz,
			(wchar_t *)PyUnicode_AS_DATA(u),
			(size_t)PyUnicode_GET_DATA_SIZE(u));
		Py_DECREF(u);
	}

	Py_INCREF(Py_None);
	return NULL;
}

static PyObject *
Path_str(PathObject *self)
{
	PyObject *s = PyString_FromString("");
	if (!xpath_puts(self->path, puts_pystring, &s)) {
		Py_XDECREF(s);
		s = NULL;
	}
	return s;
}

static PyMethodDef Path_methods[] = {
	{"eval",	(PyCFunction)Path_eval, METH_VARARGS,
				Path_eval_doc},
	{"bind",	(PyCFunction)Path_bind, METH_VARARGS},
	{NULL,		NULL}		/* sentinel */
};

static PyObject *
Path_getattr(PathObject *self, char *name)
{
	return Py_FindMethod(Path_methods, (PyObject *)self, name);
}

statichere PyTypeObject Path_Type = {
	/* The ob_type field must be initialized in the module init function
	 * to be portable to Windows without using C++. */
	PyObject_HEAD_INIT(NULL)
	0,			/*ob_size*/
	"xml.Path",		/*tp_name*/
	sizeof(PathObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)Path_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)Path_getattr,/*tp_getattr*/
	(setattrfunc)0,		/*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
        0,                      /*tp_call*/
        (reprfunc)Path_str,     /*tp_str*/
        0,                      /*tp_getattro*/
        0,                      /*tp_setattro*/
        0,                      /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT,     /*tp_flags*/
        0,                      /*tp_doc*/
        0,                      /*tp_traverse*/
        0,                      /*tp_clear*/
        0,                      /*tp_richcompare*/
        0,                      /*tp_weaklistoffset*/
        0,                      /*tp_iter*/
        0,                      /*tp_iternext*/
        Path_methods,           /*tp_methods*/
        0,                      /*tp_members*/
        0,                      /*tp_getset*/
        0,                      /*tp_base*/
        0,                      /*tp_dict*/
        0,                      /*tp_descr_get*/
        0,                      /*tp_descr_set*/
        0,                      /*tp_dictoffset*/
        0,                      /*tp_init*/
        0,                      /*tp_alloc*/
        0,                      /*tp_new*/
        0,                      /*tp_free*/
        0,                      /*tp_is_gc*/
};

/* ------------------------------------------------------------ */

/* Function of two integers returning integer */

static PyObject *
dxml_parse_xml(PyObject *self, PyObject *args)
{
	PyUnicodeObject *buf;
	xml_t root;

	if (!PyArg_ParseTuple(args, "O&:parse_string", to_unicode, &buf))
		return NULL;

	root = xml_newe(pu_new(buf), NULL, 0);
	if (!root) {
		PyErr_SetString(ErrorObject, xml_last_error);
		return NULL;
	}
	return (PyObject *)newNodeObject(root, NULL);
}
static char dxml_parse_xml_doc[] =
"parse_string(str) -> node\n"
"Parse the XML document contained in the string, and return the root node";

/* Function of no arguments returning new Xxo object */

static PyObject *
dxml_new_xpath(PyObject *self, PyObject *args)
{
	PyUnicodeObject *buf;
	xpath_t path;

	if (!PyArg_ParseTuple(args, "O&:new_xpath", to_unicode, &buf))
		return NULL;

	path = xpath_newe(pu_new(buf));
	PySys_WriteStderr("path=%p\n", path);
	if (!path) {
		PyErr_SetString(ErrorObject, xpath_last_error);
		return NULL;
	}
	return (PyObject *)newPathObject(path);
}
static char dxml_new_xpath_doc[] =
"new_xpath(str) -> xpath\n"
"Create an XPath object from the XPath expression";

static PyObject *
dxml_set_debug(PyObject *self, PyObject *args)
{
	int val;
	extern int xml_debug;

	if (!PyArg_ParseTuple(args, "i:set_debug", &val))
		return NULL;
	xml_debug = val;
	Py_INCREF(Py_None);
	return Py_None;
}

static char dxml_set_debug_doc[] =
"set_debug(int)\n"
"Change the XML library debugging level. This can result in\n"
"debugging messages being sent to stderr.";

/* List of functions defined in the module */

static PyMethodDef dxml_methods[] = {
	{"parse_string",dxml_parse_xml,		METH_VARARGS,
				dxml_parse_xml_doc},
	{"xpath",	dxml_new_xpath,		METH_VARARGS,
				dxml_new_xpath_doc},
	{"set_debug",	dxml_set_debug,		METH_VARARGS,
				dxml_set_debug_doc},
	{NULL,		NULL}		/* sentinel */
};


/* Initialization function for the module (*must* be called initxx) */

DL_EXPORT(void)
initxml(void)
{
	PyObject *m, *d;

	/* Initialize the type of the new type object here; doing it here
	 * is required for portability to Windows without requiring C++. */
	Node_Type.ob_type = &PyType_Type;
	Path_Type.ob_type = &PyType_Type;

	/* Create the module and add the functions */
	m = Py_InitModule("xml", dxml_methods);

	/* Add some symbolic constants to the module */
	d = PyModule_GetDict(m);
	ErrorObject = PyErr_NewException("xml.error", NULL, NULL);
	PyDict_SetItemString(d, "error", ErrorObject);
	PyDict_SetItemString(d, "__version__", PyString_FromFormat(
	    "%d.%d", XML_LIBRARY_VERSION / 0x100, 
		     XML_LIBRARY_VERSION % 0x100));
	PyDict_SetItemString(d, "__date__", 
		PyInt_FromLong(XML_LIBRARY_DATE));
}
