/* $Id: utf8.c,v 1.2 2002/07/14 11:38:17 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * UTF8 string encoding and decoding
 *
 * References:
 *	F Yeargeau, UTF-8: a Transformation Format of ISO 10646. (RFC2279)
 */

#include <stdlib.h>
#include "utf8.h"

static wchar_t getn(const char *p, const char **pret, int maxlen);

static struct {
	char	mask, match;
} decode[] = {
	{ 0x80, 0x00 },	/* 0xxxxxxx */
	{ 0xe0, 0xc0 },	/* 110xxxxx 10xxxxxx */
	{ 0xf0, 0xe0 },	/* 1110xxxx 10xxxxxx*2 */
	{ 0xf8, 0xf0 },	/* 11110xxx 10xxxxxx*3 */
	{ 0xfc, 0xf8 },	/* 111110xx 10xxxxxx*4 */
	{ 0xfe, 0xfc },	/* 1111110x 10xxxxxx*5 */
};

/* Read a UTF-8 encoded wchar from the string */
static wchar_t
getn(p, pret, maxlen)
	const char *p;
	const char **pret;
	int maxlen;
{
	wchar_t	c = 0;
	int i;

	if (maxlen <= 0)
		return -1;

	for (i = 0; i < 6; i++)
		if ((*p & decode[i].mask) == decode[i].match) {
			if (i >= maxlen) {
				*pret += maxlen;
				return -1;
			}
			c = *p & ~decode[i].mask;
			p++;
			for (; i; i--) {
				if ((*p & 0xc0) != 0x80) {
					*pret = p + 1;
					return -1;
				}
				c = (c << 6) | (*p & 0x3f);
				p++;
			}
			if (pret)
				*pret = p;
			return c;
		}
	/* Illegal UTF-8 chars (0xFE and 0xFF) decoded as -1 */
	if (pret)
		*pret = p + 1;
	return -1;
}

wchar_t
utf8_look(p)
	const char *p;
{
	return getn(p, NULL, 6);
}

wchar_t
utf8_get(pp)
	const char **pp;
{
	return getn(*pp, pp, 6);
}

wchar_t
utf8_getn(pp, maxlen)
	const char **pp;
	int maxlen;
{
	return getn(*pp, pp, maxlen);
}

static wchar_t encode[] = {
	0xffffff80,	/* 0xxxxxxx */
	0xfffff800,	/* 110xxxxx 10xxxxxx */
	0xffff0000,	/* 1110xxxx 10xxxxxx*2 */
	0xffe00000,	/* 11110xxx 10xxxxxx*3 */
	0xfc000000,	/* 111110xx 10xxxxxx*4 */
	0x80000000	/* 1111110x 10xxxxxx*5 */
};

void
utf8_put(c, p)
	wchar_t c;
	char **p;
{
	int i, j;

	for (i = 0; i < 6; i++)
		if ((encode[i] & c) == 0) {
			for (j = i; j > 0; j--) {
				(*p)[j] = 0x80 | (c & 0x3f);
				c >>= 6;
			}
			**p = decode[i].match | (c & ~decode[i].mask);
			(*p) += (i + 1);
			return;
		}
	/* Invalid UCS-4 characters become the illegal UTF-8 character 0xfe */
	*(*p)++ = 0xfe;
}

/* Return the number of characters that the char would consume */
int
utf8_size_put(c)
	wchar_t c;
{
	int i;

	for (i = 0; i < 6; i++)
		if ((encode[i] & c) == 0) 
			return i+1;
	return 1;
}


/*------------------------------------------------------------*/

int
wisletter(c)
	wchar_t c;
{
	return wisbasechar(c) || wisideographic(c);
}

#define R(a,b)		(a <= c && c <= b)
#define E(a)		(c == a)

int
wisbasechar(c)
	wchar_t c;
{
	return R(0x0041, 0x005A) || R(0x0061, 0x007A)
	       || R(0x00C0, 0x00D6) || R(0x00D8, 0x00F6)
	       || R(0x00F8, 0x00FF) /* ... XXX */;	
}

int
wisideographic(c)
	wchar_t c;
{
	return R(0x4E00, 0x9FA5) || E(0x3007)
		|| R(0x3021, 0x3029);
}

int
wiscombiningchar(c)
	wchar_t c;
{
	return 0; /* XXX */
}

int
wisdigit(c)
	wchar_t c;
{
	return R(0x0030, 0x0039);	/* XXX */
}

int
wisextender(c)
	wchar_t c;
{
	return E(0x00B7) || E(0x02D0) || E(0x02D1) || E(0x0387)
		|| E(0x0640) || E(0x0E46) || E(0xEC6) || E(0x3005)
		|| R(0x3031, 0x3035) || R(0x309D, 0x309E)
		|| R(0x30FC, 0x30FE);
}
