/* $Id: xfsm.c,v 1.2 2002/07/18 05:29:35 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * DFSM (Deterministic Finite State Machine) generator for regular languages.
 * Used to implement DTD validation.
 * Note: for compatibility, DTDs should be regular (deterministic) but
 * this module does not require that - non-deterministic state machines
 * are 'reduced' to a deterministic form.
 * 
 * Reference:
 *   T Pittman and J Peters, The Art of Compiler Design. Pages 60--71,
 *	Prentice-Hall, 1992.
 */

/* XXX TODO: simplify using the macros from vector.h */

#include <stdlib.h>
#include "xfsm.h"

struct xfsm {
	int accept, start;
	int flag;

	int nout, aout;
	struct outtrans {
		const void *symbol;
		int inidx;
		struct xfsm *target;
	} *out;

	int nin, ain;
	struct intrans {
		int outidx;
		struct xfsm *source;
	} *in;

	struct xfsm *next, **prevp;
};

#define EPSILON		NULL

static struct xfsm *newstate(struct xfsm **h);
static int	addtrans(struct xfsm *source, const void *symbol, 
			struct xfsm *target);
static void	deltrans(struct xfsm *x, int idx);
static void	freestate(struct xfsm *x, int clean);
static void	merge(struct xfsm *a, struct xfsm *b);
static void	copy(struct xfsm *a, struct xfsm *b);
static int	loop(struct xfsm **h, struct xfsm *xs, struct xfsm *xf, 
			struct xfsm **a, struct xfsm **b, int opt, int multi);
static void	zeroflag(struct xfsm **h);
void	reduce1(struct xfsm **h);
static void	reduce2sub(struct xfsm *a);
void	reduce2(struct xfsm **h);
static struct sunion *su_new(void);
static void	su_free(struct sunion *u);
static void	su_add(struct sunion *u, struct xfsm *x);
static int	su_eq(struct sunion *a, struct sunion *b);
void	reduce3(struct xfsm **h);
void	reduce4(struct xfsm **h);


/* create a new, empty state */
static struct xfsm *
newstate(h)
	struct xfsm **h;
{
	struct xfsm *x;

	x = (struct xfsm *)malloc(sizeof (struct xfsm));
	if (x != NULL) {
		x->accept = 0;
		x->start = 0;
		x->flag = 0;

		x->nout = 0;
		x->aout = 0;
		x->out = NULL;

		x->nin = 0;
		x->ain = 0;
		x->in = NULL;

		/* Add to linked list */
		x->next = *h;
		if (x->next)
			x->next->prevp = &x->next;
		x->prevp = h;
		*h = x;
	}
	return x;
}

/* add a transition from a state */
static int
addtrans(source, symbol, target)
	struct xfsm *source, *target;
	const void *symbol;
{
	size_t sz;
	int i;

	/* Don't add if it is an epsilon transition to self */
	if (source == target && symbol == EPSILON)
		return 1;

	/* Don't add if it already exists */
	for (i = 0; i < source->nout; i++)
		if (source->out[i].target == target &&
		    source->out[i].symbol == symbol)
			return 1;

	/* Ensure there is enough room in the transition lists */
	while (source->nout >= source->aout) {
	    source->aout = source->aout ? source->aout * 2 : 4;
	    sz = source->aout * sizeof (struct outtrans);
	    if (source->out)
		source->out = (struct outtrans *)realloc(source->out, sz);
	    else
		source->out = (struct outtrans *)malloc(sz);
	    if (source->out == NULL)
		return 0;
	}

	while (target->nin >= target->ain) {
	    target->ain = target->ain ? target->ain * 2 : 4;
	    sz = target->ain * sizeof (struct intrans);
	    if (target->in)
		target->in = (struct intrans *)realloc(target->in, sz);
	    else
		target->in = (struct intrans *)malloc(sz);
	    if (target->in == NULL)
		return 0;
	}

	/* Record incoming and outgoing transitions */
	source->out[source->nout].symbol = symbol;
	source->out[source->nout].inidx = target->nin;
	source->out[source->nout].target = target;
	target->in[target->nin].outidx = source->nout;
	target->in[target->nin].source = source;

	source->nout++;
	target->nin++;
	return 1;
}

/* delete an out transition */
static void
deltrans(x, idx)
	struct xfsm *x;
	int idx;
{
	struct xfsm *t = x->out[idx].target;
	int ti = x->out[idx].inidx;
	int n;

	/* First delete the in record */
	n = t->nin - 1;
	if (n != ti) {
		t->in[ti] = t->in[n];
		t->in[ti].source->out[t->in[ti].outidx].inidx = ti;
	}
	t->nin--;

	/* Next, delete the out record */
	n = x->nout - 1;
	if (n != idx) {
		x->out[idx] = x->out[n];
		x->out[idx].target->in[x->out[idx].inidx].outidx = idx;
	}
	x->nout--;
}

/* destroy a state */
static void
freestate(x, clean)
	struct xfsm *x;
	int clean;
{
	*(x->prevp) = x->next;
	if (x->next)
		x->next->prevp = x->prevp;
	if (clean) {
		while (x->nout)
			deltrans(x, x->nout - 1);
		while (x->nin)
			deltrans(x->in[x->nin - 1].source,
				x->in[x->nin - 1].outidx);
	}
	if (x->in)
		free(x->in);
	if (x->out)
		free(x->out);
	free(x);
}

/* merge b with a, making b redundant (it should be freed after) */
static void
merge(a, b)
	struct xfsm *a, *b;
{
	int i;

	copy(a, b);

	/* copy all x->b transitions as x->a */
	for (i = 0; i < b->nin; i++)
		addtrans(b->in[i].source, 
			b->in[i].source->out[b->in[i].outidx].symbol,
			a);

	/* NOTE: a->a epsilon transitions are always ignored */
}

/* copy the outgoing edges of b onto a */
static void
copy(a, b)
	struct xfsm *a, *b;
{
	int i;

	/* copy all b->x transitions as a->x */
	for (i = 0; i < b->nout; i++)
		addtrans(a, b->out[i].symbol, b->out[i].target);

	if (b->accept)
		a->accept = 1;
	if (b->start)
		a->start = 1;
}

/* create an empty fsm */
int
xfsm_empty(h, a, b)
	struct xfsm **h, **a, **b;
{
	struct xfsm *x;

	if (!h) return 1;
	x = newstate(h);
	if (x == NULL)
		return 0;
	x->start = 1;
	x->accept = 1;
	*a = *b = x;
	return 1;
}

/* create an fsm for a simple symbol transition */
int
xfsm_simple(h, symbol, a, b)
	struct xfsm **h;
	const void *symbol;
	struct xfsm **a, **b;
{
	struct xfsm *A, *B;

	if (!h) return 1;
	A = newstate(h);
	if (A == NULL)
		return 0;
	B = newstate(h);
	if (B == NULL) {
		freestate(A, 1);
		return 0;
	}
	if (!addtrans(A, symbol, B)) {
		freestate(A, 1);
		freestate(B, 1);
		return 0;
	}
	A->start = 1;
	B->accept = 1;
	*a = A;
	*b = B;
	return 1;
}

/* (x,y) */
int
xfsm_seq(h, xs, xf, ys, yf, a, b)
	struct xfsm **h, *xs, *xf, *ys, *yf, **a, **b;
{
	if (!h) return 1;
	if (!addtrans(xf, EPSILON, ys))
		return 0;
	xf->accept = 0;
	ys->start = 0;
	*a = xs;
	*b = yf;
	return 1;
}

/* (x|y) */
int
xfsm_alt(h, xs, xf, ys, yf, a, b)
	struct xfsm **h, *xs, *xf, *ys, *yf, **a, **b;
{
	struct xfsm *A, *B;

	if (!h) return 1;
	A = newstate(h);
	if (A == NULL)
		return 0;
	B = newstate(h);
	if (B == NULL) {
		freestate(A, 1);
		return 0;
	}

	if (!addtrans(A, EPSILON, xs) ||
	    !addtrans(A, EPSILON, ys) ||
	    !addtrans(xf, EPSILON, B) ||
	    !addtrans(yf, EPSILON, B))
	{
		freestate(A, 1);
		freestate(B, 1);
		return 0;
	}
	A->start = 1;
	xs->start = 0;
	ys->start = 0;
	xf->accept = 0;
	yf->accept = 0;
	B->accept = 1;
	*a = A;
	*b = B;
	return 1;
}

/* common code for xfsm_kleene(), xfsm_plus() and xfsm_opt() */
static int
loop(h, xs, xf, a, b, opt, multi)
	struct xfsm **h, *xs, *xf, **a, **b;
	int opt, multi;
{
	struct xfsm *A, *B;

	if (!h) return 1;
	A = newstate(h);
	if (A == NULL)
		return 0;
	B = newstate(h);
	if (B == NULL) {
		freestate(A, 1);
		return 0;
	}

	if (!addtrans(A, EPSILON, xs) ||
	    !addtrans(xf, EPSILON, B) ||
	    (opt && !addtrans(A, EPSILON, B)) ||
	    (multi && !addtrans(B, EPSILON, A)))
	{
		freestate(A, 1);
		freestate(B, 1);
		return 0;
	}
	A->start = 1;
	xs->start = 0;
	xf->accept = 0;
	B->accept = 1;
	*a = A;
	*b = B;
	return 1;
}

/* (x)* */
int
xfsm_kleene(h, xs, xf, a, b)
	struct xfsm **h, *xs, *xf, **a, **b;
{
	return loop(h, xs, xf, a, b, 1, 1);
}

/* (x)+ */
int
xfsm_plus(h, xs, xf, a, b)
	struct xfsm **h, *xs, *xf, **a, **b;
{
	return loop(h, xs, xf, a, b, 0, 1);
}

/* (x)? */
int
xfsm_opt(h, xs, xf, a, b)
	struct xfsm **h, *xs, *xf, **a, **b;
{
	return loop(h, xs, xf, a, b, 1, 0);
}

/* free all FSMs in the list */
void
xfsm_free(h)
	struct xfsm **h;
{
	while (*h) 
		freestate(*h, 0);
}

/* initialise the FSM head */
void
xfsm_init(h)
	struct xfsm **h;
{
	*h = NULL;
}

/* set the flag member of each state to zero */
static void
zeroflag(h)
	struct xfsm **h;
{
	struct xfsm *x;

	for (x = *h; x; x = x->next)
		x->flag = 0;
}

/* find cycles of epsilon transitions and merge them into single states */
void
reduce1(h)
	struct xfsm **h;
{
	int nalloc = 4;
	int len;
	struct s {
		struct xfsm *x;
		int outidx;
	} *q;
	struct xfsm *t;

	q = (struct s *)malloc(nalloc * sizeof *q);
	zeroflag(h);
	t = *h;
	while (t) {
	    if (t->flag != 0) {
		t = t->next;
		continue;
	    }
	    t->flag = 1;
	    len = 1;
	    q[len - 1].x = t;
	    q[len - 1].outidx = 0;
	    while (len) {
		struct xfsm *x = q[len - 1].x;
		struct xfsm *y;
		int i;

		i = q[len - 1].outidx++;

		if (i >= x->nout) {
			x->flag = 2;
			len--;
			continue;
		}

		if (x->out[i].symbol != EPSILON) 
			continue;

		y = x->out[i].target;
		switch (y->flag) {
		case 0:
			if (nalloc <= len) {
				while (nalloc <= len)
					nalloc *= 2;
				q = (struct s *)realloc(q, nalloc * sizeof *q);
			}
			q[len].x = y;
			q[len].outidx = 0;
			len++;
			x->flag = 1;
			break;
		case 1:
			/* cycle found - merge them all*/
			while (len && q[len - 1].x != y) {
				x = q[len - 1].x;
				merge(y, x);
				freestate(x, 1);
				len--;
			}
			/* reset all the partially traversed states */
			while (len) {
				x = q[len - 1].x;
				x->flag = 0;
				len --;
			}
			/* restart from the beginning, as indicies change */
			t = *h;
			len = 0;
			break;
		case 2:
			/* seen */
			break;
		}
	    }
	}
}

static void
reduce2sub(a)
	struct xfsm *a;
{
	int i;
	struct xfsm *b;

	for (i = 0; i < a->nout; i++) 
		if (a->out[i].symbol == EPSILON) {
			b = a->out[i].target;
			/* recursion okay because reduce1() removed cycles */
			reduce2sub(b);
			deltrans(a, i);
			copy(a, b);
			i--;
		}
}

/* remove epislon transitions by copying states */
void
reduce2(h)
	struct xfsm **h;
{
	struct xfsm *a;

	for (a = *h; a; a = a->next) 
		reduce2sub(a);
}

/* A structure to hold a set of states, also called a "state union" */
struct sunion {
	int	len, alloc;
	struct xfsm **vector;
	struct xfsm *state;		/* The representative state if any */
};

/* Create an empty state-union structure */
static struct sunion *
su_new()
{
	struct sunion *u;

	u = (struct sunion *)malloc(sizeof (struct sunion));
	u->len = 0;
	u->alloc = 4;
	u->vector = (struct xfsm **)malloc(u->alloc * sizeof (struct xfsm *));
	u->state = NULL;
	return u;
}

/* Deallocate storage for a state-union structure */
static void
su_free(u)
	struct sunion *u;
{
	free(u->vector);
	free(u);
}

/* Add state x into the vector field of u */
static void
su_add(u, x)
	struct sunion *u;
	struct xfsm *x;
{
	int i, j;

	for (i = 0; i < u->len; i++)
		if (u->vector[i] == x)
			return;
		else if (u->vector[i] > x)
			break;
	/* Grow the vector if needed */
	while (u->len + 1 >= u->alloc) {
		u->alloc *= 2;
		u->vector = (struct xfsm **)realloc(u->vector,
			u->alloc * sizeof (struct xfsm *));
	}
	/* Make space for the new fsm pointer */
	for (j = u->len; j > i; j--)
		u->vector[j] = u->vector[j - 1];
	u->vector[i] = x;
	u->len++;
}

/* Compare two unions for equality */
static int
su_eq(a, b)
	struct sunion *a, *b;
{
	return (a->len == b->len && memcmp(a->vector, b->vector,
		a->len * sizeof (struct xfsm *)) == 0);
}

/* remove nondeterminism by forming 'union' states */
void
reduce3(h)
	struct xfsm **h;
{
	struct xfsm *start;
	struct xfsm *head = NULL;

	int qalloc = 4;
	int qlen, qpos;
	int salloc = 4;
	int slen;
	int i, j, k;
	struct sunion **q = (struct sunion **)malloc(qalloc * sizeof *q);
	struct sunion *tmp = su_new();
	const void **syms = (const void **)malloc(salloc * sizeof *syms);

	/* find the start state */
	start = xfsm_getstart(h);
	if (!start) {
		xfsm_free(h);
		return;
	}

	/* Initialise the queue, with the start-class state */
	q[0] = su_new();
	q[0]->state = newstate(&head);
	q[0]->state->start = 1;
	q[0]->state->accept = start->accept;
	su_add(q[0], start);
	qpos = 0;
	qlen = 1;

	while (qpos < qlen) {
		struct sunion *s = q[qpos++];

		/* First, build a list of unique symbols out of the union */
		slen = 0;
		for (i = 0; i < s->len; i++) {
			struct xfsm *x = s->vector[i];
			for (j = 0; j < x->nout; j++) {
			    const void *sym = x->out[j].symbol;
			    for (k = 0; k < slen; k++)
				if (syms[k] == sym)
				    break;
			    if (k >= slen) {
				if (slen + 1 > salloc) {
				    salloc *= 2;
				    syms = (const void **)realloc(syms,
					salloc * sizeof *syms);
				}
				syms[slen++] = sym;
			    }
			}
		}

		/* For each symbol, construct a union target state name */
		for (k = 0; k < slen; k++) {
			const void *sym = syms[k];
			struct sunion *su;

			tmp->len = 0;
			for (i = 0; i < s->len; i++) {
			    struct xfsm *x = s->vector[i];
			    for (j = 0; j < x->nout; j++) 
				if (x->out[j].symbol == sym)
				    su_add(tmp, x->out[j].target);
			}

			/* Does the union already exist? */
			su = NULL;
			for (i = 0; i < qlen; i++)
			    if (su_eq(tmp, q[i])) {
				su = q[i];
				break;
			    }
			if (su == NULL) {
			    /* Add the new union name */
			    su = tmp;
			    tmp = su_new();
			    while (qlen + 1 >= qalloc) {
				qalloc *= 2;
				q = (struct sunion **)realloc(q,
				    qalloc * sizeof *q);
			    }
			    q[qlen++] = su;
			    su->state = newstate(&head);
			    for (i = 0; i < su->len; i++)
				if (su->vector[i]->accept) {
				    su->state->accept = 1;
				    break;
				}
			}
			/* Add the state class transition */
			addtrans(s->state, sym, su->state);
		}
	}

	/* We now have a completely new fsm list */
	xfsm_free(h);
	*h = head;
	if (*h)
		(*h)->prevp = h;

	for (i = 0; i < qlen; i++)
		su_free(q[i]);
	free(q);
	free(syms);
}

/* reduce states to a minimum number of equivalent classes */
void
reduce4(h)
	struct xfsm **h;
{
	/* XXX - didn't bother because we only need determinism */
}

/* make the fsm states deterministic through a series of reduction steps */
void
xfsm_reduce(h)
	struct xfsm **h;
{
	struct xfsm *start;

	if (!*h)
		return;

	reduce1(h);
	reduce2(h);
	reduce3(h);
	reduce4(h);

	/*
	 * Move the start state to the beginning of the list 
	 * to speed up future xfsm_getstart() calls
	 */
	start = xfsm_getstart(h);
	if (start && start != *h) {
		/* remove start from the list */
		*(start->prevp) = start->next;
		if (start->next)
			start->next->prevp = start->prevp;
		/* insert start at the head */
		start->next = *h;
		if (start->next)
			start->next->prevp = &start->next;
		*h = start;
		start->prevp = h;
	}
}

/* return the start state of the finite state machine */
struct xfsm *
xfsm_getstart(h)
	struct xfsm **h;
{
	struct xfsm *start;

	for (start = *h; start; start = start->next)
		if (start->start)
			break;
	return start;
}

/* return the next state transited to by a symbol, or return NULL if invalid */
struct xfsm *
xfsm_getnext(x, symbol)
	struct xfsm *x;
	const void *symbol;
{
	int i;

	for (i = 0; i < x->nout; i++)
		if (x->out[i].symbol == symbol)
			return x->out[i].target;
	return NULL;
}

/* return true if the given state is an accepting state */
int
xfsm_isaccept(x)
	struct xfsm *x;
{
	return x->accept;
}

/* return the ith possible symbol valid in this state, or NULL */
int
xfsm_getpossible(x, i, symp)
	struct xfsm *x;
	int i;
	const void **symp;
{
	if (i < 0 || i >= x->nout)
		return 0;
	*symp = x->out[i].symbol;
	return 1;
}

/* add a transition to the fsm */
void
xfsm_addtrans(a, sym, b)
	struct xfsm *a;
	const void *sym;
	struct xfsm *b;
{
	addtrans(a, sym, b);
}


/* XXX - undocumented - used in debugging */
struct xfsm *
xfsm_listnext(x)
	struct xfsm *x;
{
	return x->next;
}
