/* $Id: xhash.c,v 1.1.1.1 2002/07/07 14:03:33 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* Simple hash table */

#include <stdlib.h>
#include "xhash.h"

#define HASHLEN	257			/* probably not perfect */

struct hashent {
	void *key;
	int keylen;
	void *data;
	struct hashent *next;
};

struct xhash {
	void	*defaultdata;
	void	(*freefn)(void *);
	struct hashent *ent[HASHLEN];
};

/* Allocate storage and initialise an empty hash table */
struct xhash *
xhash_new(defaultdata, freefn)
	void *defaultdata;
	void (*freefn)(void *);
{
	struct xhash *x;
	int i;

	x = (struct xhash *)malloc(sizeof (struct xhash));
	if (x != NULL) {
		x->defaultdata = defaultdata;
		for (i = 0; i < HASHLEN; i++)
			x->ent[i] = NULL;
		x->freefn = freefn;
	}
	return x;
}

/* Deallocate storage associated with a hash table */
void
xhash_free(x)
	struct xhash *x;
{
	int i;
	struct hashent *e, *nexte;

	for (i = 0; i < HASHLEN; i++) {
		e = x->ent[i];
		while (e != NULL) {
			nexte = e->next;
			if (x->freefn && e->data != x->defaultdata)
				(*x->freefn)(e->data);
			free(e);
			e = nexte;
		}
	}
	free(x);
}

/* Helper function to find a linked list entry in the hash table */
static struct hashent **
find(x, key, keylen)
	struct xhash *x;
	const void *key;
	int keylen;
{
	struct hashent **ep;
	unsigned int h;
	int i;
	unsigned char *p;

	/* Compute the hash of the key */
	h = 0;
	p = (unsigned char *)key;
	for (i = 0; i < keylen; i++)
		h = (h << 1) ^ *p++;
	h = h % HASHLEN;

	for (ep = &x->ent[h]; *ep; ep = &((*ep)->next)) 
		if ((*ep)->keylen == keylen 
		    && memcmp((*ep)->key, key, keylen) == 0)
			break;
	return ep;
}

/* Find a value associated with a key */
void *
xhash_lookup(x, key, keylen)
	struct xhash *x;
	const void *key;
	int keylen;
{
	struct hashent **ep;

	ep = find(x, key, keylen);
	if (*ep == NULL)
		return x->defaultdata;
	else
		return (*ep)->data;
}

/* Associate a value with a key */
int
xhash_store(x, key, keylen, data)
	struct xhash *x;
	const void *key;
	int keylen;
	void *data;
{
	struct hashent **ep, *e;

	ep = find(x, key, keylen);
	if (*ep == NULL) {
		e = (struct hashent *)malloc(sizeof (struct hashent));
		if (e == NULL)
			return 0;
		e->key = malloc(keylen);
		if (e->key == NULL) {
			free(e);
			if (x->freefn && data != x->defaultdata)
				(*x->freefn)(data);
			return 0;
		}
		memcpy(e->key, key, keylen);
		e->keylen = keylen;
		e->next = NULL;
		*ep = e;
	} else {
		if (x->freefn && (*ep)->data != x->defaultdata)
			(*x->freefn)((*ep)->data);
	}
	(*ep)->data = data;
	return 1;
}

/*------------------------------------------------------------
 * Iterator
 */

struct xhash_iter {
	struct xhash *h;
	struct hashent *ent;
	int slot;
};

/* Move the iterator pointer to the next slot in the hash table */
static void
iter_next(i)
	struct xhash_iter *i;
{
	if (i->ent)
		i->ent = i->ent->next;
	if (i->ent == NULL)
		while (i->slot < HASHLEN) {
			i->slot++;
			if (i->h->ent[i->slot]) {
				i->ent = i->h->ent[i->slot];
				return;
			}
		}
}

struct xhash_iter *
xhash_iter_new(h)
	struct xhash *h;
{
	struct xhash_iter *i;

	i = (struct xhash_iter *)malloc(sizeof (struct xhash_iter));
	if (i) {
		i->h = h;
		i->slot = -1;
		i->ent = NULL;
		iter_next(i);
	}
	return i;
}

void
xhash_iter_free(i)
	struct xhash_iter *i;
{
	free(i);
}

int
xhash_iter_next(i, keyp, keylenp, datap)
	struct xhash_iter *i;
	void *keyp;
	int *keylenp;
	void **datap;
{
	if (i->slot >= HASHLEN) 
		return 0;
	memcpy(keyp, i->ent->key, *keylenp);
	*keylenp = i->ent->keylen;
	*datap = i->ent->data;
	iter_next(i);
	return 1;
}
