/* $Id: xml.c,v 1.16 2002/07/18 05:36:26 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * XML parser
 *
 * Reference: Extensible Markup Language (XML) 1.0 (Second edition),
 *	<http://www.w3.org/XML/>
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>

#include "utf8.h"
#include "entity.h"
#include "vector.h"
#include "wstr.h"
#include "xml.h"
#include "xpath.h"
#include "xfsm.h"
#include "xhash.h"
#include "intern.h"
#include "xmlP.h"
#include "xmlns.h"
#include "xmlnsP.h"

const char *xml_last_error;

int xml_debug = 0;

/* Stack of entities that are being read for input */
struct inputstack {
	struct entity 		*e;
	struct inputstack 	*prev;
	VECTOR_DECL(wchar_t)	 la;	/* lookahead vector */
	int			 lapos;	/* position in lookahead vector */
	int			 lineno;
};

/* Parser input and state structure maintained during entire tree parse */
struct input {
	struct inputstack	*stack;	/* stack of input sources */
	wchar_t			 next;	/* next character returned by next() */
	char			*error;	/* last error message or NULL */
	struct xmldocinfo	*doc;	/* document info */
	struct entity *(*fetch)(const char *pub, const char *sys);
	int			novalidate;
};

#define DPRINTF(fmt, args...) do { if (xml_debug)  	\
	fprintf(stderr, fmt , ## args);			\
} while (0)

#define DnPRINTF(n, fmt, args...) do {			\
	if (xml_debug >= (n))				\
		fprintf(stderr, fmt , ## args);		\
} while (0)

#define DASSERT(ex) do { if (xml_debug && !(ex)) 	\
	fprintf(stderr, "%s:%c: FAILED: %s\n", 		\
		__FILE__, __LINE__, #ex); 		\
} while (0)

#define DPRINTFN(in) \
	DPRINTF("%s() next = '%c'\n", __FUNCTION__, in->next)

/* Macro to see if there is more data to read */
#define more(in)	(!(in)->error && (in)->next != -1)

/* Prototypes */
static struct dtdelement *dtdelement_new(void);
static void dtdelement_free(struct dtdelement *de);
static struct dtdattr *dtdattr_new(enum AttType type);
static void dtdattr_free(struct dtdattr *da);
static struct externalid *externalid_new(wstr_t sysid, wstr_t pubid);
static void externalid_free(struct externalid *n);
static struct xmldocinfo *doc_new(void);
static void doc_free(struct xmldocinfo *doc);
static void doc_defconstentity(struct xmldocinfo *doc, const char *name, const char *rtext);
static struct input *input_new(struct xmldocinfo *doc, struct entity *(*fetch)(const char *pub, const char *sys), int novalidate);
static void input_free(struct input *in);
static void input_push(struct input *in, struct entity *e);
static void input_pop_raw(struct input *in);
static void input_pop(struct input *in);
static char *wstr_form(wchar_t c);
static void newline(struct input *in);
static void input_lookahead(struct input *in, int len);
static void *error(struct input *in, const char *fmt, ...);
static wchar_t next(struct input *in);
static int nextis(struct input *in, const char *s);
static int nextis_S(struct input *in, const char *s, const char *t);
static void expect(struct input *in, const char *s);
static wstr_t ns_lookup(struct input *in, struct xmlns *ns, wstr_t name);
static void ns_enter(struct xmldocinfo *doc, xml_t el, wstr_t name, wstr_t uri);
static int isNCName(struct input *in);
static wstr_t NCName(struct input *in);
static int isNCNameChar(struct input *in);
static xml_t document(struct input *in);
static void S(struct input *in);
static int isS(struct input *in);
static void Sopt(struct input *in);
static int isNameChar(struct input *in);
static wstr_t Name(struct input *in);
static int isName(struct input *in);
static wstr_t Nmtoken(struct input *in);
static wstr_t EntityValue(struct input *in);
static int isLiteral(struct input *in);
static wstr_t AttValue(struct input *in, struct dtdattr *da);
static void enumeration_check(struct input *in, struct dtdattr *da, wstr_t v);
static wstr_t SystemLiteral(struct input *in);
static wstr_t PubidLiteral(struct input *in);
static wstr_t cdata(struct input *in, int iscdata);
static void CharData(struct input *in, xml_t el);
static void Comment(struct input *in, xml_t el);
static int isComment(struct input *in);
static xml_t PI(struct input *in, xml_t parent);
static int isPI(struct input *in);
static void CDSect(struct input *in, xml_t el);
static void CDStart(struct input *in);
static int isCDStart(struct input *in);
static wstr_t CData(struct input *in);
static void CDEnd(struct input *in);
static int isCDEnd(struct input *in);
static void prolog(struct input *in, xml_t root);
static void XMLDecl(struct input *in, xml_t root);
static wstr_t VersionInfo(struct input *in);
static void Eq(struct input *in);
static wstr_t VersionNum(struct input *in);
static int isMisc(struct input *in);
static void Misc(struct input *in, xml_t root);
static void doctypedecl(struct input *in, xml_t root);
static void markupdecl(struct input *in, xml_t root);
static int SDDecl(struct input *in);
static int isSDDecl(struct input *in);
static xml_t element(struct input *in, xml_t parent);
static int STag(struct input *in, xml_t el, struct dtdelement **dep);
static xml_t newattrnode(struct input *in, xml_t el, xml_t **elp, qname_t nm, wstr_t v);
static xml_t default_attr(struct input *in, xml_t el, xml_t **elp, struct dtdattr *da, wstr_t name);
static xml_t Attribute(struct input *in, xml_t el, xml_t **elp, struct dtdelement *de);
static void ETag(struct input *in, xml_t el);
static void content(struct input *in, xml_t el, struct dtdelement *de);
static void debug_fsm_dump(struct dtdelement *de);
static void badcontent(struct input *in, struct dtdelement *de, struct xfsm *state, wstr_t actualsym);
static void pe_expand(struct input *in);
static void elementdecl(struct input *in);
static void contentspec(struct input *in, struct dtdelement *de);
static void children(struct input *in, struct xfsm **h, struct xfsm **a, struct xfsm **b);
static void cp(struct input *in, struct xfsm **h, struct xfsm **a, struct xfsm **b);
static void lchoice(struct input *in, struct xfsm **h, struct xfsm **a, struct xfsm **b);
static void lseq(struct input *in, struct xfsm **h, struct xfsm **a, struct xfsm **b);
static void lloop(struct input *in, struct xfsm **h, struct xfsm **a, struct xfsm **b);
static int islloop(struct input *in);
static void Mixed(struct input *in, struct xfsm **h, struct xfsm **a, struct xfsm **b);
static int isMixed(struct input *in);
static void AttlistDecl(struct input *in);
static void AttDef(struct input *in, struct dtdelement *de);
static struct dtdattr *AttType(struct input *in);
static struct dtdattr *NotationType(struct input *in);
static struct dtdattr *Enumeration(struct input *in);
static void DefaultDecl(struct input *in, struct dtdattr *da);
static wstr_t CharRef(struct input *in);
static int isCharRef(struct input *in);
static wstr_t Reference(struct input *in);
static int isReference(struct input *in);
static wstr_t EntityRef(struct input *in);
static struct entity *PEReference(struct input *in);
static void EntityDecl(struct input *in);
static struct entity *EntityDef(struct input *in);
static struct entity *PEDef(struct input *in);
static struct externalid *ExternalID(struct input *in);
static int isExternalID(struct input *in);
static wstr_t NDataDecl(struct input *in);
static int isNDataDecl(struct input *in);
static wstr_t EncodingDecl(struct input *in);
static int isEncodingDecl(struct input *in);
static wstr_t EncName(struct input *in);
static void NotationDecl(struct input *in);
static xml_t newtextnode(xml_t parent, wstr_t w);
static xml_t newnode(enum xmltype type, xml_t parent);
static void freenode(xml_t x);
static int xml_puts_sib(xml_t node, int (*puts)(const char *, void *), void *closure, int dosib, int dochld);
/*------------------------------------------------------------
 * DTD object management
 */
static struct dtdelement *
dtdelement_new()
{
	struct dtdelement *de;

	de = (struct dtdelement *)malloc(sizeof (struct dtdelement));
	if (de) {
		xfsm_init(&de->fsm);
		de->attrs = xhash_new(NULL, (void (*)(void *))dtdattr_free);
		de->ignorespace = 0;
		de->idattr = NULL;
	}
	return de;
}

static void
dtdelement_free(de)
	struct dtdelement *de;
{
	xhash_free(de->attrs);
	xfsm_free(&de->fsm);
	free(de);
}

static struct dtdattr *
dtdattr_new(type)
	enum AttType type;
{
	struct dtdattr *da;

	da = (struct dtdattr *)malloc(sizeof (struct dtdattr));
	if (da) {
		da->type = type;
		VECTOR_INIT(da->enums);
		da->deftype = IMPLIED;
		da->defattr = NULL;
	}
	return da;
}

static void
dtdattr_free(da)
	struct dtdattr *da;
{
	VECTOR_FREE(da->enums);
	if (da->defattr)
		VECTOR_FREE(*da->defattr);
	free(da);
}

/*------------------------------------------------------------
 * ExternalID object management
 */

static struct externalid *
externalid_new(sysid, pubid)
	wstr_t sysid, pubid;
{
	struct externalid *n;

	n = (struct externalid *)malloc(sizeof (struct externalid));
	if (n) {
		n->systemid = sysid;
		n->publicid = pubid;
	}
	return n;
}

static void
externalid_free(n)
	struct externalid *n;
{
	wstr_free(n->systemid);
	wstr_free(n->publicid);
	free(n);
}

/*------------------------------------------------------------
 * Common document info management
 */

static struct xmldocinfo *
doc_new()
{
	struct xmldocinfo *doc;

	doc = (struct xmldocinfo *)malloc(sizeof (struct xmldocinfo));
	if (doc) {
		/* Some hash tables of strings */
		doc->entity = xhash_new(NULL, (void (*)(void *))entity_decref);
		doc->pentity = xhash_new(NULL, (void (*)(void *))entity_decref);

		/* Predefined entities from section 4.6 */
		doc_defconstentity(doc, "lt", "&#60;");		/* 60 */
		doc_defconstentity(doc, "gt", "&#62;");		/* 62 */
		doc_defconstentity(doc, "amp", "&#38;");	/* 38 */
		doc_defconstentity(doc, "apos", "&#39;");	/* 39 */
		doc_defconstentity(doc, "quot", "&#34;");	/* 34 */

		doc->dtd = xhash_new(NULL, 
			(void (*)(void *))dtdelement_free);
		doc->notation = xhash_new(NULL, 
			(void (*)(void *))externalid_free);
		doc->id = xhash_new(NULL, NULL);
		doc->namespaces = NULL;

		doc->doctype = NULL;
		doc->pcdata = intern_utf8("#PCDATA");
		doc->xmlns = intern_utf8("xmlns");
		doc->empty = intern_utf8("");

		doc->nextorder = 0;
	}
	return doc;
}

static void
doc_free(doc)
	struct xmldocinfo *doc;
{
	xhash_free(doc->entity);
	xhash_free(doc->pentity);
	xhash_free(doc->dtd);
	xhash_free(doc->notation);
	xhash_free(doc->id);
	xml_free(doc->namespaces);
	free(doc);
}

/* Define a constant entity. Assumes rtext is a static UTF8 string */
static void
doc_defconstentity(doc, name, rtext)
	struct xmldocinfo *doc;
	const char *name, *rtext;
{
	struct entity *e;
	wstr_t nm;

	nm = intern_utf8(name);
	if (nm) {
		e = entity_utf8mem(rtext, rtext + strlen(rtext), 0);
		if (e)
			xhash_store(doc->entity, &nm, sizeof nm, e);
	}
}

int
doc_orderseq(doc)
	struct xmldocinfo *doc;
{
	return doc->nextorder++;
}

/*------------------------------------------------------------
 * parser input state structure
 */

static struct input *
input_new(doc, fetch, novalidate)
	struct xmldocinfo *doc;
	struct entity *(*fetch)(const char *pub, const char *sys);
	int novalidate;
{
	struct input *in;

	in = (struct input *)malloc(sizeof (struct input));
	if (in) {
		in->error = NULL;
		in->doc = doc;
		in->fetch = fetch;
		in->novalidate = novalidate;

		/* Put an EOF entity at the bottom of the input stack */
		in->stack = NULL;
		input_push(in, entity_eof());
	}
	return in;
}

/* free input state structure */
static void
input_free(in)
	struct input *in;
{
	while (in->stack)
		input_pop(in);

	if (in->error)
		free(in->error);
	free(in);
}

/* push an entity onto the input stack. (entity ref ownership is taken) */
static void
input_push(in, e)
	struct input *in;
	struct entity *e;
{
	struct inputstack *s;

	DnPRINTF(2, "input_push()\n");
	if (e->inuse) {
		error(in, "recursive entity");
		ENTITY_DECREF(e);
		return;
	}
	e->inuse++;
	ENTITY_REWIND(e);
	s = (struct inputstack *)malloc(sizeof (struct inputstack));
	s->lineno = 1;
	s->e = e;
	s->prev = in->stack;
	in->stack = s;
	VECTOR_INIT(s->la);
	s->lapos = 0;
	input_lookahead(in, 1);
	in->next = VECTOR_AT(in->stack->la, 0);
}

/* pop the top entity, but don't preserve any lookahead */
static void
input_pop_raw(in)
	struct input *in;
{
	in->stack->lapos = 0;
	VECTOR_LEN(in->stack->la) = 0;
	input_pop(in);
}

/* pop the top entity on the input stack */
static void
input_pop(in)
	struct input *in;
{
	struct inputstack *s;

	DnPRINTF(2, "input_pop()\n");

	s = in->stack;
	assert(s != NULL);
	in->stack = s->prev;

/*
	if (in->stack && (s->e->flags & ENTITY_ADDLINES))
		in->stack->lineno += (s->lineno - 1);
*/

	s->e->inuse--;
	ENTITY_DECREF(s->e);

	/*
	 * Carefully insert the lookahead buffer of the popped entity
	 * in front of the la buffer of the newly exposed entity.
	 */
	if (in->stack) {
		struct inputstack *t = in->stack;
		int i;

		/* make room */
		for (i = 0; i < VECTOR_LEN(s->la); i++)
			VECTOR_APPEND(t->la, 0);
		/* move everything from lapos upwards */
		for (i = VECTOR_LEN(t->la) - 1; 
		     i >= t->lapos + VECTOR_LEN(s->la); i--)
			VECTOR_AT(t->la, i) =
			    VECTOR_AT(t->la, i - VECTOR_LEN(s->la));
		/* copy in the lookahead of the outgoing l.a.buf */
		for (i = 0; i < VECTOR_LEN(s->la); i++)
			VECTOR_AT(t->la, i + t->lapos) = 
			    VECTOR_AT(s->la, i);
		in->next = VECTOR_AT(t->la, t->lapos);
	} else
		in->next = -1;

	VECTOR_FREE(s->la);
	free(s);
}

/* Return a simple string representation of a wide character */
static char *
wstr_form(c)
	wchar_t c;
{
	static char buf[7];

	if (c >= ' ' && c < 0x80) {
		buf[0] = c;
		buf[1] = 0;
		return buf;
	}

	buf[0] = '\\';
	buf[2] = '\0';
	switch (c) {
		case '\n':	buf[1] = 'n'; return buf;
		case '\r':	buf[1] = 'r'; return buf;
		case '\t':	buf[1] = 't'; return buf;
		case '\'':	buf[1] = c;   return buf;
		case '\"':	buf[1] = c  ; return buf;
		case -1:
			snprintf(buf, sizeof buf, "<EOF>");
			return buf;
		default:
			snprintf(buf + 1, sizeof buf - 1, "x%04x", c);
			return buf;
	}
}

static void
newline(in)
	struct input *in;
{
	in->stack->lineno++;
}

/* ensure the lookahead buffer is at least len characters long */
static void
input_lookahead(in, len)
	struct input *in;
	int len;
{
	wchar_t c;

	while (in->stack &&
	    VECTOR_LEN(in->stack->la) < len + in->stack->lapos)
	{
		while (in->stack && ENTITY_EOF(in->stack->e))
			input_pop(in);

		/*
		 * section 2.11 End of line handling
		 *  must translate both the two-character sequence
		 *  #xD #xA and any #xD that is not followed by #xA
		 *  to a single #A character
		 */
		c = ENTITY_GETCH(in->stack->e);
		if (c == 0x0d && (in->stack->e->flags & ENTITY_ISEXTERNAL)) {
			VECTOR_APPEND(in->stack->la, 0x0a);
			newline(in);
			if (!ENTITY_EOF(in->stack->e)) {
				c = ENTITY_GETCH(in->stack->e);
				if (c != 0x0a)
					VECTOR_APPEND(in->stack->la, c);
			}
		} else {
			if (c == 0x0a)
				newline(in);
			VECTOR_APPEND(in->stack->la, c);
		}
	}

	if (xml_debug >= 2) {
		int i;
		struct inputstack *is;
		for (i = 0, is = in->stack; is; i++, is = is->prev);
		fprintf(stderr, "la[%d]: lookahead is '", i);
		for (i = 0; i < VECTOR_LEN(in->stack->la); i++) {
			if (i == in->stack->lapos)
				fprintf(stderr, "\' . \'");
			fprintf(stderr, "%s", 
			    wstr_form(VECTOR_AT(in->stack->la, i)));
			if (VECTOR_AT(in->stack->la, i) == -1)
				break;
		}
		fprintf(stderr, "'\n");
	}
}

/*------------------------------------------------------------
 * basic parser building blocks
 */

/* put the input into an error state. ignores future errors */
static void *
error(struct input *in, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (in->error == NULL) {
		static char buf[4096];
		int i;
		int len;
		const char *name;
		int lineno;
		struct inputstack *s;

		/* Try to find a name:lineno in the entity stack */
		lineno = 0;
		name = NULL;
		for (s = in->stack; s; s = s->prev) {
			name = ENTITY_NAME(s->e);
			lineno = s->lineno;
			if (name)
				break;
		}
		if (name)
			snprintf(buf, sizeof buf, "%s:%d: ", name, lineno);
		else
			buf[0] = '\0';
		len = strlen(buf);

		/* Add the actual error message */
		vsnprintf(buf + len , sizeof buf - len, fmt, ap);

		strlcat(buf, " at \'", sizeof buf);
		strlcat(buf, wstr_form(in->next), sizeof buf);

		if (VECTOR_AT(in->stack->la, in->stack->lapos) != in->next) {
			DPRINTF("*** BAD LOOKAHEAD in->next='%s'",
				wstr_form(in->next));
			DPRINTF(" in->la[%d]='%s'\n",
				in->stack->lapos,
				wstr_form(VECTOR_AT(in->stack->la, 
					in->stack->lapos)));
		}

		input_lookahead(in, 10);
		for (i = in->stack->lapos + 1; 
		     i < 10 && i < VECTOR_LEN(in->stack->la);
		     i++)
		{
			if (VECTOR_AT(in->stack->la, i) == -1)
				break;
			strlcat(buf, 
			    wstr_form(VECTOR_AT(in->stack->la, i)),
			    sizeof buf);
		}
		strlcat(buf, "\'", sizeof buf);

		in->error = strdup(buf);
		in->next = -1;
		DPRINTF("**ERROR: %s\n", in->error);
	}
	va_end(ap);
	return NULL;
}

/* return the next character from the input/lookahead */
static wchar_t
next(in)
	struct input *in;
{
	wchar_t result;

	if (in->error || !in->stack)
		return -1;

	DASSERT(in->stack->lapos < VECTOR_LEN(in->stack->la));
	DASSERT(in->next == VECTOR_AT(in->stack->la, in->stack->lapos));

	result = in->next;
	in->stack->lapos++;
	if (in->stack->lapos == VECTOR_LEN(in->stack->la)) {
		VECTOR_LEN(in->stack->la) = 0;
		in->stack->lapos = 0;
		input_lookahead(in, 1);
	}
	in->next = VECTOR_AT(in->stack->la, in->stack->lapos);

	DnPRINTF(3,"next = %s ", wstr_form(result));
	DnPRINTF(3,". %s\n", wstr_form(in->next));

	return result;
}

/* Return true if the next characters match the given ASCII string */
static int
nextis(in, s)
	struct input *in;
	const char *s;
{
	int len;
	int i;
	int ret;


	if (!more(in))
		return 0;
	if (in->next != s[0])		/* quick optimisation */
		ret = 0;
	else {
		len = strlen(s);
		input_lookahead(in, len);
		ret = 1;
		for (i = 1; ret && i < len; i++)
			if (VECTOR_AT(in->stack->la, in->stack->lapos + i) 
					!= s[i])
				ret = 0;
	}
	DnPRINTF(2,"nextis(\"%s\") -> %d\n", s, ret);
	return ret;
}

/* Return true if the next characters match sSt */
static int
nextis_S(in, s, t)
	struct input *in;
	const char *s, *t;
{
	int offset, len, slen, tlen;
	wchar_t c;
	int state = 0;
	int ok;

	if (!more(in))
		return 0;

	slen = strlen(s);
	tlen = strlen(t);
	state = 0;
	len = 0;
	offset = 0;
	ok = 1;
	while (ok && state < 3) {
		input_lookahead(in, len + 1);
		c = VECTOR_AT(in->stack->la, in->stack->lapos + len);
		switch (state) {
		case 0:
			if (offset < slen) {
				ok = (c == s[offset++]);
				break;
			}
			offset = 0;
			state = 1;
		case 1:
			if (isSpace(c)) {
				offset++;
				break;
			}
			if (offset == 0) {	/* must have some space */
				ok = 0;
				break;
			}
			offset = 0;
			state = 2;
		case 2:
			if (offset < tlen) {
				ok = (c == t[offset++]);
				break;
			}
			state = 3;
		case 3:
			break;
		}
		DPRINTF("-- '%c' len %d: state=%d offset=%d ok=%d\n",
				c, len, state, offset, ok);
		len++;
	}
			
	DPRINTF("nextis_S(\"%s\", \"%s\") -> %d\n", s, t, ok);
	return ok;
}

/* Expect an ASCII string to appear on the input stream */
static void
expect(in, s)
	struct input *in;
	const char *s;
{
	if (in->error)
		return;
	if (!nextis(in, s)) {
		error(in, "expected '%s'", s);
		return;
	}
	/* skip the characters of the found string */
	while (*s++)
		next(in);
}

/*------------------------------------------------------------
 * namespaces operations
 */

static wstr_t
ns_lookup(in, ns, name)
	struct input *in;
	struct xmlns *ns;
	wstr_t name;
{
	wstr_t uri;

	uri = xmlns_lookup(ns, name);
	if (uri == xmlns_unknown) {
		error(in, "unknown namespace '%s'", wstr_str(name));
		uri = NULL;
	}
	return uri;
}

/* Create a new namespace node and apply to node el */
static void
ns_enter(doc, el, name, uri)
	struct xmldocinfo *doc;
	xml_t el;
	wstr_t name, uri;
{
	xml_t ns = newnode(NS, NULL);
	if (uri == doc->empty)	/* convert empty string ("") to NULL */
		uri = NULL;
	if (ns) {
		ns->parent = ns->ns = el->ns;
		ns->aux.ns = xmlns_bindw(el->ns ? el->ns->aux.ns : NULL, name, uri);
		ns->name.ns = NULL;
		ns->name.local = name;
		el->ns = ns;

		/* Add the ns node into the ns tree */
		ns->sibling = doc->namespaces;
		doc->namespaces = ns;
		ns->order = doc_orderseq(doc);
	}
}

/*------------------------------------------------------------
 * Namespace grammar
 */

qname_t
QName(in)
	struct input *in;
{
	wstr_t w;
	qname_t q;

	w = NCName(in);
	if (in->next == ':') {
		next(in);
		q.ns = w;
		q.local = NCName(in);
	} else {
		q.ns = NULL;
		q.local = w;
	}
	return q;
}

/* return true if input starts with a NCName */
static int
isNCName(in)
	struct input *in;
{
	return wisletter(in->next) || in->next == '_' ;
}

/* [NS4] NCName */
static wstr_t
NCName(in)
	struct input *in;
{
	wstr_t name;

	DPRINTFN(in);
	if (!isNCName(in)) {
		error(in, "expected NCName");
		return NULL;
	}

	/* Read NCNameChars into the 'name' vector */
	if (!isNCNameChar(in)) {
		error(in, "expected Nmtoken");
		return NULL;
	}

	name = wstr_new();
	while (more(in) && isNCNameChar(in)) {
		wchar_t c = next(in);
		WSTR_ADD(name, c);
	}

	/* Internalise the name */
	name = intern_wstr(name);
	return name;
}

/* [NS5] return true if next char is a NCNameChar */
static int
isNCNameChar(in)
	struct input *in;
{
	return wisletter(in->next) ||
	       wisdigit(in->next) ||
	       in->next == '.' ||
	       in->next == '-' ||
	       in->next == '_' ||
	       wiscombiningchar(in->next) ||
	       wisextender(in->next);
}


/*------------------------------------------------------------
 * The XML grammar
 */

/* [1] document */
static xml_t
document(in)
	struct input *in;
{
	xml_t root;
	xml_t top;

	DPRINTFN(in);

	root = newnode(ROOT, NULL);
	root->aux.doc = in->doc;
	root->order = doc_orderseq(in->doc);

	/* Create the default, NULL namespace */
	ns_enter(in->doc, root, NULL, NULL);

	Sopt(in);
	prolog(in, root);

	top = element(in, root);
	if (top && in->doc->doctype && 
	    !(in->doc->doctype == top->name.local && top->name.ns == NULL))
		error(in, "expected <%s> as top element", 
		    wstr_str(in->doc->doctype));

	while (more(in))
		Misc(in, root);

	return root;
}

/* [3] S */
static void
S(in)
	struct input *in;
{
	DPRINTFN(in);
	if (!isS(in)) 
		error(in, "expected space");
	else
		Sopt(in);
}

/* return true if input starts with an S */
static int
isS(in)
	struct input *in;
{
	return isSpace(in->next);
}

/* S? */
static void
Sopt(in)
	struct input *in;
{
	DPRINTFN(in);
	while (isSpace(in->next))
	    next(in);
}

/* [4] return true if next char is a NameChar */
static int
isNameChar(in)
	struct input *in;
{
	return wisletter(in->next) ||
	       wisdigit(in->next) ||
	       in->next == '.' ||
	       in->next == '-' ||
	       in->next == '_' ||
	       in->next == ':' ||
	       wiscombiningchar(in->next) ||
	       wisextender(in->next);
}

/* [5] read a Name, returning an intern'd pointer to its wchar representation */
static wstr_t
Name(in)
	struct input *in;
{
	DPRINTFN(in);
	if (!isName(in)) {	/* Name is a subset of Nmtoken */
		error(in, "expected Name");
		return NULL;
	}
	return Nmtoken(in);
}

/* return true if input starts with a Name */
static int
isName(in)
	struct input *in;
{
	return wisletter(in->next) || in->next == '_' 
		|| in->next == ':';
}

/* [7] read a Nmtoken */
static  wstr_t
Nmtoken(in)
	struct input *in;
{
	wstr_t name;

	DPRINTFN(in);
	/* Read NameChars into the 'name' vector */
	if (!isNameChar(in)) {
		error(in, "expected Nmtoken");
		return NULL;
	}

	name = wstr_new();
	while (more(in) && isNameChar(in)) {
		wchar_t c = next(in);
		WSTR_ADD(name, c);
	}

	/* Internalise the name */
	name = intern_wstr(name);

	if (xml_debug) {
		fprintf(stderr, "  Nmtoken() -> %p = ", name);
		puts_wstr(name, puts_stdio, stderr, 1);
		fprintf(stderr, "\n");
	}

	return name;
}

/* [9] EntityValue */
static wstr_t
EntityValue(in)
	struct input *in;
{
	wstr_t value;
	wchar_t quote;
	struct entity *eof;

	/*
	 * Because of section 4.4.5, entities referenced
	 * in an entity value that contain a quote, do not
	 * terminate the entity value. e.g.
	 *   <!ENTITY % YN '"Yes"' >
	 *   <!ENTITY WhatHeSaid "He said %YN;" >
	 * should have &WhatHeSaid; expand to:
	 *   He said "Yes"
	 *
	 * To achieve this, we scan the whole string,
	 * push a temporary EOF entity layer, and then the string back.
	 *
	 * See also AttValue, which has to do something similar.
	 */

	DPRINTFN(in);
	quote = next(in);	/* ['"] */
	value = wstr_new();
	while (more(in) && in->next != quote)
		WSTR_ADD(value, next(in));
	next(in);

	eof = entity_eof();
	input_push(in, eof);
	input_push(in, entity_wstr(value));

	value = wstr_new();
	while (in->next != -1)
		if (in->next == '%') {
			/* If they want a real %, then they'd use &#37; */
			struct entity *e = PEReference(in);
			if (e)
				input_push(in, e);
		} else if (isCharRef(in)) {
			wstr_t v = CharRef(in);
			if (v) {
				wstr_addwstr(value, v);
				wstr_free(v);
			}
		} else 
			WSTR_ADD(value, next(in));

	if (in->stack->e != eof)
		error(in, "internal entity error");
	else 
		input_pop_raw(in);

	return value;
}

/* return true if a literal string follows */
static int
isLiteral(in)
	struct input *in;
{
	return in->next == '"' || in->next == '\'';
}

/* [10] AttValue - caller owns wstr */
static wstr_t
AttValue(in, da)
	struct input *in;
	struct dtdattr *da;
{
	wstr_t value, v;
	wchar_t quote;
	struct entity *eof;
	enum AttType type;

	DPRINTFN(in);

	/* [see EntityValue()'s comments] */

	quote = next(in);	/* ['"] */
	value = wstr_new();
	while (more(in) && in->next != quote)
		WSTR_ADD(value, next(in));
	next(in);

	type = da ? da->type : CDATA;

	DPRINTF("** attribute value - da %p, type %d\n", da, (int)type);

	eof = entity_eof();
	input_push(in, eof);
	input_push(in, entity_wstr(value));

	value = wstr_new();
	while (in->next != -1)
		if (isReference(in)) {
			v = Reference(in);
			if (v) {
				wstr_addwstr(value, v);
				wstr_free(v);
			}
		} else {
			wchar_t c = next(in);
			/* 'Normalize' whitespace */
			if (isSpace(c))
				c = 0x20;
			WSTR_ADD(value, c);
		}

	if (in->stack->e != eof)
		error(in, "internal entity error");
	else 
		input_pop_raw(in);

	DPRINTF("** AttValue: {1} normalized: '%s'\n", wstr_str(value));

	if (type != CDATA) {
		/* parse the tokens in the attribute value */

		eof = entity_eof();
		input_push(in, eof);
		input_push(in, entity_wstr(value));

		value = wstr_new();

		Sopt(in);
		if (type == NMTOKEN || type == NMTOKENS || type == ENUMERATION)
			v = Nmtoken(in);
		else
			v = Name(in);
		if (type == ENUMERATION)
			enumeration_check(in, da, v);

		/* VECTOR_APPEND(names, v); */
		wstr_addwstr(value, v);

		if (type == IDREFS || type == ENTITIES || type == NMTOKENS)
			while (more(in) && isS(in)) {
				S(in);
				if (in->next == -1)
					break;
				WSTR_ADD(value, ' ');
				if (type == NMTOKENS)
					v = Nmtoken(in);
				else
					v = Name(in);
				/* VECTOR_APPEND(names, v); */
				wstr_addwstr(value, v);
			}
		else
			Sopt(in);
		if (more(in))
			error(in, "expected all tokens in attribute value");

		/* Clean up the rest of the data */
		while (in->next != -1)
			next(in);
		if (in->stack->e != eof)
			error(in, "internal entity error");
		else 
			input_pop_raw(in);

		DPRINTF("** AttValue: {2} parsed: '%s'\n", wstr_str(value));
	}

	/* 
	 * (from the XML spec):
	 * "Note that if the unnormalized attribute value contains a
	 *  character reference to a white space character other than
	 *  space (#x20), the normalized value contains the referenced
	 *  character itself (#xD, #xA or #x9). This contrasts with the
	 *  case where the unnormalized value contains a white space
	 *  character (not a reference), which is replaced with a space
	 *  character (#x20) in the normalized value and also contrasts
	 *  with the case where the unnormalized value contains an entity
	 *  reference whose replacement text contains a white space
	 *  character; being recursively processed, the white space
	 *  character is replaced with a space character (#x20) in the
	 *  normalized value."
	 * Of course!!!
	 */

	return value;
}

static void
enumeration_check(in, da, v)
	struct input *in;
	struct dtdattr *da;
	wstr_t v;
{
	int i;
	wstr_t ex;

	for (i = 0; i < VECTOR_LEN(da->enums); i++)
		if (VECTOR_AT(da->enums, i) == v)
			return;
	ex = wstr_new();
	for (i = 0; i < VECTOR_LEN(da->enums); i++) {
		if (i) WSTR_ADD(ex, ' ');
		wstr_addwstr(ex, VECTOR_AT(da->enums, i));
	}
	error(in, "expected %s%s", i > 1 ? "one of " : "", wstr_str(ex));
	wstr_free(ex);
}

/* [11] SystemLiteral */
static wstr_t
SystemLiteral(in)
	struct input *in;
{
	wstr_t value;
	wchar_t quote;

	DPRINTFN(in);
	quote = next(in);	/* ['"] */
	value = wstr_new();
	if (quote != '"' && quote != '\'')
		error(in, "expected quoted literal");
	while (more(in) && in->next != quote)
		WSTR_ADD(value, next(in));
	next(in);
	DPRINTF("System -> %s\n", wstr_str(value));
	return value;
}

/* [12] PubidLiteral */
static wstr_t
PubidLiteral(in)
	struct input *in;
{
	wstr_t value;
	wchar_t quote;
	static const char valid[] = 
		"\x20\x0d\x0a"
		"abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		"0123456789"
		"-'()+,./:=?;!*#@$_%";

	DPRINTFN(in);
	quote = next(in);	/* ['"] */
	value = wstr_new();
	if (quote != '"' && quote != '\'')
		error(in, "expected quoted literal");
	while (more(in) && in->next != quote) {
		if (in->next >= 0x100 
		     || strchr(valid, (char)in->next) == NULL)
		{
			error(in, "illegal Pubid character \\x%04x",
				in->next);
			return value;
		}
		WSTR_ADD(value, next(in));
	}
	next(in); /* closing quote */
	DPRINTF("Pubid -> %s\n", wstr_str(value));
	return value;
}

/* common code for CharData(14) and CDSect(18) */
static wstr_t
cdata(in, iscdata)
	struct input *in;
	int iscdata;
{
	wstr_t w;

	w = wstr_new();
	while (more(in) && in->next != -1) {
		if (isCDEnd(in))
			break;
		if (!iscdata && (in->next == '<' || in->next == '&'))
			break;
		WSTR_ADD(w, next(in));
	}
	return w;
}

/* [14] Character data */
static void
CharData(in, el)
	struct input *in;
	xml_t el;
{
	xml_t n;

	DPRINTFN(in);
	n = newtextnode(el, cdata(in, 0));
	if (!n->order)
		n->order = doc_orderseq(in->doc);
}

/* [15] Comment */
static void
Comment(in, el)
	struct input *in;
	xml_t el;
{
	wchar_t c;
	wstr_t content;

	DPRINTFN(in);
	expect(in, "<!--");
	if (el)
		content = wstr_new();
	else
		content = NULL;
	while (more(in)) {
		while (more(in) && in->next != '-') {
			c = next(in);
			if (content)
				WSTR_ADD(content, c);
		}
		if (nextis(in, "-->"))
			break;
		c = next(in);
		if (content)
			WSTR_ADD(content, c);
	}
	expect(in, "-->");
	if (el && content) {
		xml_t n = newnode(COMMENT, el);
		n->aux.args = content;
		n->order = doc_orderseq(in->doc);
	}
}

static int
isComment(in)
	struct input *in;
{
	return nextis(in, "<!--");
}

/* [16] PI */
static xml_t
PI(in, parent)
	struct input *in;
	xml_t parent;
{
	xml_t pi;
	wstr_t name;
	wstr_t arg = NULL;

	DPRINTFN(in);
	expect(in, "<?");
	name = Name(in);
	Sopt(in);
	if (more(in) && !nextis(in, "?>")) {
		arg = wstr_new();
		while (more(in) && !nextis(in, "?>"))
			WSTR_ADD(arg, next(in));
	}
	expect(in, "?>");

	if (name && VECTOR_LEN(*name) >= 3
	    && (VECTOR_AT(*name, 0) == 'X' || VECTOR_AT(*name, 0) == 'x')
	    && (VECTOR_AT(*name, 0) == 'M' || VECTOR_AT(*name, 0) == 'm')
	    && (VECTOR_AT(*name, 0) == 'L' || VECTOR_AT(*name, 0) == 'l'))
	{
		/* XXX xml-stylesheet is kind of valid */
		if (0)
		error(in, "Processing instructions must not start with XML");
	}

	pi = newnode(PROCINST, parent);
	if (pi) {
		pi->name.ns = NULL;
		pi->name.local = name;
		pi->aux.args = arg;
		pi->order = doc_orderseq(in->doc);
	}
	return pi;
}

static int
isPI(in)
	struct input *in;
{
	return nextis(in, "<?");
}

/* [18] CDSect */
static void
CDSect(in, el)
	struct input *in;
	xml_t el;
{
	xml_t n;

	DPRINTFN(in);
	CDStart(in);
	n = newtextnode(el, CData(in));
	if (!n->order)
		n->order = doc_orderseq(in->doc);
	CDEnd(in);
}

/* [19] CDStart */
static void
CDStart(in)
	struct input *in;
{
	expect(in, "<![CDATA[");
}

static int
isCDStart(in)
	struct input *in;
{
	return nextis(in, "<![CDATA[");
}

/* [20] CData */
static wstr_t
CData(in)
	struct input *in;
{
	DPRINTFN(in);
	return cdata(in, 1);
}

/* [21] CDEnd */
static void
CDEnd(in)
	struct input *in;
{
	expect(in, "]]>");
}

static int
isCDEnd(in)
	struct input *in;
{
	return nextis(in, "]]>");
}

/* [22] prolog */
static void
prolog(in, root)
	struct input *in;
	xml_t root;
{
	DPRINTFN(in);
	if (nextis(in, "<?xml"))
		XMLDecl(in, root);
	while (isMisc(in))
		Misc(in, root);
	if (nextis(in, "<!DOCTYPE")) {
		doctypedecl(in, root);
		while (isMisc(in))
			Misc(in, root);
	}
}

/* [23] XMLDecl */
static void
XMLDecl(in, root)
	struct input *in;
	xml_t root;
{
	wstr_t version;
	wstr_t encoding = NULL;
	int isstandalone = 0;

	DPRINTFN(in);
	expect(in, "<?xml");
	version = VersionInfo(in);
	if (isEncodingDecl(in))
		encoding = EncodingDecl(in);
	if (isSDDecl(in))
		isstandalone = SDDecl(in);
	Sopt(in);
	expect(in, "?>");

	/* XXX - version & encoding currently ignored! */

	if (version)
		wstr_free(version);
	if (encoding)
		wstr_free(encoding);
}

/* [24] VersionInfo */
static wstr_t
VersionInfo(in)
	struct input *in;
{
	wstr_t version = NULL;

	DPRINTFN(in);
	S(in);
	expect(in, "version");
	Eq(in);
	if (!isLiteral(in))
		error(in, "expected ' or \"");
	else
		version = VersionNum(in);
	return version;
}

/* [25] Eq */
static void
Eq(in)
	struct input *in;
{
	DPRINTFN(in);
	Sopt(in);
	expect(in, "=");
	Sopt(in);
}

/* [26] VersionNum */
static wstr_t
VersionNum(in)
	struct input *in;
{
	wstr_t value;
	wchar_t quote;
	static const char *v, valid[] = 
		"abcdefghijklmnopqrstuvwxyz"
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		"0123456789"
		"_.:-";

	DPRINTFN(in);
	quote = next(in);	/* ['"] */
	value = wstr_new();
	if (in->next == quote)
		error(in, "expected non-empty version");
	while (more(in) && in->next != quote) {
		for (v = valid; *v; v++)
			if (*v == in->next)
				break;
		if (!*v) {
			error(in, "illegal character");
			return value;
		}
		WSTR_ADD(value, next(in));
	}
	next(in);
	return value;
}

/* [27] Misc */
static int
isMisc(in)
	struct input *in;
{
	return isComment(in) || isPI(in) || isS(in);
}

/* [27] Misc */
static void
Misc(in, root)
	struct input *in;
	xml_t root;
{
	DPRINTFN(in);
	if (isComment(in))
		Comment(in, root);
	else if (isPI(in))
		PI(in, root);
	else if (isS(in))
		S(in);
	else
		error(in, "expected PI, comment or whitespace");
}

/* [28] doctypedecl */
static void
doctypedecl(in, root)
	struct input *in;
	xml_t root;
{
	struct externalid *eid = NULL;

	DPRINTFN(in);
	expect(in, "<!DOCTYPE");
	S(in);
	root->aux.doc->doctype = Name(in);

	if (isS(in)) {
		S(in);
		if (isExternalID(in))
			eid = ExternalID(in);
	}
	Sopt(in);
	if (in->next == '[') {
		next(in);
		while (more(in) && in->next != ']') {
			if (isS(in))
				S(in);
			else if (in->next == '%') {
				/* %ref; gets exanded straight away */
				struct entity *e = PEReference(in);
				if (e)
					input_push(in, e);
			} else {
				markupdecl(in, root);
			}
		}
		expect(in, "]");
		Sopt(in);
	} else {
		/* XXX - should load in the doctype from eid */
		/*
		input_push((*in->fetch)(eid->publicid, eid->systemid));
		*/
	}
	expect(in, ">");
	if (eid)
		externalid_free(eid);
}

/* [29] markupdecl */
static void
markupdecl(in, root)
	struct input *in;
	xml_t root;
{
	DPRINTFN(in);
	if (nextis(in, "<!ELEMENT"))
		elementdecl(in);
	else if (nextis(in, "<!ATTLIST"))
		AttlistDecl(in);
	else if (nextis(in, "<!ENTITY"))
		EntityDecl(in);
	else if (nextis(in, "<!NOTATION"))
		NotationDecl(in);
	else if (isPI(in))
		PI(in, root);
	else if (isComment(in))
		Comment(in, NULL);
	else
		error(in, "expected markup declarations");
}

/* [32] SDDecl */
static int
SDDecl(in)
	struct input *in;
{
	wchar_t quote;
	int result = 0;

	DPRINTFN(in);
	S(in);
	expect(in, "standalone");
	Eq(in);
	if (!isLiteral(in))
		error(in, "expected quote");
	quote = next(in);
	if (nextis(in, "no"))
		result = 0;
	else if (nextis(in, "yes"))
		result = 1;
	else
		error(in, "expected 'yes' or 'no'");
	if (in->next == quote)
		next(in);
	else
		error(in, "expected closing quote");
	return result;
}

static int
isSDDecl(in)
	struct input *in;
{
	return nextis_S(in, "", "standalone");
}

/* [39] element */
static xml_t
element(in, parent)
	struct input *in;
	xml_t parent;
{
	xml_t el;
	struct dtdelement *de;

	DPRINTFN(in);
	el = newnode(ELEMENT, parent);
	el->order = doc_orderseq(in->doc);
	el->aux.element.attr = xhash_new(NULL, NULL);
	if (!STag(in, el, &de)) {	/* note: STag() or EmptyElemTag() */
		content(in, el, de);
		ETag(in, el);
	}

	DPRINTF("** ELEMENT -> name = %s\n", 
		qname_str(el->ns->aux.ns, &el->name));

	return el;
}

/* [40] STag / [44] EmptyElemTag  - returns true if matched empty elem tag */
static int
STag(in, el, dep)
	struct input *in;
	xml_t el;
	struct dtdelement **dep;
{
	int isEmptyElemtag = 0;
	xml_t *elp = &el->aux.element.first_attr;
	struct dtdelement *de;
	qname_t name;

	DPRINTFN(in);
	expect(in, "<");

	name = QName(in);

	/* Only look for DTD entries when the qname is not prefixed */
	if (name.ns == NULL)
	   de = (struct dtdelement *)xhash_lookup(in->doc->dtd, &name.local,
		sizeof name.local);
	else
	   de = NULL;

	DPRINTF("STag: de = %p\n", de);
	*dep = de;

	while (isS(in)) {
		S(in);
		if (in->next != '>' && in->next != '/')
			Attribute(in, el, &elp, de);
	}

	/*
	 * After processing attributes, we are now in a position to
	 * convert the qname's prefix part (currently held in name.ns)
	 * into a proper URI. el->ns is always valid.
	 */
	el->name = name;
	el->name.ns = ns_lookup(in, el->ns->aux.ns, name.ns);

	/* fill in default values for missing attributes */
	if (de) {
		struct xhash_iter *i;
		wstr_t name;
		int namelen = sizeof name;
		struct dtdattr *da;

		/* walk through all known attributes of this element */
		i = xhash_iter_new(de->attrs);
		while (xhash_iter_next(i, &name, &namelen, (void **)&da)) {
		    qname_t qn;

		    qn.ns = NULL;
		    qn.local = name;
		    if (!xhash_lookup(el->aux.element.attr, &qn, sizeof qn)) {
			/* not found; see if we need to add that attribute */
			default_attr(in, el, &elp, da, name);
		    }
		}
		xhash_iter_free(i);
	}

	if (in->next == '/') {
		next(in);
		isEmptyElemtag = 1;
		/* Make sure the DTD allows empty */
		if (de && de->fsm) {
			struct xfsm *start = xfsm_getstart(&de->fsm);
			if (!xfsm_isaccept(start))
				badcontent(in, de, start, NULL);
		}
	}

	expect(in, ">");

	return isEmptyElemtag;
}

/* create a new attribute node parented by el, and chain through elp */
static xml_t
newattrnode(in, el, elp, nm, v)
	struct input *in;
	xml_t el, **elp;
	qname_t nm;
	wstr_t v;
{
	xml_t attr = newnode(ATTR, NULL);
	if (attr) {
		attr->ns = el->ns;
		attr->parent = el;
		attr->order = doc_orderseq(in->doc);
		attr->aux.attr.value = v;
		attr->name = nm;
		xhash_store(el->aux.element.attr, &attr->name,
				sizeof attr->name, attr);
		**elp = attr;
		*elp = &attr->sibling;
	}
	return attr;
}

/* An attribute wasn't specified - see if we should default add it */
static xml_t
default_attr(in, el, elp, da, name)
	struct input *in;
	xml_t el, **elp;
	struct dtdattr *da;
	wstr_t name;
{
	wstr_t value = NULL;
	qname_t nm;

	switch (da->deftype) {
	case REQUIRED:
		error(in, "missing attribute %s required", 
			wstr_str(name));
		return NULL;
	case IMPLIED:
		return NULL;
	case FIXED:
		value = wstr_dup(da->defattr);
		break;
	default:
		/* fallthough */
	}

	nm.ns = NULL;
	nm.local = name;
	return newattrnode(in, el, elp, nm, value);
}

/* [41] Attribute */
static xml_t
Attribute(in, el, elp, de)
	struct input *in;
	xml_t el, **elp;
	struct dtdelement *de;
{
	xml_t attr;
	struct dtdattr *da;
	wstr_t value;
	qname_t name;
	
	DPRINTFN(in);
	name = QName(in);
	Eq(in);

	/* See if there is an ATTLIST entry for this attribute */
	if (de && name.ns == NULL && name.local) {
		da = (struct dtdattr *)xhash_lookup(de->attrs, &name.local,
			sizeof name.local);
		DPRINTF("** looked up name=%s in de=%p -> da=%p\n",
			qname_str(el->ns->aux.ns, &name), de, da);
	} else
		da = NULL;

	value = AttValue(in, da);

	/* xmlns:<local>="<value>" introduces a new namespace */
	if (name.ns == in->doc->xmlns) {
		value = intern_wstr(value);
		ns_enter(in->doc, el, name.local, value);
		return NULL;
	}

	/* xmlns="<value>" introduces a default, anonymous namespace */
	if (name.ns == NULL && name.local == in->doc->xmlns) {
		value = intern_wstr(value);
		ns_enter(in->doc, el, NULL, value);
		return NULL;
	}

	/*
	 * special case: null namespaces do not attract the default namespace,
	 * if the name has been found in the DTD
	 */
	if (!da || name.ns != NULL)
		name.ns = ns_lookup(in, el->ns->aux.ns, name.ns);

	/* Check for duplication */
	if (xhash_lookup(el->aux.element.attr, &name, sizeof name)) {
		error(in, "attribute %s duplicated", 
			qname_str(el->ns->aux.ns, &name));
		wstr_free(value);
		return NULL;
	}

	attr = newattrnode(in, el, elp, name, value);

	/* If an ID attribute is given, store a pointer to this node by ID */
	if (da && da->type == ID) {
		wstr_t id = intern_wstr_noown(value);
		if (id && !xhash_store(in->doc->id, &id, sizeof id, attr))
			error(in, "xhash_store: memory error");
	}

	return attr;
}

/* [42] ETag */
static void
ETag(in, el)
	struct input *in;
	xml_t el;
{
	qname_t name;

	DPRINTFN(in);
	expect(in, "</");
	if (!in->error) {
		name = QName(in);
		name.ns = ns_lookup(in, el->ns->aux.ns, name.ns);
		if (el->name.local != name.local || el->name.ns != name.ns)
			error(in, "expected </%s>", 
			      qname_str(el->ns->aux.ns, &el->name));
	}
	Sopt(in);
	expect(in, ">");
}

/* [43] content */
static void
content(in, el, de)
	struct input *in;
	xml_t el;
	struct dtdelement *de;
{
	struct xfsm *state = NULL, *nextstate;
	xml_t n;
	wstr_t cdata;

	DPRINTFN(in);
	if (de && de->fsm)
		state = xfsm_getstart(&de->fsm);

	if (de && de->ignorespace)
		Sopt(in);
	if (more(in) && in->next != '<' && in->next != '&') {
		CharData(in, el);
		if (state) {
			nextstate = xfsm_getnext(state, in->doc->pcdata);
			if (!nextstate)
				badcontent(in, de, state, in->doc->pcdata);
			state = nextstate;
		}
	}
	while (more(in) && !nextis(in, "</")) {
		wstr_t sym = NULL;

		if (isComment(in)) {
			Comment(in, el);
		} else if (isCDStart(in)) {
			CDSect(in, el);
			sym = in->doc->pcdata;
		} else if (in->next == '&') {
			cdata = Reference(in);
			if (cdata) {
				n = newtextnode(el, cdata);
				if (!n->order)
					n->order = doc_orderseq(in->doc);
				sym = in->doc->pcdata;
			} 
		} else if (isPI(in)) {
			PI(in, el);
		} else if (in->next == '<') {
			n = element(in, el);
			if (n)
				sym = n->name.local;	/* XXX */
		} else
			error(in, "expected element, reference, "
				  "CDATA, <? or comment");

		if (state && sym) {
			nextstate = xfsm_getnext(state, sym);
			if (!nextstate)
				badcontent(in, de, state, sym);
			state = nextstate;
		}

		if (de && de->ignorespace)
			Sopt(in);

		if (in->next != '<' && in->next != '&') {
			CharData(in, el);
			if (state) {
				nextstate = xfsm_getnext(state,
					in->doc->pcdata);
				if (!nextstate)
					badcontent(in, de, state,
						in->doc->pcdata);
				state = nextstate;
			}
		}
	}

	if (state && !xfsm_isaccept(state))
		badcontent(in, de, state, NULL);
}

/* Print out an FSM for debugging */
static void
debug_fsm_dump(de)
	struct dtdelement *de;
{
	struct xfsm *x;
	int j;
	wstr_t sym;
	extern struct xfsm *xfsm_listnext(struct xfsm *);

	fprintf(stderr, "FSM:\n");
	for (x = de->fsm; x; x = xfsm_listnext(x)) {
		fprintf(stderr, "    %p%s:\n", x,
			xfsm_isaccept(x) ? " (accept)": "");
		for (j = 0; xfsm_getpossible(x, j, 
		    (const void **)&sym); j++)
			fprintf(stderr, "    %s -> %p\n",
				wstr_str(sym), 
				xfsm_getnext(x, sym));
	}
}

/* Called when an element not expected in the DTD is encountered */
static void
badcontent(in, de, state, actualsym)
	struct input *in;
	struct dtdelement *de;
	struct xfsm *state;
	wstr_t actualsym;
{
	wstr_t poss;
	wstr_t sym;
	int i;
	char actbuf[80], *actual;

	if (actualsym == in->doc->pcdata)
		actual = "#PCDATA";
	else if (actualsym == NULL)
		actual = "end of content";
	else {
		snprintf(actbuf, sizeof actbuf, "<%s>", wstr_str(actualsym));
		actual = actbuf;
	}

	DPRINTF("** badcontent(state=%p, actual=%p/%s)\n", state,
		actualsym, actual);

	poss = wstr_new();
	for (i = 0; xfsm_getpossible(state, i, (const void **)&sym); i++) {
		if (i) {
			WSTR_ADD(poss, ',');
			WSTR_ADD(poss, ' ');
		}
		if (!sym)
			WSTR_ADD(poss, '?');
		else
			wstr_addwstr(poss, sym);
	}
	if (i > 1)
		error(in, "expected one of (%s)%s but got %s", wstr_str(poss),
			xfsm_isaccept(state) ? " or end of content" : "",
			actual);
	else if (i == 1)
		error(in, "expected %s%s but got %s", wstr_str(poss),
			xfsm_isaccept(state) ? " or end of content" : "",
			actual);
	else if (xfsm_isaccept(state))
		error(in, "expected end of content but got %s", actual);
	else {
		error(in, "DTD state machine internal error");
		if (xml_debug)
			debug_fsm_dump(de);
	}
	wstr_free(poss);
}

/* [44] EmptyElemTag  -- see [40] Stag */

/* pe: read up to the next ">", expanding % references, then push it back */
static void
pe_expand(in)
	struct input *in;
{
	struct entity *e;
	wstr_t	decl;

	DPRINTFN(in);
	decl = wstr_new();
	while (more(in) && in->next != '>') {
		if (in->next == '%') {
			e = PEReference(in);
			if (e && e->inuse)
				error(in, "recursive reference");
			else if (e) {
				ENTITY_REWIND(e);
				while (!ENTITY_EOF(e))
					if (ENTITY_GETCH(e) == '>') {
						error(in, "illegal > in "
						      "expanded % reference");
						break;
					}
				wstr_addentity(decl, e);
				ENTITY_DECREF(e);
			}
		} else
			WSTR_ADD(decl, next(in));
	}
	input_push(in, entity_wstr(decl));
}

/* [45] elementdecl */
static void
elementdecl(in)
	struct input *in;
{
	wstr_t name;
	struct dtdelement *de;

	DPRINTFN(in);
	expect(in, "<!ELEMENT");
	S(in);
	pe_expand(in); Sopt(in);
	name = Name(in);
	S(in);

	/* Look up in case an AttlistDecl has already created it */
	de = (struct dtdelement *)xhash_lookup(in->doc->dtd, &name,
	    sizeof name);
	if (!de) {
		de = dtdelement_new();
		if (!de)
			error(in, "dtdelement_new(): memory allocation");
	} else if (de && de->fsm) {
		error(in, "duplicate element declaration");
		return;
	}
	contentspec(in, de);
	if (de && !xhash_store(in->doc->dtd, &name, sizeof name, de))
		error(in, "xhash_store: memory allocation");
	Sopt(in);
	expect(in, ">");
}

/* [46] contentspec */
static void
contentspec(in, de)
	struct input *in;
	struct dtdelement *de;
{
	struct xfsm *a, *b;

	DPRINTFN(in);
	if (nextis(in, "EMPTY")) {
		expect(in, "EMPTY");
		if (de) {
			if (!in->novalidate)
				xfsm_empty(&de->fsm, &a, &b);
			de->ignorespace = 1;
		}
	} else if (nextis(in, "ANY")) {
		expect(in, "ANY");
		if (de)
			de->ignorespace = 0;
		/* retain fsm = NULL */
	} else if (isMixed(in)) {
		Mixed(in, in->novalidate ? NULL : &de->fsm, &a, &b);
		if (de)
			de->ignorespace = 0;
	} else {
		children(in, in->novalidate ? NULL : &de->fsm, &a, &b);
		if (de)
			de->ignorespace = 1;
	}
	if (de && !in->novalidate) {
#if 0
		extern void reduce1(struct xfsm **);
		extern void reduce2(struct xfsm **);
		extern void reduce3(struct xfsm **);
		extern void reduce4(struct xfsm **);

		debug_fsm_dump(de);
		fprintf(stderr, "REDUCE 1\n");
		reduce1(&de->fsm);
		debug_fsm_dump(de);
		fprintf(stderr, "REDUCE 2\n");
		reduce2(&de->fsm);
		debug_fsm_dump(de);
		fprintf(stderr, "REDUCE 3\n");
		reduce3(&de->fsm);
		debug_fsm_dump(de);
		fprintf(stderr, "REDUCE 4\n");
		reduce4(&de->fsm);
		debug_fsm_dump(de);
#else
		/*
		 * In appendix E of the XML spec, it says that,
		 * for compatibility, the contentspecs need to be
		 * deterministic. It goes on to recommend that
		 * you should compute follows-sets and construct
		 * an FSM that way.
		 *
		 * But, I didn't read this before using a NDFSM
		 * engine I written previously. It converts
		 * non-deterministic FSMs into deterministic ones,
		 * but with a fairly large time penalty.
		 *
		 * If you want to speed it up, and assume determinism,
		 * then you really only need to rewrite xfsm.c.
		 */
		DPRINTF("computing deterministic FSM for DTD...\n");
		xfsm_reduce(&de->fsm);
		DPRINTF("\tdone\n");
#endif
	}
}

/* [47] children */
static void
children(in, h, a, b)
	struct input *in;
	struct xfsm **h, **a, **b;
{
	/*
	 * The given grammar is:
	 *
	 *    [47] children := (choice | seq) ('?' | '*' | '+')?
	 *    [48] cp       := (Name | choice | seq) ('?' | '*' | '+')?
	 *    [49] choice   := '(' S? cp (S? '|' S? cp)+ S? ')'
	 *    [50] seq      := '(' S? cp (S? ',' S? cp)* S? ')'
	 *
	 * This can be converted into an LA(1) grammar:
	 *
	 *    children  := '(' S? cp S? (lchoice+ | lseq*) ')' lloop?
	 *    cp        := children | Name lloop?
	 *    lchoice   := '|' S? cp S?
	 *    lseq      := ',' S? cp S?
	 *    lloop	:= '?' | '*' | '+'
	 */

	DPRINTFN(in);
	expect(in, "(");
	Sopt(in);
	cp(in, h, a, b);
	Sopt(in);
	if (in->next == '|')
		while (in->next == '|')
			lchoice(in, h, a, b);
	else
		while (in->next == ',')
			lseq(in, h, a, b);
	expect(in, ")");
	/* could be a bit looser here and allow Sopt() */
	if (islloop(in))
		lloop(in, h, a, b);
}

static void
cp(in, h, a, b)
	struct input *in;
	struct xfsm **h, **a, **b;
{
	wstr_t name;

	DPRINTFN(in);
	if (in->next == '(')
		children(in, h, a, b);
	else {
		name = Name(in);
		if (name) 
			xfsm_simple(h, name, a, b);
	}
	if (islloop(in))
		lloop(in, h, a, b);
}

static void
lchoice(in, h, a, b)
	struct input *in;
	struct xfsm **h, **a, **b;
{
	struct xfsm *c, *d;

	DPRINTFN(in);
	expect(in, "|");
	Sopt(in);
	cp(in, h, &c, &d);
	if (more(in) && !xfsm_alt(h, *a, *b, c, d, a, b))
		error(in, "xfsm_alt: allocation error");
	Sopt(in);
}

static void
lseq(in, h, a, b)
	struct input *in;
	struct xfsm **h, **a, **b;
{
	struct xfsm *c, *d;

	DPRINTFN(in);
	expect(in, ",");
	Sopt(in);
	cp(in, h, &c, &d);
	if (more(in) && !xfsm_seq(h, *a, *b, c, d, a, b))
		error(in, "xfsm_seq: allocation error");
	Sopt(in);
}

static void
lloop(in, h, a, b)
	struct input *in;
	struct xfsm **h, **a, **b;
{
	int ok = 1;

	DPRINTFN(in);
	if (in->next == '*') {
		next(in);
		ok = xfsm_kleene(h, *a, *b, a, b);
	} else if (in->next == '+') {
		next(in);
		ok = xfsm_plus(h, *a, *b, a, b);
	} else if (in->next == '?') {
		next(in);
		ok = xfsm_opt(h, *a, *b, a, b);
	} else
		error(in, "expected one of '*', '+' or '?'");
	if (!ok)
		error(in, "xfsm: allocation error");
}

static int
islloop(in)
	struct input *in;
{
	return in->next == '+' || in->next == '?' || in->next == '*';
}

/* [51] Mixed */
static void
Mixed(in, h, a, b)
	struct input *in;
	struct xfsm **h, **a, **b;
{
	struct xfsm *c, *d;
	wstr_t name;

	DPRINTFN(in);
	expect(in, "(");
	Sopt(in);
	expect(in, "#PCDATA");
	if (!xfsm_simple(h, in->doc->pcdata, a, b))
		error(in, "xfsm_simple: allocation error");
	Sopt(in);
	if (in->next == '|') {
		while (in->next == '|') {
			next(in);
			Sopt(in);
			name = Name(in);
			if (!xfsm_simple(h, name, &c, &d))
				error(in, "xfsm_simple: allocation error");
			else if (!xfsm_alt(h, *a, *b, c, d, a, b))
				error(in, "xfsm_alt: allocation error");
			Sopt(in);
		}
		expect(in, ")*");
	} else
		expect(in, ")");
	/*
	 * Kleene-star the lot.
	 * Note this allows (#PCDATA) to match an empty node,
	 * and also to allow multiple text nodes to match.
	 */
	if (!xfsm_kleene(h, *a, *b, a, b))
		error(in, "xfsm_kleene: allocation error");
}

static int
isMixed(in)
	struct input *in;
{
	return nextis(in, "(#PCDATA") || nextis_S(in, "(", "#PCDATA");
}

/* [52] AttlistDecl */
static void
AttlistDecl(in)
	struct input *in;
{
	wstr_t name;
	struct dtdelement *de;

	DPRINTFN(in);
	expect(in, "<!ATTLIST");
	S(in);
	pe_expand(in); Sopt(in);
	name = Name(in);
	de = xhash_lookup(in->doc->dtd, &name, sizeof name);
	if (!de && name && !in->novalidate) {
		/*
		 * According to 3.3, this is not an error.
		 * So, we create a dtd entry for the not-yet-defined element.
		 */
		de = dtdelement_new();
		if (!de || !xhash_store(in->doc->dtd, &name,
		    sizeof name, de))
			error(in, "xhash_store: memory allocation");
	}
	while (isS(in)) {
		S(in);
		if (in->next == '>')
			break;
		AttDef(in, de);
	}
	expect(in, ">");
}

/* [53] AttDef */
static void
AttDef(in, de)
	struct input *in;
	struct dtdelement *de;
{
	wstr_t name;
	struct dtdattr *da;

	DPRINTFN(in);
	name = Name(in);
	S(in);
	da = AttType(in);
	S(in);
	DefaultDecl(in, da);
	if (!de && da) {
		dtdattr_free(da);
		da = NULL;
	}
	if (de && da) {
		if (xhash_lookup(de->attrs, &name, sizeof name) != NULL) {
		    DPRINTF("AttDef: duplicate %s\n", wstr_str(name));
		    /* only the first declaration is binding - ignore others */
		    dtdattr_free(da);
		} else {
		    DPRINTF("AttDef: storing da=%p for name=%s in de=%p\n",
			da, wstr_str(name), de);
		    if (!xhash_store(de->attrs, &name, sizeof name, da))
			error(in, "xhash_store: memory allocation");
		    if (da->type == ID) {
			if (de->idattr)
			   error(in, "Only one attribute can be ID type");
			else
			   de->idattr = name;
		    }
		}
	}
}

/* [54] AttType */
static struct dtdattr *
AttType(in)
	struct input *in;
{
	DPRINTFN(in);
	if (in->next == '(')
		return Enumeration(in);

#define DO(t) if (nextis(in, #t)) { expect(in, #t); return dtdattr_new(t); }

	else DO(CDATA)				/* [55] */
	else DO(NMTOKENS)			/* [55] */
	else DO(NMTOKEN)			/* [55] */
	else DO(ENTITIES)			/* [55] */
	else DO(ENTITY)				/* [55] */
	else DO(IDREFS)				/* [55] */
	else DO(IDREF)				/* [55] */
	else DO(ID)				/* [55] */

#undef DO

	else if (nextis(in, "NOTATION"))
		return NotationType(in);
	else {
		error(in, "expected attribute type declaration");
		return NULL;
	}
}

/* [58] NotationType */
static struct dtdattr *
NotationType(in)
	struct input *in;
{
	struct dtdattr *da;
	wstr_t name;

	DPRINTFN(in);
	expect(in, "NOTATION");
	da = dtdattr_new(NOTATION);
	S(in);
	expect(in, "(");
	do {
		Sopt(in);
		name = Name(in);
		if (da && name)
			VECTOR_APPEND(da->enums, name);
		Sopt(in);
		if (!more(in) || in->next != '|')
			break;
		next(in); /* | */
	} while (1);
	expect(in, ")");
	return da;
}

/* [59] Enumeration */
static struct dtdattr *
Enumeration(in)
	struct input *in;
{
	struct dtdattr *da;
	wstr_t name;

	DPRINTFN(in);
	expect(in, "(");
	da = dtdattr_new(ENUMERATION);
	do {
		Sopt(in);
		name = Nmtoken(in);
		if (name && da)
			VECTOR_APPEND(da->enums, name);
		Sopt(in);
		if (!more(in) || in->next != '|')
			break;
		next(in); /* | */
	} while (1);
	expect(in, ")");
	return da;
}

/* [60] DefaultDecl */
static void
DefaultDecl(in, da)
	struct input *in;
	struct dtdattr *da;
{
	wstr_t value;

	DPRINTFN(in);
	if (nextis(in, "#REQUIRED")) {
		expect(in, "#REQUIRED");
		if (da)
			da->deftype = REQUIRED;
	} else if (nextis(in, "#IMPLIED")) {
		expect(in, "#IMPLIED");
		if (da)
			da->deftype = IMPLIED;
	} else {
		if (nextis(in, "#FIXED")) {
			expect(in, "#FIXED");
			S(in);
		}
		value = AttValue(in, da);
		if (da) {
			da->deftype = FIXED;
			da->defattr = value;
		}
	}
}

/* [66] CharRef - caller owns wstr */
static wstr_t
CharRef(in)
	struct input *in;
{
	wchar_t c = 0;
	wstr_t value;

	DPRINTFN(in);
	value = wstr_new();
	expect(in, "&#");
	if (more(in) && in->next == 'x') {
		next(in);
		while (more(in) && in->next != ';')
			if ('a' <= in->next && in->next <= 'f')
				c = c*16 + next(in) - 'a' + 10;
			else if ('A' <= in->next && in->next <= 'F')
				c = c*16 + next(in) - 'A' + 10;
			else if ('0' <= in->next && in->next <= '9')
				c = c*16 + next(in) - '0';
			else 
				error(in, "expected hex digits");
	} else
		while (more(in) && in->next != ';')
			if ('0' <= in->next && in->next <= '9')
				c = c*10 + next(in) - '0';
			else
				error(in, "expected numbers");
	expect(in, ";");
	WSTR_ADD(value, c);
	return value;
}

static int
isCharRef(in)
	struct input *in;
{
	return nextis(in, "&#");
}

/* [67] Reference - caller owns wstr, else NULL if data was pushed */
static wstr_t
Reference(in)
	struct input *in;
{
	DPRINTFN(in);
	if (nextis(in, "&#"))
		return CharRef(in);
	else
		return EntityRef(in);
}

static int
isReference(in)
	struct input *in;
{
	return in->next == '&';
}

/* [68] EntityRef - caller owns wstr */
static wstr_t
EntityRef(in)
	struct input *in;
{
	wstr_t name;
	struct entity *e = NULL;

	DPRINTFN(in);
	expect(in, "&");
	name = Name(in);
	expect(in, ";");
	if (name)
		e = xhash_lookup(in->doc->entity, &name, sizeof name);
	if (!e)
		error(in, "unknown entity &%s;", wstr_str(name));
	else if (e->inuse)
		error(in, "recursive use of entity &%s;", wstr_str(name));
	else {
		DPRINTF("pushing entity &%s; onto input\n", wstr_str(name));
		ENTITY_INCREF(e);
		input_push(in, e);
	}
	return NULL;
}

/* [69] PEReference - caller owns entity reference */
static struct entity *
PEReference(in)
	struct input *in;
{
	wstr_t name;
	struct entity *e = NULL;

	DPRINTFN(in);
	expect(in, "%");
	name = Name(in);
	expect(in, ";");
	if (name)
		e = xhash_lookup(in->doc->pentity, &name, sizeof name);
	if (!e)
		error(in, "unknown entity %%s;", wstr_str(name));
	else
		ENTITY_INCREF(e);
	return e;
}

/* [70] EntityDecl */
static void
EntityDecl(in)
	struct input *in;
{
	wstr_t name;
	struct entity *e;

	DPRINTFN(in);
	expect(in, "<!ENTITY");
	S(in);
	if (in->next == '%') {		/* [72] PEDecl */
		next(in);
		S(in);
		name = Name(in);
		S(in);
		e = EntityDef(in);
		if (!xhash_store(in->doc->pentity, &name, sizeof name, e))
			error(in, "out of memory");
	} else {			/* [71] GEDecl */
		name = Name(in);
		S(in);
		e = PEDef(in);
		if (!xhash_store(in->doc->entity, &name, sizeof name, e))
			error(in, "out of memory");
	}
	Sopt(in);
	expect(in, ">");
}

/* [73] EntityDef */
static struct entity *
EntityDef(in)
	struct input *in;
{
	wstr_t value, ndata;
	struct externalid *eid;

	DPRINTFN(in);
	if (in->next == '"' || in->next == '\'') {
		value = EntityValue(in);
		return entity_wstr(value);
	}

	eid = ExternalID(in);
	if (isNDataDecl(in))
		ndata = NDataDecl(in);
	else
		ndata = NULL;

	/* XXX - should look up entity using eid and (intern'd) ndata */
	if (eid)
		externalid_free(eid);
	return NULL;
}

/* [74] PEDef */
static struct entity *
PEDef(in)
	struct input *in;
{
	wstr_t value;
	struct externalid *eid;

	DPRINTFN(in);
	if (in->next == '"' || in->next == '\'') {
		value = EntityValue(in);
		return entity_wstr(value);
	} else {
		eid = ExternalID(in);
		/* XXX - should look up this entity */
		externalid_free(eid);
		return NULL;
	}
}

/* [75] ExternalID */
static struct externalid *
ExternalID(in)
	struct input *in;
{
	wstr_t sid, pid;

	DPRINTFN(in);
	if (nextis(in, "SYSTEM")) {
		expect(in, "SYSTEM");
		S(in);
		sid = SystemLiteral(in);
		return externalid_new(sid, NULL);
	}
	if (nextis(in, "PUBLIC")) {
		expect(in, "PUBLIC");
		S(in);
		pid = PubidLiteral(in);
		S(in);
		sid = SystemLiteral(in);
		return externalid_new(sid, pid);
	}
	error(in, "expected SYSTEM or PUBLIC");
	return NULL;
}

static int
isExternalID(in)
	struct input *in;
{
	return nextis(in, "SYSTEM") || nextis(in, "PUBLIC");
}

/* [76] NDataDecl */
static wstr_t
NDataDecl(in)
	struct input *in;
{
	DPRINTFN(in);
	S(in);
	expect(in, "NDATA");
	S(in);
	return Name(in);
}

static int
isNDataDecl(in)
	struct input *in;
{
	return nextis_S(in, "", "NDATA");
}

/* [77] TextDecl */
/* [78] extParsedEnt */

/* [80] EncodingDecl */
static wstr_t
EncodingDecl(in)
	struct input *in;
{
	wstr_t name = NULL;
	wchar_t quote;

	DPRINTFN(in);
	S(in);
	expect(in, "encoding");
	Eq(in);
	if (!isLiteral(in))
		error(in, "expected ' or \"");
	else {
		quote = next(in);
		name = EncName(in);
		if (quote != in->next)
			error(in, "expected %c", quote);
		else
			next(in);
	}
	return name;
}

static int
isEncodingDecl(in)
	struct input *in;
{
	return nextis_S(in, "", "encoding");
}

/* [81] EncName */
static wstr_t
EncName(in)
	struct input *in;
{
	wstr_t name = wstr_new();

	DPRINTFN(in);
	if (('A' <= in->next && in->next <= 'Z') ||
	    ('a' <= in->next && in->next <= 'z'))
	{
		WSTR_ADD(name, next(in));
		while (('A' <= in->next && in->next <= 'Z') ||
		       ('a' <= in->next && in->next <= 'z') ||
		       ('0' <= in->next && in->next <= '9') ||
		       in->next == '.' ||
		       in->next == '_' ||
		       in->next == '-')
			WSTR_ADD(name, next(in));
	}
	else
		error(in, "expected [A-Za-z]");
	return name;
}

/* [82] NotationDecl */
static void
NotationDecl(in)
	struct input *in;
{
	wstr_t name, sysid, pubid;
	struct externalid *eid;

	DPRINTFN(in);
	expect(in, "<!NOTATION");
	S(in);
	pe_expand(in);	Sopt(in);
	name = Name(in);
	S(in);
	if (nextis(in, "SYSTEM")) {
		S(in);
		pubid = NULL;
		sysid = SystemLiteral(in);
	} else {
		/* grammar at ExternalID|[83]PublicID is not LA(1) */
		expect(in, "PUBLIC");
		S(in);
		pubid = PubidLiteral(in);
		sysid = NULL;
		if (isS(in)) {
			S(in);
			if (in->next == '"' || in->next == '\'')
				sysid = SystemLiteral(in); 
		}
	}
	Sopt(in);
	expect(in, ">");
	eid = externalid_new(sysid, pubid);
	if (!xhash_store(in->doc->notation, &name, sizeof name, eid))
		error(in, "out of memory");
}

/*------------------------------------------------------------
 * XML node management
 */

/*
 * XPath section 5.7 wants text nodes to contain as much
 * consecutive text as is possible, so we either construct
 * a new text node (through newnode()) with empty text, or
 * return the last text node of the parent. In any case,
 * the givent wstr is appended to aux.args.
 */
static xml_t
newtextnode(parent, w)
	xml_t parent;
	wstr_t w;
{
	xml_t n = NULL;

	if (parent)
		for (n = parent->child; n; n = n->sibling)
			if (n->sibling == NULL)
				break;
	if (!n || n->type != TEXT) {
		n = newnode(TEXT, parent);
		n->aux.args = w;
	} else {
		wstr_addwstr(n->aux.args, w);
		wstr_free(w);
	}
	return n;
}

/* create a new XML node of the given type */
static xml_t
newnode(type, parent)
	enum xmltype type;
	xml_t parent;
{
	xml_t x = (xml_t)malloc(sizeof (struct xml));

	if (!x)
		return NULL;

	x->type = type;
	x->user = NULL;
	x->userfree = NULL;

	x->parent = parent;
	x->child = NULL;
	x->sibling = NULL;
	x->lastchildp = &x->child;
	if (parent) {
		*parent->lastchildp = x;
		parent->lastchildp = &x->sibling;
		x->ns = parent->ns;
	} else {
		x->ns = NULL;
	}
	x->id = NULL;
	x->aux.ptr = NULL;
	x->order = 0;
	return x;
}

static void
freenode(x)
	xml_t x;
{
	free(x);
}

/*------------------------------------------------------------
 * XML public API
 */

/* new XML object (entity reference is loaned) */
xml_t
xml_newe(e, fetch, novalidate)
	struct entity *e;
	struct entity *(*fetch)(const char *pub, const char *sys);
	int novalidate;
{
	xml_t root;
	struct xmldocinfo *doc;
	struct input *in = NULL;

	doc = doc_new();
	if (doc)
		in = input_new(doc, fetch, novalidate);
	if (!in) {
		if (doc) doc_free(doc);
		ENTITY_DECREF(e);
		xml_last_error = "initialisation error";
		return NULL;
	}

	e->flags |= ENTITY_ISEXTERNAL;
	input_push(in, e);

	root = document(in);

	xml_last_error = in->error;
	if (in->error) {
		xml_free(root);
		root = NULL;
	} else if (root == NULL)
		xml_last_error = "unknown error";

	input_free(in);
	return root;
}

xml_t
xml_new_utf8(buf)
	const char *buf;
{
	struct entity *e;
	xml_t x;

	e = entity_utf8mem(buf, buf + strlen(buf), 1);
	x = xml_newe(e, NULL, 0);
	return x;
}

xml_t
xml_new(buf, size)
	const wchar_t *buf;
	size_t size;
{
	struct entity *e;
	xml_t x;

	e = entity_mem(buf, buf + size / sizeof *buf, 1);
	x = xml_newe(e, NULL, 0);
	return x;
}

/* Recursively free all the parts of an XML tree */
void
xml_free(x)
	xml_t x;
{
	if (x == NULL)
		return;
	DPRINTF("xml_free(%p %s)\n", x, 
		x->type == ROOT ? "ROOT" :
		x->type == ELEMENT ? "ELEMENT" :
		x->type == TEXT ? "TEXT" :
		x->type == ATTR ? "ATTR" :
		x->type == NS ? "NS" :
		x->type == PROCINST ? "PROCINST" :
		x->type == COMMENT ? "COMMENT" :
		"?");
	xml_free(x->child); x->child = NULL;
	if (x->userfree) {
		(*x->userfree)(x->user, x);
		x->userfree = NULL;
		x->user = NULL;
	}
	switch (x->type) {
	case ROOT:
		if (x->aux.doc)
			doc_free(x->aux.doc);
		break;
	case ELEMENT:
		xml_free(x->aux.element.first_attr);
		xhash_free(x->aux.element.attr);
		break;
	case TEXT:
		wstr_free(x->aux.args);
		break;
	case ATTR:
		wstr_free(x->aux.attr.value);
		/* VECTOR_FREE(x->aux.attr.names); */
		break;
	case NS:
		xmlns_free(x->aux.ns);
		break;
	case PROCINST:
		wstr_free(x->aux.args);
		break;
	case COMMENT:
		wstr_free(x->aux.args);
		break;
	}
	xml_free(x->sibling); x->sibling = NULL;
	freenode(x);
}

void *
xml_getuser(x)
	xml_t x;
{
	return x->user;
}

void
xml_setuser(x, user, userfree)
	xml_t x;
	void *user;
	void (*userfree)(void *, xml_t);
{
	x->user = user;
	x->userfree = userfree;
}

/*
 * Write the XML document out in UTF8 representation using puts().
 * If puts returns zero, then writing is aborted and the zero is
 * returned. Otherwise non-zero is returned.
 */

static int
xml_puts_sib(node, puts, closure, dosib, dochld)
	xml_t node;
	int (*puts)(const char *, void *);
	void *closure;
	int dosib, dochld;
{
	xml_t y;
	wstr_t sv;

	if (node == NULL)
		return 1;

	if (xml_debug) {
		static char buf[10];
		snprintf(buf, sizeof buf, "[%u]", node->order);
		(*puts)(buf, closure);
	}

	switch (node->type) {
	case ROOT:
		if (!(*puts)("<?xml version=\"1.0\"?>", closure))
			return 0;
		if (dochld) 
			return (*puts)("\n", closure)
			    && xml_puts_sib(node->child, puts, closure, 1, 1);
		return 1; /* no siblings possible */
	case ELEMENT:
		if (!(*puts)("<", closure))
			return 0;
		if (!(*puts)(qname_str(node->ns->aux.ns, &node->name), closure))
			return 0;
	if (dochld) {
		/* Print introduced namespaces */
		for (y = node->ns; 
		     y && y != node->parent->ns;
		     y = y->ns)
		{
			if (y->name.local == NULL) {
				if (!(*puts)(" xmlns=", closure))
					return 0;
			} else {
				if (!(*puts)(" ", closure))
					return 0;
				if (!puts_wstr(y->name.local, 
				    puts, closure, 0))
					return 0;
				if (!(*puts)("=", closure))
					return 0;
			}
			sv = xmlns_lookup(y->aux.ns, y->name.local);
			if (sv == NULL) {
				if (!(*puts)("\"\"", closure))
					return 0;
			} else {
				if (!puts_wstr(sv, puts, closure, 1))
					return 0;
			}
		}

		/* Print other attributes */
		if (!xml_puts_sib(node->aux.element.first_attr, puts,
		    closure, 1, 1))
			return 0;
		if (!node->child) {
			if (!(*puts)("/>", closure))
				return 0;
			return 1;
		}
		if (!(*puts)(">", closure))
			return 0;
		/* Content */
		if (!xml_puts_sib(node->child, puts, closure, 1, 1))
			return 0;
		/* End tag */
		if (!(*puts)("</", closure))
			return 0;
		if (!(*puts)(qname_str(node->ns->aux.ns, &node->name), closure))
			return 0;
	}
		if (!(*puts)(">", closure))
			return 0;
		break;
	case TEXT:
		if (!puts_wstr(node->aux.args, puts, closure, 0))
			return 0;
		break;
	case ATTR:
		if (!(*puts)(" ", closure))
			return 0;
		if (!(*puts)(qname_str(node->ns->aux.ns, &node->name), closure))
			return 0;
		if (!(*puts)("=", closure))
			return 0;
		if (!puts_wstr(node->aux.attr.value, puts, closure, 1))
			return 0;
		break;
	case NS:
		/* should never reach this node! */
		break;
	case PROCINST:
		if (!(*puts)("<?", closure))
			return 0;
		if (!puts_wstr(node->name.local, puts, closure, 0))
			return 0;
		if (node->aux.args) {
			if (!(*puts)(" ", closure))
				return 0;
			if (!puts_wstr(node->aux.args, puts, closure, 0))
				return 0;
		}
		if (!(*puts)("?>", closure))
			return 0;
		break;
	case COMMENT:
		if (!(*puts)("<!--", closure))
			return 0;
		if (!puts_wstr(node->aux.args, puts, closure, -1))
			return 0;
		if (!(*puts)("-->", closure))
			return 0;
		break;
	}
	if (dosib && node->parent && node->parent->type == ROOT) {
		if (!(*puts)("\n", closure))
			return 0;
	}
	if (dosib)
		return xml_puts_sib(node->sibling, puts, closure, 1, dochld);
	return 1;
}

int
xml_puts(node, puts, closure)
	xml_t node;
	int (*puts)(const char *, void *);
	void *closure;
{
	return xml_puts_sib(node, puts, closure, 0, 1);
}

int
xml_puts1(node, puts, closure)
	xml_t node;
	int (*puts)(const char *, void *);
	void *closure;
{
	return xml_puts_sib(node, puts, closure, 0, 0);
}

/*------------------------------------------------------------
 * XML output functions
 */

int
puts_stdio(buf, closure)
	const char *buf;
	void *closure;
{
	FILE *f = (FILE *)closure;

	if (fputs(buf, f) == 0)
		return 1;
	else
		return 1;
}

#if 0
static int
puts_entity(e, puts, closure, quote)
	struct entity *e;
	int (*puts)(const char *, void *);
	void *closure;
	int quote;
{
	char s[7], *p;
	wchar_t c;

	if (e->inuse) {
		(*puts)("\n<!-- entity in use -->\n", closure);
		return 0;		/* entity in use */
	}
	ENTITY_REWIND(e);
	if (quote && !(*puts)("\"", closure))
		return 0;
	while (!ENTITY_EOF(e)) {
		p = s;
		c = ENTITY_GETCH(e);
		if (c == '<' || c == '>' || c == '&' || (quote && c == '"')) {
			*p++ = '&';
			switch (c) {
			case '<':  *p++ = 'l'; *p++ = 't'; break;
			case '>':  *p++ = 'g'; *p++ = 't'; break;
			case '&':  *p++ = 'a'; *p++ = 'm'; *p++ = 'p'; break;
			case '"':  *p++ = 'q'; *p++ = 'u'; 
				   *p++ = 'o'; *p++ = 't'; break;
			}
			*p++ = ';';
		} else
			utf8_put(c, &p);
		*p = '\0';
		if (!(*puts)(s, closure))
			return 0;
	}
	if (quote && !(*puts)("\"", closure))
		return 0;
	return 1;
}
#endif

int
puts_wstr(w, puts, closure, quote)
	wstr_t	w;
	int (*puts)(const char *, void *);
	void *closure;
	int quote;
{
	int len = 6 * VECTOR_LEN(*w) + 1;
	char *s = alloca(len + 6 + 2);
	int i;
	char *p;

	if (w == NULL)
		return (*puts)("(null)", closure);

	p = s;
	if (quote > 0) *p++ = '"';
	for (i = 0; p < s + len && i < VECTOR_LEN(*w); i++) {
		wchar_t c = VECTOR_AT(*w, i);
		if (quote >= 0 && (c == '<' || c == '>' || c == '&' 
				   || (quote>0 && c == '"'))) {
			*p++ = '&';
			switch (c) {
			case '<':  *p++ = 'l'; *p++ = 't'; break;
			case '>':  *p++ = 'g'; *p++ = 't'; break;
			case '&':  *p++ = 'a'; *p++ = 'm'; *p++ = 'p'; break;
			case '"':  *p++ = 'q'; *p++ = 'u'; 
				   *p++ = 'o'; *p++ = 't'; break;
			}
			*p++ = ';';
		} else
			utf8_put(c, &p);
	}
	if (quote > 0 && p < s + len) *p++ = '"';
	if (p < s + len) *p++ = '\0';
	return (*puts)(s, closure);
}

/* return a representation of a qname for printing */
const char *
qname_str(ns, q)
	struct xmlns *ns;
	qname_t *q;
{
	static char buf[1024];
	wstr_t prefix;
	int pos;

	prefix = xmlns_finduri(ns, q->ns);
	if (prefix) {
		snprintf(buf, sizeof buf, "%s:", wstr_str(prefix));
		pos = strlen(buf);
	} else
		pos = 0;
	snprintf(buf + pos, sizeof buf - pos, "%s", wstr_str(q->local));
	return buf;
}

