/* $Id: xmlP.h,v 1.6 2002/07/18 05:30:07 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * Private data structures for XML documents
 * 
 * XML 1.0, http://www.w3.org/TR/2000/REC-xml-20001006
 * Namespaces, http://www.w3.org/TR/1999/REC-xml-names-19990114
 *
 * This implementation strongly influenced by the XPath Data Model 
 * (see section 5 of the XPath spec).
 */

/* A qualified name */
typedef struct qname {
	wstr_t	ns;		/* (loaned) URI of namespace (sometimes the prefix) */
	wstr_t	local;		/* (interned) local part of name */
} qname_t;

/* Types of XML nodes */
enum xmltype {
	ROOT, ELEMENT, TEXT, ATTR, NS, PROCINST, COMMENT
};

/* An XML node representation */
struct xml {
	enum xmltype	type;		/* node type */
	void *		user;		/* user data */
	void		(*userfree)(void *, struct xml *);

	xml_t		parent;		/* linkage */
	xml_t		child;
	xml_t		sibling;
	xml_t		*lastchildp;

	xml_t		ns;		/* namespace scope tree or NULL */

	qname_t		name;
	wstr_t		id;		/* unique identifier */
	union {
		struct xmldocinfo *doc;
		void *	ptr;
		wstr_t	args;
		struct {
			xml_t	first_attr;
			struct xhash *attr;
		} element;
		struct {
			wstr_t	value;
			/* VECTOR_DECL(wstr_t) names; */
		} attr;
		struct xmlns *ns;
	} aux;

	int		order;		/* document-order number */
};

/*
	all:
		user:		reserved pointer for user data
		parent:		parent pointer
		child:		first child pointer
		sibling:	sibling pointer
		lastchildp:	addr of own .child or last child's .sibling
		ns:		namespace scope chain
		id:		identifier (intern'd or NULL)
		name:		(see below, default NULL:NULL)
		aux:		(see below, default NULL)
	ROOT:
		aux.doc:	document information
	ELEMENT
		name:			element name (intern'd)
		aux.element.first_attr:	attribute nodes (linked by sibling)
		aux.element.attr:	hash table of qname_t -> xml_t
	TEXT
		aux.args:	expanded text
	ATTR
		name:		attribute name (intern'd)
		attr.value:	normalised value (not interned)
		attr.names:	vector of interned names if value not PCDATA
	NS
		ns:		namespace nodes of enclosing scope
		name.local:	namespace's name in scope
		aux.ns:		xmlns tree, mirroring scope tree
	PROCINST
		name.local:	procinst target name (intern'd)
		aux.args:	(wstr_t) procinst arguments
	COMMENT
		aux.args:	comment text
*/

/* The document-wide information structure */
struct xmldocinfo {
	struct xhash	*entity;	/* &ref; (struct entity *) */
	struct xhash	*pentity;	/* %ref; (struct entity *) */
	struct xhash	*dtd;		/* <tag> (struct dtdelement *) */
	struct xhash	*notation;	/* ntn (struct externalid *) */
	struct xhash	*id;		/* ID (wstr_t -> shared xml_t) */
	xml_t		namespaces;	/* list of namespaces */
	wstr_t		 doctype;	/* doctype (interned) or null */
	wstr_t		 pcdata;	/* interned #PCDATA symbol */
	wstr_t		 xmlns;		/* interned 'xmlns' symbol */
	wstr_t		 empty;		/* interned '' string */
	int		 nextorder;	/* for document-order */
};

/* An attribute specification for an entity */
struct dtdattr {
	enum AttType { 
		CDATA,
		ID, IDREF, IDREFS, ENTITY, ENTITIES, NMTOKEN, NMTOKENS,
		NOTATION, ENUMERATION
	} type;
	VECTOR_DECL(wstr_t) enums;
	enum DefType { REQUIRED, IMPLIED, FIXED } deftype;
	wstr_t defattr;
};

struct dtdelement {
	struct xfsm *fsm;
	struct xhash *attrs;
	int ignorespace;	/* ignore CDATA */
	wstr_t	idattr;		/* name of the attribute with an ID type */
};

struct externalid {
	wstr_t	systemid;
	wstr_t	publicid;
};

#define isSpace(c) ((c) == 0x20 || (c) == 0x9 || (c) == 0xd || (c) == 0xa)

/* Hide library internal functions with a dxml prefix */
#define	puts_wstr	dxml_puts_wstr
#define	qname_str	dxml_qname_str

int	puts_wstr(wstr_t, int (*)(const char *, void *), void *, int);
const char *qname_str(struct xmlns *ns, qname_t *);
