/* $Id: xpath.c,v 1.17 2002/07/21 04:23:11 d Exp $ */
/* David Leonard, 2002. Public domain. */

/*
 * XPath
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include "xml.h"
#include "xpath.h"
#include "xmlns.h"

#include "vector.h"
#include "wstr.h"
#include "entity.h"
#include "xmlP.h"

#include "utf8.h"
#include "xhash.h"
#include "intern.h"
#include "xmlnsP.h"

#include "xpathP.h"

static const double NaN = 0.0/0.0;

static struct xhash *function_hash;
static struct xhash *operator_hash;
static struct xhash *axis_hash;
static struct xhash *nodetype_hash;

const char *xpath_last_error;

extern int xml_debug;

#define more(in)	((in)->next != -1 && (in)->error == NULL)

#define	DPRINTF(fmt, args...)  do { if (xml_debug) \
	fprintf(stderr, fmt , ##args); \
} while (0)
#define	DEBUG_FN  do { if (xml_debug) \
	fprintf(stderr, "%s . %s\n", __FUNCTION__, tokname(in->next)); \
} while (0)

/* Prototypes */
static const char *f_last(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_position(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_count(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_id(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_local_name(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_namespace_uri(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_name(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_string(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_concat(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_starts_with(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_contains(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_substring_before(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_substring_after(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_substring(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_string_length(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_normalize_space(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_translate(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_boolean(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_not(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_true(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_false(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_lang(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_number(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_sum(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_floor(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_ceiling(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *f_round(struct context *context, struct value *argv, int argc, struct value *ret);
static const char *xp_eval(xpath_t xp, xml_t node, struct xmlns *ns, struct value *v);
static char *value_to_utf8(struct value *v, int *);
static void var_bind_utf8(xpath_t xp, const char *nm, struct value *v);
static void var_bind(xpath_t xp, const wchar_t *uri, size_t urisz, const wchar_t *nm, size_t nmsz, struct value *v);
static struct input *input_new(struct entity *e);
static void input_free(struct input *in);
static void error(struct input *in, const char *fmt, ...);
static const char *tokname(int token);
static void expect(struct input *in, int token);
static wchar_t nextch(struct input *in);
static int next(struct input *in);
static struct expr *expr_new(enum exprtype type);
static struct expr *expr_newbin(enum exprtype type, struct expr *e1, struct expr *e2);
static void expr_free(struct expr *e);
static struct path *path_new(enum axis axis);
static struct path *path_new_dslash(void);
static void path_free(struct path *p);
static struct value *value_alloc(void);
static void value_free(struct value *v);
static void value_copy(struct value *dst, struct value *src);
static void value_addnode(struct value *v, xml_t n);
static int xmlcmp(const void *a, const void *b);
static void value_sort_nodes(struct value *v);
static void sort_nodes(nodeset_t *ns);
static void value_addnodes(struct value *v1, struct value *v2);
static void value_append_wstr(struct value *v, wstr_t w);
static xml_t first_node(struct value *v);
static void toboolean(struct value *v);
static int asboolean(struct value *v);
static void tostring(struct value *v);
static void asstring(struct value *v, struct value *res);
static void remove_exponent(char *buf);
static void tonumber(struct value *v);
static double asnumber(struct value *v);
static void append_string_value(struct value *v, xml_t n);
static void append_text_nodes(struct value *v, xml_t n, int sib);
static qname_t expanded_name(xml_t n);
static struct path *LocationPath(struct input *in);
static struct path *AbsoluteLocationPath(struct input *in);
static struct path *RelativeLocationPath(struct input *in);
static int isStep(struct input *in);
static struct path *Step(struct input *in);
static enum axis AxisSpecifier(struct input *in);
static void NodeTest(struct input *in, struct path *p);
static struct expr *Predicate(struct input *in);
static struct expr *PredicateExpr(struct input *in);
static struct path *AbbreviatedAbsoluteLocationPath(struct input *in);
static struct path *AbbreviatedStep(struct input *in);
static enum axis AbbreviatedAxisSpecifier(struct input *in);
static struct expr *Expr(struct input *in);
static struct expr *PrimaryExpr(struct input *in);
static struct expr *FunctionCall(struct input *in);
static struct expr *Argument(struct input *in);
static struct expr *UnionExpr(struct input *in);
static struct expr *PathExpr(struct input *in);
static struct expr *FilterExpr(struct input *in);
static struct expr *OrExpr(struct input *in);
static struct expr *AndExpr(struct input *in);
static struct expr *EqualityExpr(struct input *in);
static struct expr *RelationalExpr(struct input *in);
static struct expr *AdditiveExpr(struct input *in);
static struct expr *MultiplicativeExpr(struct input *in);
static struct expr *UnaryExpr(struct input *in);
static const char *expr_eval(struct context *c, struct expr *e, struct value *result);
static int value_compare(struct value *v1, struct value *v2, enum exprtype cmp);
static const char *path_eval(struct context *c, struct path *p, nodeset_t *src, nodeset_t *res);
static const char *path_addnode(struct context *c, struct path *p, xml_t n, nodeset_t *result);
static const char *path_generate(struct context *c, struct path *p, enum axis ax, xml_t n, nodeset_t *result);
static const char *filter(struct context *c, nodeset_t *src, struct expr *e, nodeset_t *result);
static int expr_put(struct expr *e, int (*puts)(const char *, void *), void *closure);
static int path_put(struct path *p, int (*puts)(const char *, void *), void *closure);
static int value_put(struct value *v, int (*puts)(const char *, void *), void *closure);

/*------------------------------------------------------------
 * Functions
 *
 * notes
 * - return an error string on error
 * - functions must initialise and fill *ret, when returning NULL
 * - the function's caller will immediately fini the argument valuess
 */

struct function functab[] = {
	{ "last",	f_last,			""	},
	{ "position",	f_position,		""	},
	{ "count",	f_count,		"n"	},
	{ "id",		f_id,			"o"	},
	{ "local-name",	f_local_name,		"|n"	},
	{ "namespace-uri", f_namespace_uri,	"|n"	},
	{ "name",	f_name,			"|n"	},
	{ "string",	f_string,		"|o"	},
	{ "concat",	f_concat,		"sss*"	},
	{ "starts-with",f_starts_with,		"ss"	},
	{ "contains",	f_contains,		"ss"	},
	{ "substring-before",f_substring_before,"ss"	},
	{ "substring-after",f_substring_after,	"ss"	},
	{ "substring",	f_substring,		"sd|d"	},
	{ "string-length",f_string_length,	"|s"	},
	{ "normalize-space",f_normalize_space,	"|s"	},
	{ "translate",	f_translate,		"sss"	},
	{ "boolean",	f_boolean,		"o"	},
	{ "not",	f_not,			"b"	},
	{ "true",	f_true,			""	},
	{ "false",	f_false,		""	},
	{ "lang",	f_lang,			"s"	},
	{ "number",	f_number,		"|o"	},
	{ "sum",	f_sum,			"n"	},
	{ "floor",	f_floor,		"d"	},
	{ "ceiling",	f_ceiling,		"d"	},
	{ "round",	f_round,		"d"	}
};

/* Check if the given number of args is compatible with the function */
static char *
function_arity_check(f, argc)
	struct function *f;
	int argc;
{
	const char *a;
	int isopt = 0;
	static char errbuf[80];

	for (a = f->signature; *a; a++)
		if (*a == '|') 
			isopt = 1;
		else if (*a == '*' && argc >= -1) /* previous+ is optional */
			argc = 0;
		else if (!isopt || argc > 0)
			argc--;
	if (argc == 0)
		return NULL;	/* compatible arity */
	snprintf(errbuf, sizeof errbuf, 
	    "too %s arguments to %s()",
	    argc > 0 ? "many" : "few",
	    f->utf8_name);
	return errbuf;
}

/* Check types of the values; assumes arity is correct */
static char *
function_type_check(f, argv, argc)
	struct function *f;
	struct value *argv;
	int argc;
{
	const char *a;
	enum valuetype exptype;
	static char errbuf[80];
	int pos;

#define TYPENAME(t) (					\
	    (t) == STRING ? "string" :			\
	    (t) == NODESET ? "node-set" :		\
	    (t) == NUMBER ? "number" :			\
	    (t) == BOOLEAN ? "boolean" : "?")

	pos = 0;
	for (a = f->signature; pos < argc && *a; a++) {
		if (*a == '*')
			a--;
		if (*a == '|')
			continue;
		if (*a != 'o') {
			switch (*a) {
			case 's':	exptype = STRING; break;
			case 'n':	exptype = NODESET; break;
			case 'i':	exptype = NUMBER; break;
			case 'b':	exptype = BOOLEAN; break;
			default:	
			   return "function_type_check: internal error";
			}
			if (exptype != argv[pos].type) {
				snprintf(errbuf, sizeof errbuf,
				    "expected %s but got %s for argument %d of %s()",
				    TYPENAME(exptype),
				    TYPENAME(argv[pos].type),
				    pos + 1,
				    f->utf8_name);
				return errbuf;
			}
		}
		pos++;
	}
	return NULL;
}

static const char *
f_last(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	ret->type = NUMBER;
	ret->v_number = context->size;
	return NULL;
}

static const char *
f_position(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	ret->type = NUMBER;
	ret->v_number = context->position;
	return NULL;
}

static const char *
f_count(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	ret->type = NUMBER;
	ret->v_number = VECTOR_LEN(argv[0].v_nodeset);
	return NULL;
}

static const char *
f_id(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int i;
	wstr_t w;
	xml_t n;

	VALUE_INIT_NODESET(*ret);

	if (argv[0].type == NODESET) {
		struct value s, r;
		const char *err;

		/* the union of the result of applying id to the
		 * string-value of each of the nodes in the 
		 * argument node-set */
		for (i = 0; i < VECTOR_LEN(argv[0].v_nodeset); i++) {
			/* convert each argument node to a string s */
			VALUE_INIT_STRING(s);
			append_string_value(&s, VECTOR_AT(argv[0].v_nodeset, i));
			/* id(s) */
			err = f_id(context, &s, 1, &r);
			VALUE_FINI(s);
			if (err) {
				VALUE_FINI(*ret);
				return err;
			}
			value_addnodes(ret, &r);
			VALUE_FINI(r);
		}
		value_sort_nodes(ret);
		return NULL;
	}

	/*
	 * ...converted to a string ... split into a whitespace-separated
	 * list of tokens, result is a node-set containing the elements
	 * ... that have a unique ID equal to any of the tokens in the list
	 */
	tostring(&argv[0]);
	/* skip leading whitespace */
	for (i = 0; i < VECTOR_LEN(argv[0].v_string); i++) 
		if (!isSpace(VECTOR_AT(argv[0].v_string, i)))
			break;
	while (i < VECTOR_LEN(argv[0].v_string)) {
		/* pick out non-whitespace token */
		int start = i;
		for (; i < VECTOR_LEN(argv[0].v_string); i++)
			if (isSpace(VECTOR_AT(argv[0].v_string, i)))
				break;
		/*
		 * To avoid wastefully interning strings, we check to see
		 * if the token name already has an interned representation.
		 */
		w = (wstr_t)xhash_lookup(intern_hash,
		    &VECTOR_AT(argv[0].v_string, start),
		    (i - start) * sizeof (wchar_t));
		/* see if there's a node with this name */
		if (w) {
		    n = (xml_t)xhash_lookup(context->doc->id, &w, sizeof w);
		    if (n)
			value_addnode(ret, n);
		}
		value_sort_nodes(ret);
		/* skip whitespace after token*/
		for (; i < VECTOR_LEN(argv[0].v_string); i++) 
			if (!isSpace(VECTOR_AT(argv[0].v_string, i)))
				break;
	}
	return NULL;
}

static const char *
f_local_name(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	xml_t n;
	qname_t q;

	if (argc == 0)
		n = context->node;
	else
		n = first_node(&argv[0]);

	VALUE_INIT_STRING(*ret);
	if (n) {
		q = expanded_name(n);
		if (q.local)
			value_append_wstr(ret, q.local);
	}
	return NULL;
}

static const char *
f_namespace_uri(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	xml_t n;
	qname_t q;

	if (argc == 0)
		n = context->node;
	else
		n = first_node(&argv[0]);

	VALUE_INIT_STRING(*ret);
	if (n) {
		q = expanded_name(n);
		if (q.ns)
			value_append_wstr(ret, q.ns);
	}
	return NULL;
}

static const char *
f_name(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	xml_t n;

	if (argc == 0)
		n = context->node;
	else
		n = first_node(&argv[0]);

	VALUE_INIT_STRING(*ret);
	if (n && (n->type == ELEMENT || n->type == ATTR)) {
		struct xmlns *ns = n->ns->aux.ns;
		wstr_t prefix;
		prefix = xmlns_finduri(ns, n->name.ns);
		if (prefix) {
			value_append_wstr(ret, prefix);
			VECTOR_APPEND(ret->v_string, ':');
		}
		value_append_wstr(ret, n->name.local);
	}
	return NULL;
}

static const char *
f_string(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	if (argc == 0) {
		VALUE_INIT_STRING(*ret);
		append_string_value(ret, context->node);
		return NULL;
	}
	value_copy(ret, &argv[0]);
	tostring(ret);
	return NULL;
}

static const char *
f_concat(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int i;

	VALUE_INIT_STRING(*ret);
	for (i = 0; i < argc; i++)
		VECTOR_CONCAT(ret->v_string, argv[i].v_string);
	return NULL;
}

static const char *
f_starts_with(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int truth;

	truth = VECTOR_LEN(argv[1].v_string) == 0 ||
		(VECTOR_LEN(argv[0].v_string) >= VECTOR_LEN(argv[1].v_string)
		 && memcmp(VECTOR_PTR(argv[0].v_string), 
		          VECTOR_PTR(argv[1].v_string),
		          VECTOR_SIZE(argv[1].v_string)) == 0);
	VALUE_INIT_BOOLEAN(*ret, truth);
	return NULL;
}

static const char *
f_contains(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int i, truth;

	if (VECTOR_LEN(argv[1].v_string) == 0)
		truth = 1;
	else
		for (truth = 0, i = 0; 
		      i <= VECTOR_LEN(argv[0].v_string) - 
			   VECTOR_LEN(argv[1].v_string); i++)
		    if (memcmp(&VECTOR_AT(argv[0].v_string, i), 
			      VECTOR_PTR(argv[1].v_string),
			      VECTOR_SIZE(argv[1].v_string)) == 0)
		    {
			truth = 1;
			break;
		    }

	VALUE_INIT_BOOLEAN(*ret, truth);
	return NULL;
}

static const char *
f_substring_before(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int i, j;

	VALUE_INIT_STRING(*ret);
	for (i = 0; i < VECTOR_LEN(argv[0].v_string) -
			VECTOR_LEN(argv[1].v_string); i++)
		if (memcmp(&VECTOR_AT(argv[0].v_string, i),
			VECTOR_PTR(argv[1].v_string),
			VECTOR_SIZE(argv[1].v_string)) == 0)
			break;
	for (j = 0; j < i; j++)
		VECTOR_APPEND(ret->v_string, VECTOR_AT(argv[0].v_string, j));
	return NULL;
}

static const char *
f_substring_after(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int i, j;

	VALUE_INIT_STRING(*ret);
	if (VECTOR_LEN(argv[1].v_string) == 0)
		return NULL;
	for (i = 0; i < VECTOR_LEN(argv[0].v_string) -
			VECTOR_LEN(argv[1].v_string); i++)
		if (memcmp(&VECTOR_AT(argv[0].v_string, i),
			VECTOR_PTR(argv[1].v_string),
			VECTOR_SIZE(argv[1].v_string)) == 0)
			break;
	for (j = i + VECTOR_LEN(argv[1].v_string);
	     j < VECTOR_LEN(argv[0].v_string); j++)
		VECTOR_APPEND(ret->v_string, VECTOR_AT(argv[0].v_string, j));
	return NULL;
}

static const char *
f_substring(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int i, n2, n3;

	n2 = rint(argv[1].v_number);
	n3 = argc == 3 ? rint(argv[2].v_number) : VECTOR_LEN(argv[0].v_string);

	VALUE_INIT_STRING(*ret);
	for (i = n2 - 1; i < n3; i++)
		VECTOR_APPEND(ret->v_string, VECTOR_AT(argv[0].v_string, i));
	return NULL;
}

static const char *
f_string_length(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int len;
	struct value v;

	if (argc == 1) {
		VALUE_INIT_NUMBER(*ret, VECTOR_LEN(argv[0].v_string));
		return NULL;
	}
	VALUE_INIT_STRING(v);
	append_string_value(&v, context->node);
	len = VECTOR_LEN(v.v_string);
	VALUE_FINI(v);
	VALUE_INIT_NUMBER(*ret, len);
	return NULL;
}

static const char *
f_normalize_space(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	struct value lv, *v;
	int i;
	int waswhite;

	if (argc == 1) {
		VALUE_INIT_BOOLEAN(lv, 0);
		v = &argv[0];
	} else {
		VALUE_INIT_STRING(lv);
		append_string_value(&lv, context->node);
		v = &lv;
	}
	VALUE_INIT_STRING(*ret);
	waswhite = -1;
	for (i = 0; i < VECTOR_LEN(v->v_string); i++) {
		wchar_t c = VECTOR_AT(v->v_string, i);
		if (isSpace(c)) {
			if (waswhite == 0)
				waswhite = 1;
		} else {
			if (waswhite == 1)
				VECTOR_APPEND(ret->v_string, ' ');
			VECTOR_APPEND(ret->v_string, c);
			waswhite = 0;
		}
	}
	VALUE_FINI(lv);
	return NULL;
}

static const char *
f_translate(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	int i, j;
	wchar_t c;

	VALUE_INIT_STRING(*ret);
	for (i = 0; i < VECTOR_LEN(argv[0].v_string); i++) {
		c = VECTOR_AT(argv[0].v_string, i);
		for (j = 0; j < VECTOR_LEN(argv[1].v_string); j++)
		    if (VECTOR_AT(argv[1].v_string, j) == c) {
			if (j < VECTOR_LEN(argv[2].v_string))
			    VECTOR_APPEND(ret->v_string,
				VECTOR_AT(argv[2].v_string, j));
			break;
		    }
		if (j >= VECTOR_LEN(argv[1].v_string))
			VECTOR_APPEND(ret->v_string, c);
	}
	return NULL;
}

static const char *
f_boolean(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	VALUE_INIT_BOOLEAN(*ret, asboolean(&argv[0]));
	return NULL;
}

static const char *
f_not(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	VALUE_INIT_BOOLEAN(*ret, !argv[0].v_boolean);
	return NULL;
}

static const char *
f_true(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	VALUE_INIT_BOOLEAN(*ret, 1);
	return NULL;
}

static const char *
f_false(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	VALUE_INIT_BOOLEAN(*ret, 0);
	return NULL;
}

static const char *
f_lang(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	return "not implemented";
}

static const char *
f_number(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	if (argc == 0) {
		struct value v;
		VALUE_INIT_NODESET(v);
		VECTOR_APPEND(v.v_nodeset, context->node);
		VALUE_INIT_NUMBER(*ret, asnumber(&v));
		VALUE_FINI(v);
		return NULL;
	}
	VALUE_INIT_NUMBER(*ret, asnumber(&argv[0]));
	return NULL;
}

static const char *
f_sum(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	double sum = 0.0;
	struct value v;
	int i;

	VALUE_INIT_NODESET(v);
	VECTOR_APPEND(v.v_nodeset, NULL);
	for (i = 0; i < VECTOR_LEN(argv[0].v_nodeset); i++) {
		VECTOR_AT(v.v_nodeset, 0) = VECTOR_AT(argv[0].v_nodeset, i);
		sum += asnumber(&v);
	}
	VALUE_FINI(v);
	VALUE_INIT_NUMBER(*ret, sum);
	return NULL;
}

static const char *
f_floor(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	VALUE_INIT_NUMBER(*ret, floor(argv[0].v_number));
	return NULL;
}

static const char *
f_ceiling(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	VALUE_INIT_NUMBER(*ret, ceil(argv[0].v_number));
	return NULL;
}

static const char *
f_round(context, argv, argc, ret)
	struct context *context;
	struct value *argv;
	int argc;
	struct value *ret;
{
	VALUE_INIT_NUMBER(*ret, rint(argv[0].v_number));
	return NULL;
}


/*------------------------------------------------------------
 * Exported API
 */

/* create a new xpath expression */
xpath_t
xpath_new_utf8(expr)
	const char *expr;
{
	return xpath_newe(entity_utf8mem(expr, expr + strlen(expr), 1));
}

xpath_t
xpath_new(buf, bufsz)
	const wchar_t *buf;
	size_t bufsz;
{
	return xpath_newe(entity_mem(buf, buf + (bufsz / sizeof (wchar_t)), 1));
}

xpath_t
xpath_newe(e)
	struct entity *e;
{
	xpath_t xp;
	struct input *in;
	const char *ret;

	xp = (struct xpath *)malloc(sizeof (struct xpath));
	if (xp) {
		xp->variables = xhash_new(NULL, (void (*)(void *))value_free);
		in = input_new(e);
		xp->expr = Expr(in);
		if (in->error == NULL && in->next != -1)
			error(in, "unexpected characters");
		ret = in->error;
		if (in->error) {
			xpath_free(xp);
			xp = NULL;
		}
		input_free(in);
	} else {
		ret = "malloc";
		ENTITY_DECREF(e);
	}
	xpath_last_error = ret;
	return xp;
}

/* Discard an xpath expression */
void
xpath_free(xp)
	xpath_t xp;
{
	xhash_free(xp->variables);
	expr_free(xp->expr);
	free(xp);
}

/* evaluate an expression with the user-provided node */
static const char *
xp_eval(xp, node, ns, v)
	xpath_t xp;
	xml_t node;
	struct xmlns *ns;
	struct value *v;
{
	struct context c;
	xml_t n;

	if (!node)
		return "null context node";
	for (n = node; n; n = n->parent)
		if (n->type == ROOT)
			break;
	if (!n)
		return "no root node";
	c.doc = n->aux.doc;
	c.node = node;
	c.position = 1;
	c.size = 1;
	c.variables = xp->variables;
	c.reverse = 0;
	c.ns = ns;

	return expr_eval(&c, xp->expr, v);
}

/* evaluate an expression that results in a node-set - caller must free */
xml_t *
xpath_select(xp, node, ns)
	xpath_t xp;
	xml_t node;
	struct xmlns *ns;
{
	struct value v;
	const char *m;

	m = xp_eval(xp, node, ns, &v);
	if (m) {
		xpath_last_error = m;
		return NULL;
	}
	if (v.type != NODESET) {
		xpath_last_error = "did not evaulate to a node-set";
		VALUE_FINI(v);
		return NULL;
	}
	/* Add a NULL to the end of the node vector */
	VECTOR_APPEND(v.v_nodeset, NULL);
	/* NOTE: caller to free the returned pointer */
	return VECTOR_PTR(v.v_nodeset);
}

/* Return an alloc'd UTF8 string */
static char *
value_to_utf8(v, lenp)
	struct value *v;
	int *lenp;
{
	char *res, *p;
	int sz, i;

	/* assert(v->type == STRING); */
	for (sz = 0, i = 0; i < VECTOR_LEN(v->v_string); i++)
		sz += utf8_size_put(VECTOR_AT(v->v_string, i));
	res = (char *)malloc(sz + 1);
	for (p = res, i = 0; i < VECTOR_LEN(v->v_string); i++)
		utf8_put(VECTOR_AT(v->v_string, i), &p);
	*p = '\0';
	if (lenp)
		*lenp = p - res;
	return res;

}

/* evaluate expression - unknown return type */
int
xpath_eval(xp, node, ns, uv)
	xpath_t xp;
	xml_t node;
	struct xmlns *ns;
	union xpath_value *uv;
{
	struct value v;
	const char *m;

	m = xp_eval(xp, node, ns, &v);
	if (m) {
		xpath_last_error = m;
		return XPATH_VALUE_ERROR;
	}
	switch (v.type) {
	case NODESET:
		value_sort_nodes(&v);
		VECTOR_APPEND(v.v_nodeset, NULL);
		uv->x_nodeset = VECTOR_PTR(v.v_nodeset);
		return XPATH_VALUE_NODESET;
	case BOOLEAN:
		uv->x_boolean = v.v_boolean;
		VALUE_FINI(v);
		return XPATH_VALUE_BOOLEAN;
	case NUMBER:
		uv->x_number = v.v_number;
		VALUE_FINI(v);
		return XPATH_VALUE_NUMBER;
	case STRING:
		uv->x_string.data = value_to_utf8(&v,
			&uv->x_string.size);
		VALUE_FINI(v);
		return XPATH_VALUE_STRING_UTF8;
	}
	VALUE_FINI(v);
	xpath_last_error = "xpath_eval(): internal error";
	return XPATH_VALUE_ERROR;
}

/* evaluate expression; return a UTF-8 string - caller must free */
char *
xpath_eval_utf8string(xp, node, ns)
	xpath_t xp;
	xml_t node;
	struct xmlns *ns;
{
	struct value v;
	const char *m;
	char *res;

	m = xp_eval(xp, node, ns, &v);
	if (m) {
		xpath_last_error = m;
		return NULL;
	}
	if (v.type != STRING) {
		xpath_last_error = "did not evaluate to a string";
		VALUE_FINI(v);
		return NULL;
	}
	res = value_to_utf8(&v, NULL);
	VALUE_FINI(v);
	return res;
}

/* evaluate expression; return a boolean */
int
xpath_eval_boolean(xp, node, ns)
	xpath_t xp;
	xml_t node;
	struct xmlns *ns;
{
	struct value v;
	const char *m;
	int res;

	m = xp_eval(xp, node, ns, &v);
	if (m) {
		xpath_last_error = m;
		return -1;
	}
	if (v.type != BOOLEAN) {
		xpath_last_error = "did not evaluate to a boolean";
		VALUE_FINI(v);
		return -1;
	}
	res = v.v_boolean;
	VALUE_FINI(v);
	return res;
}

/* evaluate expression; return a double */
double
xpath_eval_number(xp, node, ns)
	xpath_t xp;
	xml_t node;
	struct xmlns *ns;
{
	struct value v;
	const char *m;
	double res;

	m = xp_eval(xp, node, ns, &v);
	if (m) {
		xpath_last_error = m;
		return NaN;
	}
	if (v.type != NUMBER) {
		xpath_last_error = "did not evaluate to a number";
		VALUE_FINI(v);
		return NaN;
	}
	res = v.v_number;
	VALUE_FINI(v);
	xpath_last_error = NULL;
	return res;
}

static void
var_bind_utf8(xp, nm, v)
	xpath_t xp;
	const char *nm;
	struct value *v;
{
	struct qname q;
	q.ns = NULL;
	q.local = intern_utf8(nm);
	xhash_store(xp->variables, &q, sizeof q, v);
}

static void
var_bind(xp, uri, urisz, nm, nmsz, v)
	xpath_t xp;
	const wchar_t *uri;
	size_t urisz;
	const wchar_t *nm;
	size_t nmsz;
	struct value *v;
{
	struct qname q;
	q.ns = intern(uri, urisz);
	q.local = intern(nm, nmsz);
	xhash_store(xp->variables, &q, sizeof q, v);
}

/* Unbind a variable (reset to NULL) */
void
xpath_unbind_utf8(xp, nm)
	xpath_t xp;
	const char *nm;
{
	var_bind_utf8(xp, nm, NULL);
}

void
xpath_unbind(xp, uri, urisz, nm, nmsz)
	xpath_t xp;
	const wchar_t *uri;
	size_t urisz;
	const wchar_t *nm;
	size_t nmsz;
{
	var_bind(xp, uri, urisz, nm, nmsz, NULL);
}

/* Bind a string to variable $s */
void
xpath_bind_utf8_string(xp, nm, s)
	xpath_t xp;
	const char *nm, *s;		/* s is UTF-8 encoded */
{
	struct value *v;

	v = value_alloc();
	VALUE_INIT_STRING(*v);
	while (*s) {
		wchar_t c = utf8_get(&s);
		VECTOR_APPEND(v->v_string, c);
	}
	var_bind_utf8(xp, nm, v);
}

void
xpath_bind_string(xp, uri, urisz, nm, nmsz, str, strsz)
	xpath_t xp;
	const wchar_t *uri;
	size_t urisz;
	const wchar_t *nm;
	size_t nmsz;
	const wchar_t *str;
	size_t strsz;
{
	struct value *v;

	v = value_alloc();
	VALUE_INIT_STRING(*v);
	while (strsz > 0) {
		VECTOR_APPEND(v->v_string, *str);
		str++;
		strsz -= sizeof (wchar_t);
	}
	var_bind(xp, uri, urisz, nm, nmsz, v);
}

/* Bind a number to the variable $nm */
void
xpath_bind_utf8_number(xp, nm, d)
	xpath_t xp;
	const char *nm;
	double d;
{
	struct value *v;

	v = value_alloc();
	VALUE_INIT_NUMBER(*v, d);
	var_bind_utf8(xp, nm, v);
}

void
xpath_bind_number(xp, uri, urisz, nm, nmsz, d)
	xpath_t xp;
	const wchar_t *uri;
	size_t urisz;
	const wchar_t *nm;
	size_t nmsz;
	double d;
{
	struct value *v;

	v = value_alloc();
	VALUE_INIT_NUMBER(*v, d);
	var_bind(xp, uri, urisz, nm, nmsz, v);
}

/* Bind a set of nodes to the variable $nm */
void
xpath_bind_utf8_nodeset(xp, nm, nodes, len)
	xpath_t xp;
	const char *nm;
	xml_t *nodes;
	int len;
{
	struct value *v;
	int i;

	v = value_alloc();
	VALUE_INIT_NODESET(*v);
	if (len < 0)
		for (i = 0; nodes[i]; i++) 
			value_addnode(v, nodes[i]);
	else
		for (i = 0; i < len; i++) 
			value_addnode(v, nodes[i]);
	value_sort_nodes(v);
	var_bind_utf8(xp, nm, v);
}

void
xpath_bind_nodeset(xp, uri, urisz, nm, nmsz, nodes, len)
	xpath_t xp;
	const wchar_t *uri;
	size_t urisz;
	const wchar_t *nm;
	size_t nmsz;
	xml_t *nodes;
	int len;
{
	struct value *v;
	int i;

	v = value_alloc();
	VALUE_INIT_NODESET(*v);
	if (len < 0)
		for (i = 0; nodes[i]; i++) 
			value_addnode(v, nodes[i]);
	else
		for (i = 0; i < len; i++) 
			value_addnode(v, nodes[i]);
	value_sort_nodes(v);
	var_bind(xp, uri, urisz, nm, nmsz, v);
}


/* Bind a truth value to the variable $nm */
void
xpath_bind_utf8_boolean(xp, nm, truth)
	xpath_t xp;
	const char *nm;
	int truth;
{
	struct value *v;

	v = value_alloc();
	VALUE_INIT_BOOLEAN(*v, truth);
	var_bind_utf8(xp, nm, v);
}

void
xpath_bind_boolean(xp, uri, urisz, nm, nmsz, truth)
	xpath_t xp;
	const wchar_t *uri;
	size_t urisz;
	const wchar_t *nm;
	size_t nmsz;
	int truth;
{
	struct value *v;

	v = value_alloc();
	VALUE_INIT_BOOLEAN(*v, truth);
	var_bind(xp, uri, urisz, nm, nmsz,v);
}

/*------------------------------------------------------------
 * Lexical analyser
 */


/* Initialise the input state */
static struct input *
input_new(e)
	struct entity *e;
{
	struct input *in = (struct input *)malloc(sizeof (struct input));
	int i;

	in->entity = e;
	in->error = NULL;
	in->next_raw = wstr_new();

	/* Internalise some keywords so that lexical analysis is sped up */

#define DEF(h, str, v)						\
	do {							\
		wstr_t w = intern_utf8(str);			\
		xhash_store(h, &w, sizeof w, (void *)(v));	\
	} while (0)

	if (!operator_hash) {
		operator_hash = xhash_new(NULL, NULL);
		DEF(operator_hash, "and", AND);
		DEF(operator_hash, "or",  OR);
		DEF(operator_hash, "mod", MOD);
		DEF(operator_hash, "div", DIV);
	}

	if (!axis_hash) {
		axis_hash = xhash_new(NULL, NULL);
		DEF(axis_hash, "ancestor",		ANCESTOR_AXIS);
		DEF(axis_hash, "ancestor-or-self",	ANCESTOR_OR_SELF_AXIS);
		DEF(axis_hash, "attribute",		ATTRIBUTE_AXIS);
		DEF(axis_hash, "child",			CHILD_AXIS);
		DEF(axis_hash, "descendant",		DESCENDANT_AXIS);
		DEF(axis_hash, "descendant-or-self",	DESCENDANT_OR_SELF_AXIS);
		DEF(axis_hash, "following",		FOLLOWING_AXIS);
		DEF(axis_hash, "following-sibling",	FOLLOWING_SIBLING_AXIS);
		DEF(axis_hash, "namespace",		NAMESPACE_AXIS);
		DEF(axis_hash, "parent",		PARENT_AXIS);
		DEF(axis_hash, "preceding",		PRECEDING_AXIS);
		DEF(axis_hash, "preceding-sibling",	PRECEDING_SIBLING_AXIS);
		DEF(axis_hash, "self",			SELF_AXIS);
	}

	if (!nodetype_hash) {
		nodetype_hash = xhash_new(NULL, NULL);
		DEF(nodetype_hash, "comment",	COMMENT_TYPE);
		DEF(nodetype_hash, "text",	TEXT_TYPE);
		DEF(nodetype_hash, "processing-instruction", PI_TYPE);
		DEF(nodetype_hash, "node",	NODE_TYPE);
	}

	if (!function_hash) {
		function_hash = xhash_new(NULL, NULL);
		for (i = 0; i < sizeof functab / sizeof functab[0]; i++) {
			wstr_t w = intern_utf8(functab[i].utf8_name);
			xhash_store(function_hash, &w, sizeof w, &functab[i]);
		}
	}

	in->nextc = 0;
	in->next = -2;
	in->next_literal = NULL;

	ENTITY_REWIND(in->entity);
	nextch(in);
	next(in);

#undef DEF
	return in;
}

static void
input_free(in)
	struct input *in;
{
	if (in->next_literal)
		wstr_free(in->next_literal);
	ENTITY_DECREF(in->entity);
	wstr_free(in->next_raw);
	free(in);
}

/* Put the lexical analyzer into an error state */
static void
error(struct input *in, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	if (!in->error) {
		static char errbuf[1024];

		vsnprintf(errbuf, sizeof errbuf, fmt, ap);
		if (VECTOR_LEN(*in->next_raw) == 0) {
			strlcat(errbuf, " at end of input", sizeof errbuf);
		} else {
			strlcat(errbuf, " at '", sizeof errbuf);
			strlcat(errbuf, wstr_str(in->next_raw), sizeof errbuf);
			strlcat(errbuf, "'", sizeof errbuf);
		}
		in->error = errbuf;
		if (xml_debug)
			fprintf(stderr, "error(): %s\n", errbuf);
	}
	va_end(ap);
}

/* Return a string fom of a token */
static const char *
tokname(token)
	int token;
{
	static char buf[7];
	static char hex[] = "0123456789abcdef";

	switch (token) {
	case -1:	return "{eof}";
	case DCOLON:	return "::";
	case MULTIPLY:	return "{multiply}";
	case LITERAL:	return "{literal}";
	case NUMERIC:	return "{numeric}";
	case AND:	return "and";
	case OR:	return "or";
	case MOD:	return "mod";
	case DIV:	return "div";
	case DSLASH:	return "//";
	case NEQ:	return "!=";
	case LEQ:	return "<=";
	case GEQ:	return ">=";
	case NAMETEST:	return "{name}";
	case FUNCNAME:	return "{function}";
	case NODETYPE:	return "{nodetype}";
	case AXISNAME:	return "{axis}";
	case DPERIOD:	return "..";
	default:
		if (token < ' ' || token > '~') {
			buf[0] = '\\';
			buf[1] = 'x';
			buf[2] = hex[(token >> (3*4)) & 0xf];
			buf[3] = hex[(token >> (2*4)) & 0xf];
			buf[4] = hex[(token >> (1*4)) & 0xf];
			buf[5] = hex[(token >> (0*4)) & 0xf];
			buf[6] = '\0';
		} else {
			buf[0] = token;
			buf[1] = '\0';
		}
		return buf;
	}
}

/* Consume an expected token */
static void
expect(in, token)
	struct input *in;
	int token;
{
	if (in->next != token) {
	    if (!in->error) {
		static char buf[80];
		snprintf(buf, sizeof buf, "expected %s", tokname(token));
		error(in, buf);
	    }
	} else
	    next(in);
}

static wchar_t
nextch(in)
	struct input *in;
{
	wchar_t c = in->nextc;

	if (ENTITY_EOF(in->entity))
		in->nextc = -1;
	else
		in->nextc = ENTITY_GETCH(in->entity);

	if (xml_debug)
	    fprintf(stderr, "nextch() -> \\x%04x '%c'; nextc  \\x%04x '%c'\n",
		c, c, in->nextc, in->nextc);

	VECTOR_APPEND(*in->next_raw, c);
	return c;

}

/* Lexical analysis */
static int
next(in)
	struct input *in;
{
	wchar_t	c;
	int prev1;
	wstr_t name;
	int havecolon;

	in->prev = in->next;

	if (in->error)
		goto fail;

	/* skip white space: S */
	while (isSpace(in->nextc))
		nextch(in);

	VECTOR_LEN(*in->next_raw) = 0;

	/* end-of-file */
	if (in->nextc == -1) {
		in->next = -1;
		goto out;
	}

	/* quoted literals */
	if (in->nextc == '"' || in->nextc == '\'')  {
		wchar_t quote;
		wstr_t value;

		quote = nextch(in);
		value = wstr_new();
		while (in->nextc != quote && in->nextc != -1) {
			c = nextch(in);
			WSTR_ADD(value, c);
		}
		if (in->nextc == -1) {
			wstr_free(value);
			error(in, "unclosed quote");
			goto fail;
		}
		nextch(in);	/* consume the closing quote */
		in->next = LITERAL;
		if (in->next_literal)
			wstr_free(in->next_literal);
		in->next_literal = value;
		goto out;
	}

	/* numbers */
	if (wisdigit(in->nextc)) {
		double value = 0;

		while (wisdigit(in->nextc)) {
			value = value * 10 + nextch(in) - '0';
		}
		if (in->nextc == '.') {
			double mult = 1.0;
			nextch(in);
			while (wisdigit(in->nextc)) {
				mult /= 10;
				value = value + mult * (nextch(in) - '0');
			}
		}
		in->next = NUMERIC;
		in->next_number = value;
		goto out;
	}

	/* doubled things (::, //,  ..) */
	if (in->nextc == ':' || in->nextc == '.' || in->nextc == '/') {
		c = nextch(in);
		if (in->nextc == c) {
			nextch(in);
			in->next = c == ':' ? DCOLON 
				 : c == '/' ? DSLASH 
				 :            DPERIOD;
			goto out;
		}
		in->next = c;
		goto out;
	}

	/* partial equality <= >= != or < and > */
	if (in->nextc == '<' || in->nextc == '>' || in->nextc == '!') {
		c = nextch(in);
		if (in->nextc == '=') {
			nextch(in);
			in->next = (c == '<' ? LEQ : c == '>' ? GEQ : NEQ);
			goto out;
		}
		in->next = c;
		goto out;
	}

	/* 
         * If there is a preceding token and the preceding token
         * is not one of @, ::, (, [, , or an Operator, then a *
         * must be recognized as a MultiplyOperator and an NCName
         * must be recognized as an OperatorName.
	 */
	prev1 = (in->prev == '@' || in->prev == DCOLON || in->prev == '('
	      || in->prev == '[' || in->prev == ',' || isOperator(in->prev)
	      || in->prev == AXISNAME || in->prev == -2 /* start */);

	DPRINTF("next(): prev1 = %d (in->prev=%s)\n", prev1, tokname(in->prev));

	/* MultiplyOperator or '*' */
	if (in->nextc == '*') {
		nextch(in);
		if (prev1) {
			in->next_name.ns = NULL;
			in->next_name.local = NULL;
			in->next = NAMETEST;
		} else
			in->next = MULTIPLY;
		goto out;
	}

	/*
	 * NCName     : (Letter | '_') (NCNameChar)*
	 * NCNameChar : Letter | Digit | '.' | '-' | '_' 
	 *	      | CombiningChar | Extender
	 */
	if (in->nextc != '_' && !wisletter(in->nextc))  {
		/* All other tokens are represented by themselves */
		in->next = nextch(in);
		goto out;
	}
	name = wstr_new();
	while (in->nextc == '.' || in->nextc == '-' || in->nextc == '_'
	       || wisdigit(in->nextc) || wisletter(in->nextc))
	{
		c = nextch(in);
		WSTR_ADD(name, c);
	}
	name = intern_wstr(name);

	/* skip whitespace */
	while (isSpace(in->nextc)) 
		nextch(in);

	/* If !prev1 then an NCName must be recognised as an OperatorName */
	if (!prev1) {
		void *x;
		/* see if it's an Operator name */
		x = xhash_lookup(operator_hash, &name, sizeof name);
		if (x) {
			in->next = (int)x;	/* one of AND, OR, DIV etc */
			goto out;
		}
		error(in, "expected an operator");
		goto fail;
	}

	/* 
	 * Consume a colon if an operator name not recognised
	 */
	havecolon = 0;
	if (in->nextc == ':') {
		DPRINTF("next(): havecolon\n");
		nextch(in);
		havecolon = 1;
	}

	/*
         * If the two characters following an NCName (possibly after
         * intervening ExprWhitespace) are ::, then the token must
         * be recognized as an AxisName.
	 */
	if (havecolon && in->nextc == ':') {
		void *x;

		nextch(in);	/* XXX consumes the DCOLON token */
		DPRINTF("next(): saw a DCOLON\n");
		x = xhash_lookup(axis_hash, &name, sizeof name);
		if (!x) {
			error(in, "expected axis name before double colon");
			goto fail;
		}
		in->next = AXISNAME;
		in->next_axis = (enum axis)x;
		goto out;
	}

	if (!havecolon) {
		/* its just a local name */
		in->next_name.ns = NULL;
		in->next_name.local = name;
	} else {
		/* it is a QName */
		in->next_name.ns = name;

		if (in->nextc == '*') {
			in->next_name.local = NULL;
			nextch(in);
		} else {
			if (in->nextc != '_' && !wisletter(in->nextc)) {
				error(in, "expected name to follow colon");
				goto fail;
			}
			name = wstr_new();
			while (in->nextc == '.' || in->nextc == '-' 
			    || in->nextc == '_' || wisdigit(in->nextc)
			    || wisletter(in->nextc))
			{
				c = nextch(in);
				WSTR_ADD(name, c);
			}
			in->next_name.local = intern_wstr(name);
		}
	} 

	/*
         * If the character following an QName (possibly after
         * intervening ExprWhitespace) is (, then the token must
         * be recognized as a NodeType or a FunctionName.
	 */
	if (in->nextc == '(') {
		void *x;

		if (in->next_name.ns != NULL) {
			error(in, "prefixes not allowed on node types"
				  " or function names");
			goto fail;
		}
		name = in->next_name.local;
		/* assert(name != NULL) */

		x = xhash_lookup(nodetype_hash, &name, sizeof name);
		if (x) {
			in->next = NODETYPE;
			in->next_nodetype = (enum nodetype)x;
			goto out;
		}

		x = xhash_lookup(function_hash, &name, sizeof name);
		if (x) {
			in->next = FUNCNAME;
			in->next_function = (struct function *)x;
			goto out;
		}
		error(in, "unknown function or node type %s", 
			wstr_str(name));
		goto fail;
	}

	in->next = NAMETEST;
	goto out;

   fail:
	in->next = -1;

   out:

	if (xml_debug) {
		fprintf(stderr, "next() -> %s ", tokname(in->prev));
		fprintf(stderr, "... in->next = %s", tokname(in->next));
		switch (in->next) {
		case NUMERIC: fprintf(stderr, ", next_number = %f", 
			in->next_number); break;
		case LITERAL: fprintf(stderr, ", next_literal = \"%s\"",
			wstr_str(in->next_literal)); break;
		case AXISNAME: fprintf(stderr, ", next_axis = %d", 
			(int)in->next_axis); break;
		case FUNCNAME: fprintf(stderr, ", next_function = %p %s",
			in->next_function, in->next_function->utf8_name); 
			break;
		case NODETYPE: fprintf(stderr, ", next_nodetype = %s",
			in->next_nodetype == COMMENT_TYPE ? "COMMENT_TYPE" :
			in->next_nodetype == TEXT_TYPE ? "TEXT_TYPE" :
			in->next_nodetype == PI_TYPE ? "PI_TYPE" :
			in->next_nodetype == NODE_TYPE ? "NODE_TYPE" :
						"???"); break;
		case NAMETEST: 
			fprintf(stderr, ", next_name = ");
			if (in->next_name.ns)
				fprintf(stderr, "%s:", 
				    wstr_str(in->next_name.ns));
			if (in->next_name.local)
				fprintf(stderr, "%s", 
				    wstr_str(in->next_name.local));
			else
				fprintf(stderr, "*");
			break;
		}
		fprintf(stderr, "\n");
	}

	return in->prev;
}

/*------------------------------------------------------------
 * expression and path data structure management
 */

static struct expr *
expr_new(type)
	enum exprtype type;
{
	struct expr *e;

	e = (struct expr *)malloc(sizeof (struct expr));
	if (e) {
		e->type = type;
		switch (type) {
		default:
			e->u.bin.exp1 = e->u.bin.exp2 = NULL; break;
		case eNEG:
			e->u.bin.exp1 = NULL; break;
		case eVAR:
			e->u.name.ns = NULL;
			e->u.name.local = NULL;
			break;
		case eLIT:
			e->u.value.type = BOOLEAN;
			break;
		case eFUNC:
			e->u.func.func = NULL;
			VECTOR_INIT(e->u.func.arg);
			break;
		case ePATH:
			e->u.path.path = NULL;
			e->u.path.expr = NULL;
			break;
		}
	}
	return e;
}

static struct expr *
expr_newbin(type, e1, e2)
	enum exprtype type;
	struct expr *e1, *e2;
{
	struct expr *e;
	e = expr_new(type);
	e->u.bin.exp1 = e1;
	e->u.bin.exp2 = e2;
	return e;
}

static void
expr_free(e)
	struct expr *e;
{
	int i;

	if (e == NULL)
		return;

	switch (e->type) {
	default:
		expr_free(e->u.bin.exp1);
		expr_free(e->u.bin.exp2);
		break;
	case eNEG:
		expr_free(e->u.bin.exp1);
		break;
	case eVAR:
		break;
	case eLIT:
		VALUE_FINI(e->u.value);
		break;
	case eFUNC:
		for (i = 0; i < VECTOR_LEN(e->u.func.arg); i++)
			expr_free(VECTOR_AT(e->u.func.arg, i));
		VECTOR_FREE(e->u.func.arg);
		break;
	case ePATH:
		expr_free(e->u.path.expr);
		path_free(e->u.path.path);
		break;
	}
	free(e);
}

/* create a path step of form <axis>::node() with no filters */
static struct path *
path_new(axis)
	enum axis axis;
{
	struct path *p;

	p = (struct path *)malloc(sizeof (struct path));
	if (p) {
		p->axis = axis;
		p->nodetype = NODE_TYPE;
		p->nodename.ns = NULL;
		p->nodename.local = NULL;
		VECTOR_INIT(p->pred);
		p->next = NULL;
	}
	return p;
}

static struct path *
path_new_dslash()
{
	return path_new(DESCENDANT_OR_SELF_AXIS);
}

static void
path_free(p)
	struct path *p;
{
	int i;

	if (p == NULL)
		return;
	path_free(p->next);
	for (i = 0; i < VECTOR_LEN(p->pred); i++)
		expr_free(VECTOR_AT(p->pred, i));
	free(p);
}

/*------------------------------------------------------------
 * value conversions and operations
 */

static struct value *
value_alloc()
{
	return (struct value *)malloc(sizeof (struct value));
}

static void
value_free(v)
	struct value *v;
{
	if (v) {
		VALUE_FINI(*v);
		free(v);
	}
}

/* duplicate a value from src into (uninitialised) dst */
static void
value_copy(dst, src)
	struct value *dst, *src;
{
	dst->type = src->type;
	switch (src->type) {
	case NODESET:
		VECTOR_COPY(dst->v_nodeset, src->v_nodeset);
		return;
	case BOOLEAN:
		dst->v_boolean = src->v_boolean;
		return;
	case NUMBER:
		dst->v_number = src->v_number;
		return;
	case STRING:
		VECTOR_COPY(dst->v_string, src->v_string);
		return;
	}
}

/* add node n to the node-set v */
static void
value_addnode(v, n)
	struct value *v;
	xml_t n;
{
	int i;

	/* assert(v->type == NODESET); */
	for (i = 0; i < VECTOR_LEN(v->v_nodeset); i++)
		if (VECTOR_AT(v->v_nodeset, i) == n)
			return;
	VECTOR_APPEND(v->v_nodeset, n);
}

void
nodeset_add(ns, n)
	nodeset_t *ns;
	xml_t n;
{
	int i;

	/* XXX - should replace with bsearch */
	for (i = 0; i < VECTOR_LEN(*ns); i++)
		if (VECTOR_AT(*ns, i) == n)
			return;
	VECTOR_APPEND(*ns, n);
	/* XXX - and sort after adding */
}

/* compare two nodes using document order */
static int
xmlcmp(a, b)
	const void *a, *b;
{
	xml_t xa = (xml_t)a;
	xml_t xb = (xml_t)b;

	return xa->order - xb->order;
}

/* sort a nodeset */
static void
value_sort_nodes(v)
	struct value *v;
{
	/* assert(v->type == NODESET); */
	sort_nodes(&v->v_nodeset);
}

static void
sort_nodes(ns)
	nodeset_t *ns;
{
	VECTOR_SORT(*ns, xmlcmp);
}

/* add all the nodes in v2 into v1 */
static void
value_addnodes(v1, v2)
	struct value *v1, *v2;
{
	int i;
	/* assert(v1->type == NODESET); */
	/* assert(v2->type == NODESET); */

	for (i = 0; i < VECTOR_LEN(v2->v_nodeset); i++)
		value_addnode(v1, VECTOR_AT(v2->v_nodeset, i));
}

/* append w to the end of string v */
static void
value_append_wstr(v, w)
	struct value *v;
	wstr_t w;
{
	/* assert(v->type == STRING); */
	VECTOR_CONCAT(v->v_string, *w);
}

/* return first node of node-set, in document order */
static xml_t
first_node(v)
	struct value *v;
{
	int i;
	xml_t m, n;

	/* assert(v->type == NODESET); */
	n = NULL;
	for (i = 0; i < VECTOR_LEN(v->v_nodeset); i++) {
		m = VECTOR_AT(v->v_nodeset, i);
		if (n == NULL || m->order < n->order)
			n = m;
	}
	return n;
}

/* convert v to boolean */
static void
toboolean(v)
	struct value *v;
{
	int truth;

	if (v->type == BOOLEAN)
		return;
	truth = asboolean(v);
	VALUE_FINI(*v);
	VALUE_INIT_BOOLEAN(*v, truth);
}

/* evaluate v as a boolean */
static int
asboolean(v)
	struct value *v;
{
	switch (v->type) {
	case BOOLEAN:
		return v->v_boolean;
	case NODESET:
		return VECTOR_LEN(v->v_nodeset) != 0;
	case NUMBER:
		return !(isnan(v->v_number) || v->v_number == 0);
	case STRING:
		return VECTOR_LEN(v->v_string) != 0;
	}
	return 0;
}

/* convert v to string */
static void
tostring(v)
	struct value *v;
{
	struct value ret;

	if (v->type == STRING)
		return;

	VALUE_INIT_STRING(ret);
	asstring(v, &ret);
	VALUE_FINI(*v);
	memcpy(v, &ret, sizeof *v);
}

/* note: res must have already been initialised as string */
static void
asstring(v, res)
	struct value *v, *res;
{
	int i;
	double d;
	char buf[400];
	char *p, *bufp = "error";

	VALUE_INIT_STRING(*res);

	switch (v->type) {
	case NODESET:
		for (i = 0; i < VECTOR_LEN(v->v_nodeset); i++)
			append_string_value(res, VECTOR_AT(v->v_nodeset, i));
		return;
	case BOOLEAN:
		if (v->v_boolean) 
			bufp = "true";
		else	
			bufp = "false";
		break;
	case NUMBER:
		if (isnan(v->v_number)) {
			bufp = "NaN";
			break;
		}
		if (v->v_number == 0) {
			bufp = "0";
			break;
		}
		d = v->v_number;
		if (isinf(d)) {
			if (d < 0)
				bufp = "-Infinity";
			else
				bufp = "Infinity";
			break;
		}
		snprintf(buf, sizeof buf, "%.17g", d);	/* 17 digit precision for doubles */
		remove_exponent(buf);
		bufp = buf;
		break;
	case STRING:
		VECTOR_CONCAT(res->v_string, v->v_string);
		return;
	}

	/* Use buf as the result */
	for (p = bufp; *p; p++)
		VECTOR_APPEND(res->v_string, *p);
}

/*
 * remove exponents (e-123 or e+999) from the buffer;
 * assumes sufficient space (e+308 .. e-325 for double precision)
 */
static void
remove_exponent(buf)
	char *buf;
{
	char *p, *e;
	int exp;
	int negexp;
	int ilen;
	int isneg;

	/* scan for the 'e' */
	for (p = buf; *p; p++)
		if (*p == 'e')
			break;
	if (*p != 'e')
		return;	/* no exponent to remove */
	e = p;
	p++;
	if (*p == '-')
		negexp = 1;
	else if (*p == '+')
		negexp = 0;
	else
		return; /* malformed */
	for (exp = 0, p++; *p; p++) {
		if (*p < '0' || *p > '9')
			return; /* malformed */
		exp = exp * 10 + (*p - '0');
	}
	isneg = (buf[0] == '-');

	if (!negexp) {
		/* -1234.5678e99 */
		for (p = buf; *p; p++)
			if (*p == '.' || *p == 'e')
				break;
		/* -12345<678e99 */
		while (exp && *p != 'e') {
			*p = *(p+1);
			exp--;
			p++;
		}
		if (exp == 0 && *p == 'e') {
			*p = '\0';
			return;
		}
		if (exp == 0) {
			*p = '.';
			/* skip to the 'e' part and zap it */
			while (*p && *p != 'e')
				p++;
			*p = '\0';
			return;
		}
		for (; exp; p++, exp--)
			*p = '0';
		*p = '\0';
		return;
	}

	/* for negative exponents, we have to do some planning */
	/* first, find out where the point is */
	/* -1234.5678e-99 */

	/* find the decimal point */
	for (p = buf; *p; p++)
		if (*p == 'e' || *p == '.')
			break;
	ilen = exp - (p - buf) + 1;
	*(e+ilen) = '\0';
	/* move everything after the . forward ilen */
	for (p = e-1; p > buf && *p != '.'; p--)
		*(p+ilen) = *p;
	/* move everything before the . forward ilen+1 */
	for (p--; p >= buf && *p != '-'; p--)
		*(p+ilen+1) = *p;
	p = buf;
	if (*p == '-') p++;
	*p++ = '0';
	*p++ = '.';
	for (; ilen > 1; ilen--, p++)
		*p = '0';
}

/* convert v to a number */
static void
tonumber(v)
	struct value *v;
{
	double d;

	if (v->type == NUMBER)
		return;
	d = asnumber(v);
	VALUE_FINI(*v);
	VALUE_INIT_NUMBER(*v, d);
}

static double
asnumber(v)
	struct value *v;
{
	int i, neg;
	double d, e;
	wchar_t c;
	struct value s;

	switch (v->type) {
	case NUMBER:
		return v->v_number;
	case STRING:
		for (i = 0; i < VECTOR_LEN(v->v_string); i++) {
			c = VECTOR_AT(v->v_string, i);
			if (!isSpace(c))
				break;
		}
		if (VECTOR_AT(v->v_string, i) == '-') {
			neg = 1;
			i++;
		} else 
			neg = 0;
		if (i == VECTOR_LEN(v->v_string))
			d = NaN;
		else
		    /* XXX - does this do IEEE 754 round-to-nearest? */
		    for (d = 0; i < VECTOR_LEN(v->v_string); i++) {
			c = VECTOR_AT(v->v_string, i);
			if (c == '.') {
				for (i++, e = 1; i < VECTOR_LEN(v->v_string); i++) {
					c = VECTOR_AT(v->v_string, i);
					if (c < '0' || c > '9')
						break;
					e /= 10.0;
					d += e * (c - '0');
				}
				break;
			}
			if (c < '0' || c > '9') 
				break;
			d = d * 10.0 + (c - '0');
		    }
		for (; i < VECTOR_LEN(v->v_string); i++) {
			c = VECTOR_AT(v->v_string, i);
			if (!isSpace(c))
				break;
		}
		if (i < VECTOR_LEN(v->v_string))
			d = NaN;
		return d;
	case BOOLEAN:
		return v->v_boolean ? 1 : 0;
	case NODESET:
		VALUE_INIT_STRING(s);
		asstring(v, &s);
		d = asnumber(&s);
		VALUE_FINI(s);
		return d;
	}
	return NaN;
}


/* append the 'string-value' of n to the end of string v */
static void
append_string_value(v, n)
	struct value *v;
	xml_t n;
{
	switch (n->type) {
	case ROOT:
		/*
		 * The string-value of the root node is the concatenation of
		 * the string-values of all text node descendants of the root
		 * node in document order.
		 */
		append_text_nodes(v, n, 0);
		break;
	case ELEMENT:
		/* 
		 * The string-value of an element node is the concatenation of the 
		 * string-values of all text node descendants of the element node 
		 * in document order.
		 */
		append_text_nodes(v, n, 0);
		break;
	case TEXT:
		/* 
		 * The string-value of a text node is the character data.
		 */
		VECTOR_CONCAT(v->v_string, *n->aux.args);
		break;
	case ATTR:
		/* 
		 * The string-value is the normalized value
		 * as specified by the XML Recommendation
		 */
		VECTOR_CONCAT(v->v_string, *n->aux.attr.value);
		break;
	case NS:
		/*
		 * The string-value of a namespace node is the namespace
		 * URI that is being bound to the namespace prefix
		 */
		VECTOR_CONCAT(v->v_string, *n->aux.args);
		break;
	case PROCINST:
		/*
		 * The string-value of a processing instruction node is 
		 * the part of the processing instruction following
		 * the target and any whitespace. 
		 */
		VECTOR_CONCAT(v->v_string, *n->aux.args);
		break;
	case COMMENT:
		/*
		 * The string-value of comment is the content of the
		 * comment not including the opening <!-- or the closing -->
		 */
		VECTOR_CONCAT(v->v_string, *n->aux.args);
		break;
	}
}

/* append string-value of descendent text nodes to v */
static void
append_text_nodes(v, n, sib)
	struct value *v;
	xml_t n;
	int sib;
{
	if (n == NULL)
		return;
	if (n->type == TEXT) 
		append_string_value(v, n);
	append_text_nodes(v, n->child, 1);
	if (sib)
		append_text_nodes(v, n->sibling, sib);
}

static qname_t
expanded_name(n)
	xml_t n;
{
	qname_t q;

	switch (n->type) {
	case ELEMENT:
	case ATTR:
	case NS:
	case PROCINST:
		return n->name;
	default:
		q.ns = NULL;
		q.local = NULL;
		break;
	}
	return q;
}

/*------------------------------------------------------------
 * Syntax: paths
 */

/* [1] LocationPath */
static struct path *
LocationPath(in)
	struct input *in;
{
	DEBUG_FN;
	if (in->next == '/' || in->next == DSLASH)
		return AbsoluteLocationPath(in);
	else
		return RelativeLocationPath(in);
}

/* [2] AbsoluteLocationPath { / // } */
static struct path *
AbsoluteLocationPath(in)
	struct input *in;
{
	struct path *p;

	DEBUG_FN;
	if (in->next == DSLASH) 
		return AbbreviatedAbsoluteLocationPath(in);
	expect(in, '/');
	if (in->error)
		return NULL;
	p = path_new(ROOT_AXIS);
	if (isStep(in))
		p->next = RelativeLocationPath(in);
	return p;
}

/* [3] RelativeLocationPath and [11] AbbreviatedRelativeLocationPath */
/*
 * RelativeLocationPath = Step
 *			| Step '/' RelativeLocationPath
 *			| Step '//' RelativeLocationPath
 */
static struct path *
RelativeLocationPath(in)
	struct input *in;
{
	struct path *p;
	DEBUG_FN;
	p = Step(in);
	if (in->next == '/') {
		next(in);
		p->next = RelativeLocationPath(in);
	} if (in->next == DSLASH) {
		next(in);
		p->next = path_new_dslash();
		p->next->next = RelativeLocationPath(in);
	}
	return p;
}

/* Return true if one of the 'first' tokens of the Step production */
static int
isStep(in)
	struct input *in;
{
	return in->next == AXISNAME	/* AxisSpecifier.AxisName */
	    || in->next == '@'		/* AxisSpecifier.AbbreviatedAxisSpecifier */
	    || in->next == NAMETEST	/* AbbreviatedAxisSpecifier NodeTest */
	    || in->next == NODETYPE	/* AbbreviatedAxisSpecifier NodeTest */
	    || in->next == '.'		/* AbbreviatedStep */
	    || in->next == DPERIOD	/* AbbreviatedStep */
	    ;
}

/* [4] Step */
static struct path *
Step(in)
	struct input *in;
{
	struct path *p;
	enum axis axis;

	DEBUG_FN;
	if (in->next == '.' || in->next == DPERIOD)
		return AbbreviatedStep(in);

	axis = AxisSpecifier(in);
	p = path_new(axis);
	NodeTest(in, p);
	while (more(in) && in->next == '[')  {
		struct expr *e = Predicate(in);
		if (e) VECTOR_APPEND(p->pred, e);
	}
	return p;
}

/* [5] AxisSpecifier */
static enum axis
AxisSpecifier(in)
	struct input *in;
{
	enum axis ax;

	DEBUG_FN;
	if (in->next == AXISNAME) {
		ax = in->next_axis;
		next(in);
		/* note - DCOLON is part of the AXISNAME token */
		return ax;
	} else
		return AbbreviatedAxisSpecifier(in);
}

/* [7] NodeTest */
static void
NodeTest(in, p)
	struct input *in;
	struct path *p;
{
	DEBUG_FN;
	if (in->next == NAMETEST) {
		p->nodetype = NAMETEST_TYPE;
		p->nodename = in->next_name;
		next(in);
		return;
	}
	if (in->next == NODETYPE) {
		p->nodetype = in->next_nodetype;
		next(in);
		expect(in, '(');
		if (p->nodetype == PI_TYPE) {
		    if (in->next == LITERAL) {
			p->nodename.local = intern_wstr(in->next_literal);
			in->next_literal = NULL; /* we now own it that literal */
			next(in);
		    } else
			p->nodename.local = NULL;
		}
		expect(in, ')');
		return;
	}
	error(in, "expected node type or name test");
}

/* [8] Predicate */
static struct expr *
Predicate(in)
	struct input *in;
{
	struct expr *e;

	DEBUG_FN;
	expect(in, '[');
	e = PredicateExpr(in);
	expect(in, ']');
	return e;
}

/* [9] PredicateExpr */
static struct expr *
PredicateExpr(in)
	struct input *in;
{
	DEBUG_FN;
	return Expr(in);
}

/* [10] AbbreviatedAbsoluteLocationPath { // } */
static struct path *
AbbreviatedAbsoluteLocationPath(in)
	struct input *in;
{
	struct path *p;

	DEBUG_FN;
	expect(in, DSLASH);
	p = path_new(ROOT_AXIS);
	p->next = path_new_dslash();
	p->next->next = RelativeLocationPath(in);
	return p;
}

/* [12] AbbreviatedStep { . DPERIOD } */
static struct path *
AbbreviatedStep(in)
	struct input *in;
{
	DEBUG_FN;
	if (in->next == '.') {
		next(in);
		return path_new(SELF_AXIS);
	}
	if (in->next == DPERIOD) {
		next(in);
		return path_new(PARENT_AXIS);
	}
	error(in, "expected . or ..");
	return NULL;
}

/* [13] AbbreviatedAxisSpecifier { any } */
static enum axis
AbbreviatedAxisSpecifier(in)
	struct input *in;
{
	DEBUG_FN;
	if (in->next == '@') {
		next(in);
		return ATTRIBUTE_AXIS;
	} else
		return CHILD_AXIS;
}

/*------------------------------------------------------------
 * Syntax: expressions
 */

/* [14] Expr */
static struct expr *
Expr(in)
	struct input *in;
{
	DEBUG_FN;
	return OrExpr(in);
}

/* [15] PrimaryExpr */
static struct expr *
PrimaryExpr(in)
	struct input *in;
{
	struct expr *e;

	DEBUG_FN;
	if (in->next == '$') {
		next(in);
		if (in->next != NAMETEST) {
			error(in, "expected variable name after $");
			return NULL;
		}
		if (in->next_name.local == NULL) {
			error(in, "invalid variable name");
			return NULL;
		}
		e = expr_new(eVAR);
		e->u.name = in->next_name;
		next(in);
		return e;
	}
	if (in->next == '(') {
		next(in);
		e = Expr(in);
		expect(in, ')');
		return e;
	}
	if (in->next == LITERAL) {
		e = expr_new(eLIT);
		VALUE_INIT_STRING(e->u.value);
		memcpy(&e->u.value.v_string, in->next_literal, sizeof (vector_t));
		in->next_literal = NULL;
		next(in);
		return e;
	}
	if (in->next == NUMERIC) {
		e = expr_new(eLIT);
		VALUE_INIT_NUMBER(e->u.value, in->next_number);
		next(in);
		return e;
	}
	if (in->next == FUNCNAME)
		return FunctionCall(in);
	error(in, "expected $, (, string, number or function call");
	return NULL;
}

/* [16] FunctionCall */
static struct expr *
FunctionCall(in)
	struct input *in;
{
	struct expr *e, *arg;
	struct function *f;
	const char *m;

	DEBUG_FN;
	f = in->next_function;
	expect(in, FUNCNAME);
	expect(in, '(');
	if (in->error)
		return NULL;
	e = expr_new(eFUNC);
	e->u.func.func = f;
	if (more(in) && in->next != ')') {
		arg = Argument(in);
		VECTOR_APPEND(e->u.func.arg, arg);
		while (more(in) && in->next == ',') {
			expect(in, ',');
			arg = Argument(in);
			VECTOR_APPEND(e->u.func.arg, arg);
		}
	}
	m = function_arity_check(f, VECTOR_LEN(e->u.func.arg));
	if (m)
		error(in, "%s", m);
	expect(in, ')');
	return e;
}

/* [17] Argument */
static struct expr *
Argument(in)
	struct input *in;
{
	DEBUG_FN;
	return Expr(in);
}

/* [18] UnionExpr */
static struct expr *
UnionExpr(in)
	struct input *in;
{
	struct expr *e;

	DEBUG_FN;
	e = PathExpr(in);
	if (in->next != '|')
		return e;
	expect(in, '|');
	return expr_newbin(eUNION, e, UnionExpr(in));
}

/* [19] PathExpr */
static struct expr *
PathExpr(in)
	struct input *in;
{
	struct expr *e, *u;

	DEBUG_FN;
	if (in->next == '$'
	 || in->next == '('
	 || in->next == LITERAL
	 || in->next == NUMERIC
	 || in->next == FUNCNAME)
	{
		e = FilterExpr(in);
		if (in->next == '/') {
			expect(in, '/');
			u = expr_new(ePATH);
			u->u.path.expr = e;
			u->u.path.path = RelativeLocationPath(in);
			return u;
		}
		if (in->next == DSLASH) {
			expect(in, DSLASH);
			u = expr_new(ePATH);
			u->u.path.expr = e;
			u->u.path.path = path_new_dslash();
			u->u.path.path->next = RelativeLocationPath(in);
			return u;
		}
		return e;
	} else {
		e = expr_new(ePATH);
		e->u.path.path = LocationPath(in);
		e->u.path.expr = expr_new(eCONTEXT);
		return e;
	}
}

/* [20] FilterExpr */
static struct expr *
FilterExpr(in)
	struct input *in;
{
	struct expr *e;

	DEBUG_FN;
	e = PrimaryExpr(in);
	while (more(in) && in->next == '[') {
		struct expr *t = expr_new(eFILTER);
		t->u.bin.exp1 = e;
		t->u.bin.exp2 = Predicate(in);
		e = t;
	}
	return e;
}

/* [21] OrExpr */
static struct expr *
OrExpr(in)
	struct input *in;
{
	struct expr *e;

	DEBUG_FN;
	e = AndExpr(in);
	while (more(in) && in->next == OR) {
		expect(in, OR);
		e = expr_newbin(eOR, e, AndExpr(in));
	}
	return e;
}

/* [22] AndExpr */
static struct expr *
AndExpr(in)
	struct input *in;
{
	struct expr *e;

	DEBUG_FN;
	e = EqualityExpr(in);
	while (more(in) && in->next == AND) {
		expect(in, AND);
		e = expr_newbin(eAND, e, EqualityExpr(in));
	}
	return e;
}

/* [23] EqualityExpr */
static struct expr *
EqualityExpr(in)
	struct input *in;
{
	struct expr *e;
	int tok;

	DEBUG_FN;
	e = RelationalExpr(in);
	while (more(in) && (in->next == '=' || in->next == NEQ)) {
		tok = next(in);
		e = expr_newbin(tok == '=' ? eEQ : eNEQ,
		    e, RelationalExpr(in));
	}
	return e;
}

/* [24] RelationalExpr */
static struct expr *
RelationalExpr(in)
	struct input *in;
{
	struct expr *e;
	int tok;

	DEBUG_FN;
	e = AdditiveExpr(in);
	while (more(in) && (in->next == '<' || in->next == '>' ||
	       in->next == LEQ || in->next == GEQ))
	{
		tok = next(in);
		e = expr_newbin(
			tok == '<' ? eLT :
			tok == '>' ? eGT :
			tok == LEQ ? eLEQ :
			             eGEQ,
			e, AdditiveExpr(in));
	}
	return e;
}

/* [25] AdditiveExpr */
static struct expr *
AdditiveExpr(in)
	struct input *in;
{
	struct expr *e;
	int tok;

	DEBUG_FN;
	e = MultiplicativeExpr(in);
	while (more(in) && (in->next == '+' || in->next == '-')) {
		tok = next(in);
		e = expr_newbin(tok == '+' ? eADD : eSUB,
		    e, MultiplicativeExpr(in));
	}
	return e;
}

/* [25] MultiplicativeExpr */
static struct expr *
MultiplicativeExpr(in)
	struct input *in;
{
	struct expr *e;
	int tok;

	DEBUG_FN;
	e = UnaryExpr(in);
	while (more(in) && (in->next == DIV || in->next == MOD || 
	    in->next == MULTIPLY)) {
		tok = next(in);
		e = expr_newbin(tok == DIV ? eDIV : 
				tok == MOD ? eMOD : 
					     eMUL,
			        e,
				UnaryExpr(in));
	}
	return e;
}

/* [25] MultiplicativeExpr */
static struct expr *
UnaryExpr(in)
	struct input *in;
{
	struct expr *e;

	DEBUG_FN;
	if (in->next == '-') {
		next(in);
		e = expr_new(eNEG);
		e->u.unary = UnaryExpr(in);
		return e;
	}
	return UnionExpr(in);
}

/*------------------------------------------------------------
 * expresion evaluation
 */

static const char *
expr_eval(c, e, result)
	struct context *c;
	struct expr *e;
	struct value *result;
{
	const char *m;
	struct value v1, v2, *vp;
	double d1, d2, dr;
	int truth;
	struct value *args;
	int i, len;
	qname_t qname;

	switch (e->type) {
	case eUNION:
		m = expr_eval(c, e->u.bin.exp1, result);
		if (m) return m;
		if (result->type != NODESET) {
			VALUE_FINI(*result);
			return "lhs of | operator not a node-set";
		}
		m = expr_eval(c, e->u.bin.exp2, &v1);
		if (m) return m;
		if (v1.type != NODESET) {
			VALUE_FINI(*result);
			VALUE_FINI(v1);
			return "rhs of | operator not a node-set";
		}
		value_addnodes(result, &v1);
		VALUE_FINI(v1);
		value_sort_nodes(result);
		return NULL;
	case eOR:
	case eAND:
		m = expr_eval(c, e->u.bin.exp1, result);
		if (m) return m;
		toboolean(result);
		/* short circuit */
		if (e->type == eOR && result->v_boolean)
			return NULL;
		if (e->type == eAND && !result->v_boolean)
			return NULL;
		m = expr_eval(c, e->u.bin.exp2, result);
		if (m) return m;
		toboolean(result);
		return NULL;
	case eEQ:
	case eNEQ:
	case eLT:
	case eGT:
	case eLEQ:
	case eGEQ:
		m = expr_eval(c, e->u.bin.exp1, &v1);
		if (m) return m;
		m = expr_eval(c, e->u.bin.exp2, &v2);
		if (m) { VALUE_FINI(v1); return m; }
		truth = value_compare(&v1, &v2, e->type);
		VALUE_FINI(v1);
		VALUE_FINI(v2);
		VALUE_INIT_BOOLEAN(*result, truth);
		return NULL;
	case eADD:
	case eSUB:
	case eMUL:
	case eDIV:
	case eMOD:
		m = expr_eval(c, e->u.bin.exp1, &v1);
		if (m) return m;
		m = expr_eval(c, e->u.bin.exp2, &v2);
		if (m) { VALUE_FINI(v1); return m; }
		d1 = asnumber(&v1);
		VALUE_FINI(v1);
		d2 = asnumber(&v2);
		VALUE_FINI(v2);
		switch (e->type) {
		case eADD:	dr = d1 + d2; break;
		case eSUB:	dr = d1 - d2; break;
		case eMUL:	dr = d1 * d2; break;
		case eDIV:	dr = d1 / d2; break;
		case eMOD:	dr = d1 - d2*(int)(d1 / d2); break;
		default:	dr = 0.0; /* NOTREACHED */
		}
		VALUE_INIT_NUMBER(*result, dr);
		return NULL;
	case eNEG:
		m = expr_eval(c, e->u.unary, &v1);
		if (m) return m;
		dr = -asnumber(&v1);
		VALUE_FINI(v1);
		VALUE_INIT_NUMBER(*result, dr);
		return NULL;
	case eVAR:
		qname.local = e->u.name.local;
		/*
		 * XXX $QName with no prefix always assumed to 
		 * have a null URI.
		 */
		if (e->u.name.ns)
			qname.ns = xmlns_lookup(c->ns, e->u.name.ns);
		else
			qname.ns = NULL;
		vp = (struct value *)xhash_lookup(c->variables,
			&qname, sizeof qname);
		if (!vp)
			return "variable not bound";
		value_copy(result, vp);
		return NULL;
	case eLIT:
		value_copy(result, &e->u.value);
		return NULL;
	case eFUNC:
		len = VECTOR_LEN(e->u.func.arg);
		if (len)
			args = (struct value *)malloc(len * 
				sizeof (struct value));
		else
			args = NULL;
		for (i = 0; i < len; i++) {
			m = expr_eval(c, VECTOR_AT(e->u.func.arg, i), &args[i]);
			if (m) {
				while (--i >= 0)
					VALUE_FINI(args[i]);
				if (len) free(args);
				return m;
			}
		}
		m = function_type_check(e->u.func.func, args, len);
		if (!m)
			m = (*e->u.func.func->fn)(c, args, len, result);
		for (i = 0; i < len; i++)
			VALUE_FINI(args[i]);
		if (args)
			free(args);
		return m;
	case ePATH:
		/* generate a nodeset from a path and initial nodeset */
		m = expr_eval(c, e->u.path.expr, &v1);
		if (m) return m;
		if (v1.type != NODESET) {
			VALUE_FINI(v1);
			return "expected node-set result before path";
		}
		VALUE_INIT_NODESET(*result);
		m = path_eval(c, e->u.path.path, &v1.v_nodeset, &result->v_nodeset);
		VALUE_FINI(v1);
		return m;
	case eFILTER:
		/* nodeset-yielding expression, filtered by a predicate  */
		m = expr_eval(c, e->u.bin.exp1, &v1);
		if (m) return m;
		if (v1.type != NODESET) {
			VALUE_FINI(v1);
			return "expected node-set result before predicate";
		}
		sort_nodes(&v1.v_nodeset);
		VALUE_INIT_NODESET(*result);
		m = filter(c, &v1.v_nodeset, e->u.bin.exp2, &result->v_nodeset);
		VALUE_FINI(v1);
		return m;

	case eCONTEXT:
		/* generate a node-set containing the context node */
		VALUE_INIT_NODESET(*result);
		VECTOR_APPEND(result->v_nodeset, c->node);
		return NULL;
	default:
		return "expr_eval(): internal error";
	}
}

/* Compare two values */
static int
value_compare(v1, v2, cmp)
	struct value *v1, *v2;
	enum exprtype cmp;
{
	int res;
	double d1, d2;

/* reverse the direction of the comparison operator */
#define REVERSE(cmp)	(cmp == eLT ? eGT : \
			 cmp == eGT ? eLT : \
			 cmp == eGEQ ? eLEQ : \
			 cmp == eLEQ ? eGEQ :  cmp)

	if (v1->type == NODESET && v2->type == NODESET) {
		int i, j;
	        struct value s1, s2;
		xml_t n1, n2;

		VALUE_INIT_STRING(s1);
		VALUE_INIT_STRING(s2);
		res = 0;
		for (i = 0; !res && i < VECTOR_LEN(v1->v_nodeset); i++) {
		    n1 = VECTOR_AT(v1->v_nodeset, i);
		    VECTOR_LEN(s1.v_string) = 0;
		    append_string_value(&s1, n1);
		    for (j = 0; !res && j < VECTOR_LEN(v2->v_nodeset); j++) {
		        n2 = VECTOR_AT(v2->v_nodeset, j);
		        VECTOR_LEN(s2.v_string) = 0;
		        append_string_value(&s2, n2);
			res = value_compare(&s1, &s2, cmp);
		    }
		}
		VALUE_FINI(s1);
		VALUE_FINI(s2);
		return res;
	}

	if (v1->type == NODESET) {
		int i;
		struct value s;
		xml_t n;

		res = 0;
		for (i = 0; !res && i < VECTOR_LEN(v1->v_nodeset); i++) {
			n = VECTOR_AT(v1->v_nodeset, i);
			VALUE_INIT_STRING(s);
		        append_string_value(&s, n);
			if (v2->type == NUMBER)
				tonumber(&s);
			else if (v2->type == BOOLEAN)
				toboolean(&s);
			res = value_compare(&s, v2, cmp);
			VALUE_FINI(s);
		}
		return res;
	}
	if (v2->type == NODESET)
		return value_compare(v2, v1, REVERSE(cmp));

	if (cmp == eEQ || cmp == eNEQ) {
		struct value s1;
		if (v2->type == BOOLEAN)
			return value_compare(v2, v1, cmp);
		if (v1->type == BOOLEAN) {
			if (cmp == eEQ)
				return v1->v_boolean != !asboolean(v2); /* tricky */
			else
				return v1->v_boolean != asboolean(v2);
		}
		if (v2->type == NUMBER)
			return value_compare(v2, v1, cmp);
		if (v1->type == NUMBER) {
			if (cmp == eEQ)
				return v1->v_number == asnumber(v2);
			else
				return v1->v_number != asnumber(v2);
		}
		if (v1->type == STRING && v2->type == STRING) {
			int eq = VECTOR_EQ(v1->v_string, v2->v_string);
			return cmp == eEQ ? eq : !eq;
		}
		if (v1->type == STRING)
			return value_compare(v2, v1, cmp);
		VALUE_INIT_STRING(s1);
		asstring(v1, &s1);
		res = value_compare(v2, &s1, cmp);
		VALUE_FINI(s1);
		return res;
	}

	d1 = asnumber(v1);
	d2 = asnumber(v2);
	switch (v2->type) {
		case eLT:	return d1 < d2;
		case eLEQ:	return d1 <= d2;
		case eGT:	return d1 > d2;
		case eGEQ:	return d1 >= d2;
		default:	return 0;		/* XXX */
	}
}

/*------------------------------------------------------------
 * path evaluation
 */

static const char *
path_eval(c, p, src, res)
	struct context *c;
	struct path *p;
	nodeset_t *src, *res;
{
	const char *m;
	nodeset_t nodes;
	int i;

	VECTOR_INIT(nodes);
	for (i = 0; i < VECTOR_LEN(*src); i++) 
		if ((m = path_generate(c, p, p->axis, 
		    VECTOR_AT(*src, i), &nodes)))
		{
			VECTOR_FREE(nodes);
			return m;
		}

	sort_nodes(&nodes);

	/* Apply the filter to each node generated by the path */
	for (i = 0; i < VECTOR_LEN(p->pred); i++) {
		nodeset_t leftover;
		m = filter(c, &nodes, VECTOR_AT(p->pred, i), &leftover);
		if (m) {
			VECTOR_FREE(nodes);
			return m;
		}
		VECTOR_FREE(nodes);
		memcpy(&nodes, &leftover, sizeof nodes);
	}
	if (p->next) {
		m = path_eval(c, p->next, &nodes, res);
		VECTOR_FREE(nodes);
	} else {
		memcpy(res, &nodes, sizeof *res);
		m = NULL;
	}
	return m;
}

static const char *
path_addnode(c, p, n, result)
	struct context *c;
	struct path *p;
	xml_t n;
	nodeset_t *result;
{
	switch (p->nodetype) {
	case COMMENT_TYPE:
		if (n->type != COMMENT)
			return NULL;
		break;
	case TEXT_TYPE:
		if (n->type != TEXT)
			return NULL;
		break;
	case PI_TYPE:
		if (n->type != PROCINST)
			return NULL;
		if (p->nodename.local && 
		    p->nodename.local != n->name.local)
			return NULL;
		break;
	case NAMETEST_TYPE:
		if (n->type != AXIS_PRINCIPAL(p->axis))
			return NULL;
		if (p->nodename.local && 
		    p->nodename.local != n->name.local)
			return NULL;
		if (p->nodename.ns == NULL && 
		    p->nodename.ns != NULL)
			return NULL;
		if (p->nodename.ns != NULL) {
			wstr_t uri = xmlns_lookup(c->ns,
			    p->nodename.ns);
			if (uri == xmlns_unknown)
			    return "unknown namespace";
			if (uri != n->name.ns)
				return NULL;
		}
		break;
	case NODE_TYPE:
		/* match everything */
		break;
	}

	nodeset_add(result, n);
	return NULL;
}

/* Append all the nodes from the given axis about n to result */
static const char *
path_generate(c, p, ax, n, result)
	struct context *c;
	struct path *p;
	enum axis ax;
	xml_t n;
	nodeset_t *result;
{
	xml_t n2;
	const char *m;

	switch (ax) {
	case ROOT_AXIS:
		/* the root node of the document */
		while (n && n->type != ROOT)
		    n = n->parent;
		if (!n)
			return "path_generate(): no root";
		return path_addnode(c, p, n, result);
	case ANCESTOR_AXIS:
		for (n = n->parent; n; n = n->parent)
		    if ((m = path_addnode(c, p, n, result)))
			return m;
		return NULL;
	case ANCESTOR_OR_SELF_AXIS:
		if ((m = path_addnode(c, p, n, result)))
		    return m;
		return path_generate(c, p, ANCESTOR_AXIS, n, result);
	case ATTRIBUTE_AXIS:
		if (n->type == ELEMENT) 
		    for (n = n->aux.element.first_attr; n; n = n->sibling)
			if ((m = path_addnode(c, p, n, result)))
			    return m;
		return NULL;
	case CHILD_AXIS:
		for (n = n->child; n; n = n->sibling)
		    if ((m = path_addnode(c, p, n, result)))
			return m;
		return NULL;
	case DESCENDANT_AXIS:
		/* XXX this case is probably a bottleneck! */
		for (n = n->child; n; n = n->sibling) {
		    if ((m = path_addnode(c, p, n, result)))
			return m;
		    if ((m = path_generate(c, p, ax, n, result)))
			return m;
		}
		return NULL;
	case DESCENDANT_OR_SELF_AXIS:
		if ((m = path_addnode(c, p, n, result)))
			return m;
		return path_generate(c, p, DESCENDANT_AXIS, n, result);
	case FOLLOWING_AXIS:
		if (n->type != ATTR && n->type != NS)
		    for (n = n->sibling; n; n = n->sibling) {
			if ((m = path_addnode(c, p, n, result)))
				return m;
			if ((m = path_generate(c, p, DESCENDANT_AXIS,
			    n, result)))
				return m;
		    }
		return NULL;
	case FOLLOWING_SIBLING_AXIS:
		if (n->type != ATTR && n->type != NS)
		    for (n = n->sibling; n; n = n->sibling)
			if ((m = path_addnode(c, p, n, result)))
			    return m;
		return NULL;
	case NAMESPACE_AXIS:
		if (n->type == ELEMENT)
		    for (n = n->ns; n; n = n->ns)
			if ((m = path_addnode(c, p, n, result)))
			    return m;
		return NULL;
	case PARENT_AXIS:
		if (n->parent)
		    return path_addnode(c, p, n->parent, result);
		return NULL;
	case PRECEDING_AXIS:
		if ((n->type != ATTR && n->type != NS) && n->parent)
		    for (n2 = n->parent->child; n2 && n2 != n; n2 = n2->sibling) {
			if ((m = path_addnode(c, p, n2, result)))
			    return m;
			if ((m = path_generate(c, p, DESCENDANT_AXIS, n2, result)))
			    return m;
		    }
		return NULL;
	case PRECEDING_SIBLING_AXIS:
		if ((n->type != ATTR && n->type != NS) && n->parent)
		    for (n2 = n->parent->child; n2 && n2 != n; n2 = n2->sibling)
			if ((m = path_addnode(c, p, n2, result)))
			    return m;
		return NULL;
	case SELF_AXIS:
		return path_addnode(c, p, n, result);
	}
	return "path_generate(): internal error";
}

/* initialise and fill result, with those of src that predicate e is true for */
static const char *
filter(c, src, e, result)
	struct context *c;
	nodeset_t *src;
	struct expr *e;
	nodeset_t *result;
{
	struct context sc;
	struct value v;
	int truth;
	int i;
	const char *m;

	VECTOR_INIT(*result);
	sc.doc = c->doc;
	sc.variables = c->variables;
	sc.reverse = c->reverse;
	sc.ns = c->ns;

	sc.size = VECTOR_LEN(*src);
	for (i = 0; i < VECTOR_LEN(*src); i++) {
		sc.node = VECTOR_AT(*src, i);
		sc.position = c->reverse ? sc.size - i : 1 + i;
		m = expr_eval(&sc, e, &v);
		if (m) {
			VECTOR_FREE(*result);
			return m;
		}
		if (v.type == NUMBER)
			truth = (rint(v.v_number) == sc.position);
		else
			truth = asboolean(&v);
		VALUE_FINI(v);
		if (truth)
			VECTOR_APPEND(*result, sc.node);
	}
	return NULL;
}

/*------------------------------------------------------------
 * display of an xpath for debugging
 */
int
xpath_puts(xp, puts, closure)
	xpath_t xp;
	int (*puts)(const char *, void *);
	void *closure;
{
	return expr_put(xp->expr, puts, closure);
}

static int
expr_put(e, puts, closure)
	struct expr *e;
	int (*puts)(const char *, void *);
	void *closure;
{
	int i;

	switch (e->type) {
	case eUNION:
	case eOR:
	case eAND:
	case eEQ:
	case eNEQ:
	case eLT:
	case eGT:
	case eLEQ:
	case eGEQ:
	case eADD:
	case eSUB:
	case eMUL:
	case eDIV:
	case eMOD:
		if (e->u.bin.exp1->type < e->type)
		     if (!(*puts)("(", closure))
			return 0;
		if (!expr_put(e->u.bin.exp1, puts, closure))
			return 0;
		if (e->u.bin.exp1->type < e->type)
		     if (!(*puts)(")", closure))
			return 0;
		if (!(*puts)(
			e->type == eUNION ? " | " :
			e->type == eOR    ? " or " :
			e->type == eAND   ? " and " :
			e->type == eEQ    ? " = " :
			e->type == eNEQ   ? " != " :
			e->type == eLT    ? " < " :
			e->type == eGT    ? " > " :
			e->type == eLEQ   ? " <= " :
			e->type == eGEQ   ? " >= " :
			e->type == eADD   ? " + " :
			e->type == eSUB   ? " - " :
			e->type == eMUL   ? " * " :
			e->type == eDIV   ? " div " :
			e->type == eMOD   ? " mod " :
					    " ? ", closure))
				return 0;
		if (e->u.bin.exp2->type < e->type)
		     if (!(*puts)("(", closure))
			return 0;
		if (!expr_put(e->u.bin.exp2, puts, closure))
			return 0;
		if (e->u.bin.exp2->type < e->type)
		     if (!(*puts)(")", closure))
			return 0;
		return 1;
	case eNEG:
		if (e->u.unary->type < e->type)
		     if (!(*puts)("(", closure))
			return 0;
		if (!expr_put(e->u.unary, puts, closure))
			return 0;
		if (e->u.unary->type < e->type)
		     if (!(*puts)(")", closure))
			return 0;
		return 1;
	case eVAR:
		if (!(*puts)("$", closure))
			return 0;
		if (e->u.name.ns) {
			if (!puts_wstr(e->u.name.ns, puts, closure, 0))
				return 0;
			if (!(*puts)(":", closure))
				return 0;
		}
		if (!puts_wstr(e->u.name.local, puts, closure, 0))
			return 0;
		return 1;
	case eLIT:
		return value_put(&e->u.value, puts, closure);
	case eFUNC:
		if (!(*puts)(e->u.func.func->utf8_name, closure))
			return 0;
		if (!(*puts)("(", closure))
			return 0;
		for (i = 0; i < VECTOR_LEN(e->u.func.arg); i++) {
			if (i > 0 && !(*puts)(",", closure))
				return 0;
			expr_put(VECTOR_AT(e->u.func.arg, i),
			    puts, closure);
		}
		return (*puts)(")", closure);
	case ePATH:
		if (e->u.path.expr->type == eCONTEXT)
		    return path_put(e->u.path.path, puts, closure);
		return expr_put(e->u.path.expr, puts, closure)
		    && (*puts)("/", closure)
		    && path_put(e->u.path.path, puts, closure)
		    ;
	case eFILTER:
		return expr_put(e->u.bin.exp1, puts, closure)
		    && (*puts)("[", closure)
		    && expr_put(e->u.bin.exp2, puts, closure)
		    && (*puts)("]", closure)
		    ;
	case eCONTEXT:
		return (*puts)("{context-node}", closure);
	default:
		return (*puts)("?expr?", closure);
	}
}

static int
path_put(p, puts, closure)
	struct path *p;
	int (*puts)(const char *, void *);
	void *closure;
{
	static struct {
		enum axis ax;
		const char *name;
	} axnames[] = {
#ifdef DEBUG
		{ ROOT_AXIS, "{root::}" },
#else
		{ ROOT_AXIS, "" },
#endif
		{ ANCESTOR_AXIS, "ancestor::" },
		{ ANCESTOR_OR_SELF_AXIS, "ancestor-or-self::" },
		{ ATTRIBUTE_AXIS, "@" },
#ifdef DEBUG
		{ CHILD_AXIS, "child::" },
#else
		{ CHILD_AXIS, "" },
#endif
		{ DESCENDANT_AXIS, "descendant::" },
		{ DESCENDANT_OR_SELF_AXIS, "descendant-or-self::" },
		{ FOLLOWING_AXIS, "following::" },
		{ FOLLOWING_SIBLING_AXIS, "following-sibling::" },
		{ NAMESPACE_AXIS, "namespace::" },
		{ PARENT_AXIS, "parent::" },
		{ PRECEDING_AXIS, "preceding::" },
		{ PRECEDING_SIBLING_AXIS, "preceding-sibling::" },
		{ SELF_AXIS, "self::" }
	};
	int i;
		
	/* abbreviate // */
	if (p->axis == DESCENDANT_OR_SELF_AXIS &&
	    p->nodetype == NODE_TYPE &&
	    VECTOR_LEN(p->pred) == 0)
		goto next;

	/* abbreviate . */
	if (p->axis == SELF_AXIS && p->nodetype == NODE_TYPE) {
		if (!(*puts)(".", closure))
			return 0;
		goto preds;
	}

	/* abbreviate .. */
	if (p->axis == PARENT_AXIS && p->nodetype == NODE_TYPE) {
		if (!(*puts)("..", closure))
			return 0;
		goto preds;
	}

	for (i = 0; i < sizeof axnames / sizeof axnames[0]; i++)
	    if (axnames[i].ax == p->axis) {
		if (!(*puts)(axnames[i].name, closure))
			return 0;
		break;
	    }

    if (p->axis != ROOT_AXIS)
	switch (p->nodetype) {
	case COMMENT_TYPE:
		if (!(*puts)("comment()", closure))
			return 0;
		break;
	case TEXT_TYPE:
		if (!(*puts)("text()", closure))
			return 0;
		break;
	case PI_TYPE:
		if (!(*puts)("processing-instruction(", closure))
			return 0;
		if (p->nodename.local)
		    if (!puts_wstr(p->nodename.local, puts, closure, 0))
			return 0;
		if (!(*puts)(")", closure))
			return 0;
		break;
	case NODE_TYPE:
		if (!(*puts)("node()", closure))
			return 0;
		break;
	case NAMETEST_TYPE:
		if (p->nodename.ns != NULL) {
		    if (!puts_wstr(p->nodename.ns, puts, closure, 0))
			return 0;
		    if (!(*puts)(":", closure))
			return 0;
		} 
		if (p->nodename.local == NULL) {
		    if (!(*puts)("*", closure))
			return 0;
		} else
		    if (!puts_wstr(p->nodename.local, puts, closure, 0))
			return 0;
		break;
	default:
		if (!(*puts)("?nodetype?", closure))
			return 0;
	}

    preds:

	for (i = 0; i < VECTOR_LEN(p->pred); i++) {
		if (!(*puts)("[", closure))
			return 0;
		if (!expr_put(VECTOR_AT(p->pred, i), puts, closure))
			return 0;
		if (!(*puts)("]", closure))
			return 0;
	}

    next:
	if (p->next) {
		if (!(*puts)("/", closure))
			return 0;
		return path_put(p->next, puts, closure);
	}
	return 1;
}

static int
value_put(v, puts, closure)
	struct value *v;
	int (*puts)(const char *, void *);
	void *closure;
{
	struct value t;
	int ret;

	switch (v->type) {
	case STRING:
		ret = puts_wstr((wstr_t)&v->v_string, puts, closure, 1);
		break;
	default:
		VALUE_INIT_STRING(t);
		asstring(v, &t);
		ret = puts_wstr((wstr_t)&t.v_string, puts, closure, 0);
		VALUE_FINI(t);
		break;
	}
	return ret;
}
