/* $Id: xpathP.h,v 1.8 2002/07/17 13:50:48 d Exp $ */
/* David Leonard, 2002. Public domain. */

/* forward declarations */
struct expr;
struct path;

typedef VECTOR_DECL(xml_t) nodeset_t;

/* A generic value type.. called an 'object' in the spec */
struct value {
	enum valuetype { NODESET, BOOLEAN, NUMBER, STRING } type;
	union {
		nodeset_t		u_nodeset;
		int			u_boolean;
		double			u_number;
		VECTOR_DECL(wchar_t)	u_string;
	} u;
};
#define v_nodeset	u.u_nodeset
#define v_boolean	u.u_boolean
#define v_number	u.u_number
#define v_string	u.u_string

#define VALUE_INIT_BOOLEAN(v, t) do {				\
		(v).type = BOOLEAN;				\
		(v).v_boolean = (t);				\
	} while (0)
#define VALUE_INIT_NUMBER(v, n) do {				\
		(v).type = NUMBER;				\
		(v).v_number = (n);				\
	} while (0)
#define VALUE_INIT_NODESET(v) do {				\
		(v).type = NODESET;				\
		VECTOR_INIT((v).v_nodeset);			\
	} while (0)
#define VALUE_INIT_STRING(v) do {				\
		(v).type = STRING;				\
		VECTOR_INIT((v).v_string);			\
	} while (0)

#define VALUE_FINI(v)	do {					\
	if ((v).type == NODESET || (v).type == STRING)		\
		VECTOR_FREE((v).v_nodeset);			\
    } while (0)

/* An axis in a path step */
enum axis {
	ROOT_AXIS,
	ANCESTOR_AXIS, ANCESTOR_OR_SELF_AXIS, ATTRIBUTE_AXIS, CHILD_AXIS,
	DESCENDANT_AXIS, DESCENDANT_OR_SELF_AXIS, FOLLOWING_AXIS,
	FOLLOWING_SIBLING_AXIS, NAMESPACE_AXIS, PARENT_AXIS,
	PRECEDING_AXIS, PRECEDING_SIBLING_AXIS, SELF_AXIS
};

#define IS_REVERSE_AXIS(ax) \
	((ax) == ANCESTOR_AXIS || \
	 (ax) == ANCESTOR_OR_SELF_AXIS || \
	 (ax) == PRECEDING_AXIS || \
	 (ax) == PRECEDING_SIBLING_AXIS)
#define AXIS_PRINCIPAL(ax) \
	((ax) == ATTRIBUTE_AXIS ? ATTR : \
	 (ax) == NAMESPACE_AXIS ? NS : \
	                          ELEMENT)

enum nodetype {
	NAMETEST_TYPE,
	COMMENT_TYPE,
	TEXT_TYPE,
	PI_TYPE,
	NODE_TYPE
};

/* Context used at evaluation */
struct context {
	struct xmldocinfo *doc;
	xml_t	node;
	int	position, size;
	struct xhash *variables;
	int	reverse;
	struct xmlns	*ns;
};

/* a function description */
struct function {
	const char *utf8_name;
	const char *  (*fn)(struct context *, 
			    struct value *, int, struct value *);
	const char *signature;
};

/* Lexical tokens - (single ASCII characters are also tokens) */
#define DCOLON		'a'	/* :: */
#define MULTIPLY	'b'	/* * */
#define LITERAL		'c'	/* "literal text" */
#define NUMERIC		'd'	/* 9999.9999 */
#define	AND		'e'	/* and */
#define OR		'f'	/* or */
#define MOD		'g'	/* mod */
#define DIV		'h'	/* div */
#define DSLASH		'i'	/* // */
#define NEQ		'j'	/* != */
#define LEQ		'k'	/* <= */
#define GEQ		'l'	/* >= */
#define NAMETEST	'm'	/* *, foo:bar, foo:* next_name */
#define FUNCNAME	'n'	/* next_func?? */
#define NODETYPE	'o'	/* next_nodetype */
#define AXISNAME	'p'	/* next_axis */
#define DPERIOD		'q'	/* .. */

/* Is a token an 'Operator' non-terminal in the grammar? */
#define isOperator(c) (		\
	    (c) == AND || 	\
	    (c) == OR || 	\
	    (c) == MOD || 	\
	    (c) == DIV || 	\
	    (c) == MULTIPLY || 	\
	    (c) == '/' || 	\
	    (c) == DSLASH || 	\
	    (c) == '|' || 	\
	    (c) == '+' || 	\
	    (c) == '-' || 	\
	    (c) == '=' || 	\
	    (c) == NEQ || 	\
	    (c) == '<' || 	\
	    (c) == LEQ || 	\
	    (c) == '>' || 	\
	    (c) == GEQ)

/* input state used when parsing an xpath expression */
struct input {
	/* Tokenizer input state */
	struct entity	*entity;
	wchar_t		 nextc;
	const char	*error;
	const char	*error_pos;
	wstr_t		 next_raw;

	int		 next;		/* next token type */
	wstr_t		 next_literal;	/* LITERAL */
	double		 next_number;	/* NUMBER */
	qname_t		 next_name;	/* NAMETEST */
	enum xmltype	 next_nodetype;	/* NODETYPE */
	enum axis	 next_axis;	/* AXISNAME */
	struct function	*next_function;	/* FUNCTION */

	int prev;
};

/* Expression node - to be evaluated in a context */
struct expr {
	enum exprtype {
		eUNION,	/* bin.exp1 | bin.exp2 */
		eOR,	/* bin.exp1 or bin.exp2 */
		eAND,	/* bin.exp1 and bin.exp2 */
		eEQ,	/* bin.exp1 = bin.exp2 */
		eNEQ,	/* bin.exp1 != bin.exp2 */
		eLT,	/* bin.exp1 < bin.exp2 */
		eGT,	/* bin.exp1 > bin.exp2 */
		eLEQ,	/* bin.exp1 <= bin.exp2 */
		eGEQ,	/* bin.exp1 >= bin.exp2 */
		eADD,	/* bin.exp1 + bin.exp2 */
		eSUB,	/* bin.exp1 - bin.exp2 */
		eMUL,	/* bin.exp1 * bin.exp2 */
		eDIV,	/* bin.exp1 div bin.exp2 */
		eMOD,	/* bin.exp1 mod bin.exp2 */
		eNEG,	/* - bin.exp1 */
		eVAR,	/* $ name */
		eLIT,	/* value */
		eFUNC,	/* func.name ( arg0, arg1 ... argn ) */
		ePATH,	/* expr/path */
		eFILTER,/* expr[expr] */
		eCONTEXT/* implicit expression before a path */
	} type;
	union {
		struct expr *unary;
		struct {
			struct expr *exp1, *exp2;
		} 	bin;
		qname_t name;
		struct value value;
		struct {
			struct function *func;
			VECTOR_DECL(struct expr *)arg;
		} func;
		struct {
			struct expr *expr;
			struct path *path;
		} path;
	} u;
};

/* Path step - evaluated in a context to generate a node-set */
struct path {
	enum axis	axis;		/* generator axis */
	enum nodetype	nodetype;	/* node type test */	
	qname_t		nodename;	/* node name match */
	VECTOR_DECL(struct expr *) pred;/* predicate exprs */
	struct path	*next;
};

struct xpath {
	struct xhash	*variables;
	struct expr	*expr;
};

typedef struct xpath_t *xpath;
